# -*- coding: utf-8 -*-
#
# Copyright 2020 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls_S (CoMPUter-based case fiLeS -
# Deutschlandstipendium).
#
# mpuls_S is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls_S is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls_S. If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple, defaultdict
from decimal import Decimal
from contextlib import contextmanager

from mpulsweb.controllers.formularpage import FormularpageController as \
    MpulsFormularpageController, PageRenderer
from mpulsweb.lib.formedrenderer import ViewRenderer as MpulsViewRenderer, \
    ViewPrintRenderer as MpulsViewPrintRenderer

from formed.instance.tree import IDENTIFIER_SEPARATOR
from formed.renderer.helpers import escape, tag
from formed.renderer.html import Item
from formed.meta.data import IntLeaf, ConditionalNode, RepeatNode
from formed.meta import UNKNOWN_INT, UNKNOWN_NUMERIC
from formed.meta.document import conditional_parents, containing_repeat_group
from formed.meta.semantic import convert, SemanticError, ErrorItem, \
    check_decimal_step


def conditional_parents_in_rg(nc):
    """Return a list of all conditionals containing rg up to the first RG"""
    items = []
    for item in reversed(nc.pathToRoot()[:-1]):
        if isinstance(item, ConditionalNode):
            items.append(item)
        elif isinstance(item, RepeatNode):
            break
    return items


class TabRGRow(object):

    """Description of row of a tabular repeatgroup for rendering"""

    def __init__(self, choice, description, rg, fields):
        self.choice = choice
        self.description = description
        self.rg = rg
        self.fields = fields

    def get_nc_id(self, data, nc):
        """Determine the to use for the nodecomponent in the form.

        data should be the FormData instance used to render the form. nc
        is expected to be one of the ncs in this row
        """
        id = data.get_id(nc)
        if self.rg is None:
            prefix, suffix = id.split(IDENTIFIER_SEPARATOR, 2)
            id = IDENTIFIER_SEPARATOR.join([prefix,
                                            "choice-%s" % (self.choice,),
                                            suffix])
        return id


class TabularRG(object):

    def __init__(self, rg_nc, document):
        self.rg_nc = rg_nc
        basename = rg_nc.getName()
        self.nc_choices = document.name2nc[basename + '_institution_art']
        self.nc_partner = document.name2nc[basename + '_partner']
        self.nc_schriftlich = document.name2nc[basename + '_schriftlich']
        self.option_nodes = [node for node in self.nc_choices.get_option_nodes()
                             if node.getValue() >= 0]

    def headings(self):
        return [self.nc_choices, self.nc_partner, self.nc_schriftlich]

    def legend_entries(self):
        return [self.nc_partner, self.nc_schriftlich]

    def rendering_description(self, data):
        instance_values = self.get_instance_values(data)
        rows = []
        for option in self.option_nodes:
            value = option.getValue()
            desc = option.getDescription()
            row = instance_values.get(value)
            if row is None:
                if option.isInvisible():
                    # This option is invisible (usually because it's an
                    # old option that should not be used in new cases,
                    # so if there's no existing instances (i.e. row is
                    # None) we ignore it here and do not render a row
                    # for it.
                    continue
                row = TabRGRow(value, desc, None,
                               ((self.nc_partner, None),
                                (self.nc_schriftlich, None)))
            else:
                row.description = desc
            rows.append(row)
        return rows

    def get_instance_values(self, data):
        values = {}
        for rg in data.get_repeat_group(self.rg_nc):
            data.push_node(rg)
            try:
                choice = data.get_value(self.nc_choices)
                partner = data.get_value(self.nc_partner)
                schriftlich  = data.get_value(self.nc_schriftlich)
                values[choice] = TabRGRow(choice, None, rg,
                                          ((self.nc_partner, partner),
                                           (self.nc_schriftlich, schriftlich)))
            finally:
                data.pop_node()
        return values


class ViewRenderer(MpulsViewRenderer):

    def render_nc(self, nc):
        rg_koop_set = ("rg_koop_1", "rg_koop_2", "rg_koop_3",
                       "rg_koop_4", "rg_koop_5")
        if nc.getName() in rg_koop_set:
            return self.render_koop_rg(nc)
        if nc.getName().startswith("row-grp-dauer-"):
            return self.render_grp_dauer(nc)
        return super(ViewRenderer, self).render_nc(nc)

    def format_numeric(self, nc, value):
        name = nc.getName()
        if (isinstance(value, Decimal)
            and value != UNKNOWN_NUMERIC
            and name in ("kjp_grp_dauer_stunden",
                         "nichtkjp_grp_dauer_stunden")):
            unit_nc_name = name.rpartition("_")[0] + "_einheit"
            unit_nc = self.data.get_meta_node(unit_nc_name)
            unit = self.data.get_value(unit_nc)
            if unit == 1:
                # 1 == doppelstunden, i.e. 1.5h
                value = value / Decimal("1.5")
            # The dauer fields are rendered with an input field that has
            # a type="number" attribute. In this case, some browsers
            # (e.g. Edge) require the value attribute to use '.' as a
            # decimal separator, not a ','. The base class method would
            # use ',' because the base class doesn't use type="number".
            # To force a '.' we do not use the base class method here.
            #
            # Also, use quantize to make sure the value always ends up
            # with an exponent of -2 to avoid sending a value of e.g. 80
            # as "8e+1" to the client.
            return str(value.quantize(Decimal(".01")))
        return super(ViewRenderer, self).format_numeric(nc, value)

    def render_koop_rg(self, nc):
        description = nc.getDescription()
        tabrg = self.preprocess_tabular_rg(nc)

        legend = []
        for entry in tabrg.legend_entries():
            legend.append('<small>%s: %s</small>' % (entry.getDescription(), entry.get_title()))
        out = []
        out.append('<div class="col-10">')
        out.append('<legend>%s %s</legend>'
                       % (escape(description),
                          "".join(self.render_annotation_links(nc))))
        out.append(", ".join(legend))
        out.append('<div class="table-responsive">')
        out.append('<table class="table table-sm table-hover custom-table-striped">')
        out.append('<thead>')
        out.append('<tr>')
        for field in tabrg.headings():
            out.append('<th scope="col">')
            out.append('<div class="text-nowrap">')
            out.append(escape(field.getDescription()))
            out.append('</div>')
            out.append('</th>')
        out.append('</tr>')
        out.append('</thead>')

        out.append('<tbody>')
        for row in tabrg.rendering_description(self.data):
            out.append('<tr>')
            out.append('<th scope="row" style="width: 80%">')
            out.append(escape(row.description))
            out.append('</th>')
            for field_nc, field_value in row.fields:
                out.append('<td style="width: 10%">')
                if row.rg is not None:
                    self.data.push_node(row.rg)
                try:
                    out.extend(self.render_koop_field(row, field_nc,
                                                      field_value))
                finally:
                    if row.rg is not None:
                        self.data.pop_node()
                out.append('</td>')
            out.append('</tr>')
        out.append('</tbody>')
        out.append('</table>')
        out.append('</div>')
        out.append('</div>')

        return Item("\n".join(out))

    def preprocess_tabular_rg(self, rg_nc):
        return TabularRG(rg_nc, self.data.formed)

    def render_koop_field(self, row, nc, value):
        if row.rg is not None:
            # If there's no RG instance for the row, it cannot have
            # values that would generate a warning, except for warnings
            # about required fields.
            self.check_rules(nc)
        if isinstance(nc, IntLeaf):
            return self.render_koop_field_int(row, nc, value)
        raise ValueError("Cannot render koop nc %r" % nc)

    @contextmanager
    def koop_read_only(self, nc, rg):
        """Determine whether the koop field nc should be read only

        The rg parameter should be the repeat group instance containing
        the field or None if the instance does not yet exist.

        The field is read only, if the render is already in read-only
        mode (self.read_only) or if any conditional containing nc
        between the nc and its containing repeat group evaluates to
        false. If the repeat group instance does not exist yet (rg is
        None) such conditionaly are considered false, mostly because
        that just happens to be sensible in the particular case of the
        fields we use this for.
        """
        read_only = self.read_only
        for conditional in conditional_parents_in_rg(nc):
            if read_only:
                break
            if rg is not None:
                editable = self.data.evaluate_conditional(conditional)
            else:
                editable = False
            read_only = not editable
        old_read_only = self.read_only
        try:
            self.read_only = read_only
            self.update_disabled_fields(nc)
            yield
        finally:
            self.read_only = old_read_only

    def render_koop_field_int(self, row, nc, value):
        if value == UNKNOWN_INT:
            value = self.gettext("unknown")
        id = row.get_nc_id(self.data, nc)
        with self.koop_read_only(nc, row.rg):
            css_classes = self.error_css_class(nc) if row.rg is not None else ""
            out = [
                tag('input', name=id, id=("f_" + id), value=value, type="text",
                    placeholder=nc.getPlaceholder(),
                    title=nc.get_title(),
                    class_= css_classes + " form-control form-control-sm",
                    disabled=self.read_only, **self.set_modification_attrs),
                ]
            self.render_tooltip(nc, out)
        return out

    def render_grp_dauer(self, nc):
        variant = nc.getName().rpartition("-")[-1]
        dauer_nc_name = variant + "_grp_dauer_stunden"
        dauer_nc = self.data.get_meta_node(dauer_nc_name)
        dauer_id = self.data.get_id(dauer_nc)
        unit_nc_name = variant + "_grp_dauer_einheit"
        unit_nc = self.data.get_meta_node(unit_nc_name)
        unit_id = self.data.get_id(unit_nc)
        unit_value = self.data.get_value(unit_nc)

        # make sure we handle rules including those for required fields
        # properly. This is normally done by render_nc, but we bypass
        # that here.
        self.check_rules(dauer_nc)
        self.update_disabled_fields(dauer_nc)
        self.check_rules(unit_nc)
        self.update_disabled_fields(unit_nc)

        out = []
        out.append(tag("div", class_="form-row"))
        out.append(tag("div", class_="form-group col-form col-8"))
        out.extend(self.render_label(dauer_nc, dauer_id))

        out.append(tag("div", class_="input-group-prepend"))
        out.append('<div class="input-group-text d-block text-left">')

        for child in unit_nc.getChildren():
            if child.isInvisible():
                continue
            child_value = child.getValue()

            # the child name is usually only unique within the radio
            # node, so prefix the id used in the HTML form with the
            # radio instance id so that it hopefully is unique on the
            # page.
            radio_id = unit_id + "." + child.getName()

            out.append('<div class="form-check">')
            out.append(tag('input', type="radio", id=radio_id, name=unit_id,
                           value=child_value, disabled=self.read_only,
                           checked=(child_value == unit_value),
                           tabindex=self.next_tabindex(),
                           class_=(self.error_css_class(nc)
                                   + " form-check-input position-static"),
                           **self.set_modification_attrs))
            out.append(tag("label", class_="form-check-label", for_=radio_id))
            out.append(tag("a", name="f_" + radio_id))
            out.append(escape(child.getDescription()))
            out.append("</a>")
            out.append("</label>")
            out.append("</div>")

        out.append("</div>")

        formatted = self.format_numeric(dauer_nc, self.data.get_value(dauer_nc))
        out.append(tag('input', name=dauer_id, id=dauer_id,
                       value=formatted, type="number", step="0.25",
                       placeholder=dauer_nc.getPlaceholder(),
                       class_=(self.error_css_class(dauer_nc)
                               + " form-control form-control-sm input-grp-dauer"),
                       disabled=self.read_only, **self.set_modification_attrs))

        out.append("</div>")
        self.render_tooltip(dauer_nc, out)
        out.append("</div>")
        out.append("</div>")

        return Item("".join(out))


class FormularpageController(MpulsFormularpageController):

    def _before_set_data_hook(self, case, form_result):
        """Apply application specific modififications to the form data.
        Here this means:

          - Reset mitarbeit_netzwerk if network doesn't exist

          - Handle the tabular repeatgroup instances
         """
        self._handle_zielgruppen_netzwerk(case, form_result)
        self._handle_tabular_repeatgroups(case, form_result)
        self._handle_grp_dauer(case, form_result)

    def _handle_zielgruppen_netzwerk(self, case, form_result):
        """Reset mitarbeit_netzwerk if network doesn't exist"""
        instance_tree = case.get_formed_instance_tree()
        networks = instance_tree.getAllItemsForName("zielgruppen_netzwerk")
        for network in networks:
            network_id = network.getIdentifier()
            # Note: The following condition matches the expression in
            # the conditional around mitarbeit_netzwerk in the
            # formedtree, so that this code resets mitarbeit_netzwerk
            # exactly when the conditional makes the field uneditable.
            #
            # We need to make sure that we only process fields that
            # actually are on the current page. We do that by checking
            # that the value for zielgruppen_netzwerk in form_result is
            # not None as that is only the case if the value was on the
            # page.
            if form_result.get(network_id) not in (None, "1", "2"):
                parent_ids = network_id.split(IDENTIFIER_SEPARATOR)[1:]
                mitarbeit_id = IDENTIFIER_SEPARATOR.join(["mitarbeit_netzwerk"]
                                                         + parent_ids)
                form_result[mitarbeit_id] = "-1"

    def _handle_tabular_repeatgroups(self, case, form_result):
        """Handle values of tabular rg fields."""
        # FIXME: We just delete all of them for now. We need to create
        # suitable RG instances instead.
        instance_tree = case.get_formed_instance_tree()

        # extract the items that refer to choices for which we did not
        # have RG isntances when rendering. This is stored in
        # new_rg_fields which maps RG-IDs that could be used to create
        # new instances (see the create_repeat_group_instance method of
        # the case objects) to dictionaries that in turn map the
        # institution_art options for which we need new instances to
        # lists with the fieldname and value of the other fields in that
        # RG instance
        new_rg_fields = defaultdict(lambda: defaultdict(list))
        for key, value in form_result.items():
            # the identifiers we're interested in look like:
            # rg_koop_1_partner:choice-5:138
            keyparts = key.split(IDENTIFIER_SEPARATOR, 2)
            if len(keyparts) == 3 and keyparts[1].startswith("choice-"):
                del form_result[key]
                nodename, encoded_choice, rgpath = keyparts
                choice = encoded_choice.partition("-")[-1]
                nc = instance_tree.formed.name2nc.get(nodename)
                if nc is not None:
                    rgname = containing_repeat_group(nc).getName()
                    rg_id = IDENTIFIER_SEPARATOR.join([rgname, rgpath])
                    new_rg_fields[rg_id][choice].append((nodename, value))

        # form_result now only holds the "normal" fields that do not
        # need special treatment here. So now we need to process the
        # items in new_rg_fields and create instances and add
        # corresponding entries to form_result so that the values will
        # be written to the databse with the other changes.
        form_result.update(self._create_tabular_instances(case, new_rg_fields))

    def _create_tabular_instances(self, case, new_rg_fields):
        """Create instances of the tabular RGs described in new_rg_fields"""
        # collects the new fields to insert
        new_formdata = {}

        instance_tree = case.get_formed_instance_tree()
        for rg_id, choice_fields in new_rg_fields.items():
            # gather information about the existing instances
            known_instances = {}
            rg_container = instance_tree.getItem(rg_id)
            choice_field_name = rg_container.meta.getName() + "_institution_art"
            for item in instance_tree.getAllItemsForName(choice_field_name):
                suffix = item.getIdentifier().split(IDENTIFIER_SEPARATOR, 1)[-1]
                known_instances[str(item.getValue())] = suffix

            # process instances that may have to be created
            for choice, fields in choice_fields.items():
                suffix = known_instances.get(choice)
                if suffix is None:
                    newrg = case.create_repeat_group_instance(rg_id)
                    suffix = newrg.getIdentifier().split(IDENTIFIER_SEPARATOR,
                                                         1)[-1]
                def formname(basename):
                    return IDENTIFIER_SEPARATOR.join([basename, suffix])
                new_formdata[formname(choice_field_name)] = choice
                for name, value in fields:
                    new_formdata[formname(name)] = value
        return new_formdata

    def _handle_grp_dauer(self, case, form_result):
        # Dictionary to keep track of the actual values the user entered
        # so that we can show exactly those when there are problems.
        # Since we do unit conversion for the *_grp_dauer* values, the
        # values stored in the the 'uncommited_fields' session entry,
        # are not the ones the user entered. The dictionary is
        # initialized here to make it's always set after the form data
        # has been processed.
        self._grp_dauer_original_values = {}

        # Dictionary to keep track of validation errors.
        self._grp_dauer_errors = {}

        instance_tree = case.get_formed_instance_tree()
        for name in ("kjp_grp_dauer_stunden", "nichtkjp_grp_dauer_stunden"):
            for nc in instance_tree.getAllItemsForName(name):
                if nc.getIdentifier() in form_result:
                    self._handle_dauer_conversion(case, form_result, nc)

    def _handle_dauer_conversion(self, case, form_result, dauer_nc):
        dauer_id = dauer_nc.getIdentifier()
        dauer_string = form_result[dauer_id].strip()

        if dauer_string:
            try:
                # Do not use checkAndConvert here, because
                # checkAndConvert would perform not only the conversion
                # from string to decimal, but also other checks, such as
                # range checks, before the unit conversion is applied,
                # leading to incorrect error messages. We do need to
                # perform the step check before the unit conversion, so
                # we do this here as well.
                dauer = convert(dauer_nc.meta, dauer_string)
                check_decimal_step(dauer, Decimal("0.25"))
            except SemanticError, err:
                # The value is not valid. libformed might accept it,
                # though (e.g. if only the step check failed), so remove
                # it from form_result.
                del form_result[dauer_id]
                # Record the error so that we can add it to the
                # "uncomited" fields later.
                self._grp_dauer_errors[dauer_id] = \
                    ErrorItem(ErrorItem.SEMANTIC, dauer_id, err.value,
                              dauer_string)
            else:
                basename, rg_ids = dauer_id.split(IDENTIFIER_SEPARATOR, 1)
                unit_id = basename.rpartition("_")[0] + "_einheit:" + rg_ids
                unit = form_result.get(unit_id)
                if unit is not None and unit.strip() == '1':
                    form_result[dauer_id] = str(Decimal("1.5") * dauer)
                    self._grp_dauer_original_values[dauer_id] = (dauer_string,
                                                                 dauer_nc.meta)

    def _save_uncomited_fields(self, uncommitted):
        for fieldname, (value, meta) in self._grp_dauer_original_values.items():
            for item in uncommitted.get(fieldname, []):
                if item.type == ErrorItem.SEMANTIC:
                    if u"kleiner" in item.msg:
                        item.msg = (u"%s Doppelstunden ist kleiner als der"
                                    u" zulässige Minimalwert %.2f Zeitstunden."
                                    % (value, meta.getMinValue()))
                    elif u"größer" in item.msg:
                        item.msg = (u"%s Doppelstunden ist größer als der"
                                    u" zulässige Maximalwert %.2f Zeitstunden."
                                    % (value, meta.getMaxValue()))
                    item.bad = value
        for fieldname, item in self._grp_dauer_errors.items():
            if u"nähesten" in item.msg:
                item.msg = (u"Bitte wählen sie einen gültigen Wert."
                            u" Eine Eingabe ist nur in"
                            u" 0,25er Schritten möglich.")
            uncommitted.setdefault(fieldname, []).append(item)
        super(FormularpageController, self)._save_uncomited_fields(uncommitted)

    def _get_formpage_renderer(self):
        pagerenderer = PageRenderer()
        pagerenderer.set_viewrenderer(ViewRenderer)
        return pagerenderer.get_rendered_page
