# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Torsten Irländer <torsten.irlaender@intevation.de>
# Sascha L. Teichmann <teichmann@intevation.de>

import logging
import datetime

from pylons.controllers.util import redirect, abort

import formencode

from mpulsweb.lib.translation import _

from mpulsweb.lib.base import BaseController, c, render, request, \
    render_htmlfill, restrict_to_POST, abort_unless_POST, session
from mpulsweb.lib.helpers import url
from mpulsweb.lib.security import checkRole
from mpulsweb.model.appointment import GlobalAppointmentOverview, \
    GlobalAppointmentFactory, MissingAppointmentError, CaseReminderFactory

from mpulsweb.lib.formedforms import FormedForm, Override

log = logging.getLogger(__name__)

def trigger_to_timedelta(meta, value):
    if value is None:
        return None
    return datetime.timedelta(value)

class AppointmentForm(FormedForm):

    def __init__(self, action):
        super(AppointmentForm, self).__init__(
                "appointment.xml", "appointment", action,
                modification_checks=True,
                overrides=[Override("trigger",
                                    to_form_result=trigger_to_timedelta)])


appointment_new_form = AppointmentForm("create_action")
appointment_edit_form = AppointmentForm("edit_action")

class AppointmentController(BaseController):

    def _load_global_appointment(self, id):
        try:
            return GlobalAppointmentFactory().loadById(id)
        except MissingAppointmentError:
            abort(404)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def index(self):
        redirect(url(controller='appointment', action='overview'))

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def overview(self):
        c.appointmentlist = GlobalAppointmentOverview()
        return render('/appointments/overview.mako')

    def _render_new_form(self, instance, errors):
        c.rendered_form = appointment_new_form.render(instance, errors)
        return render('/appointments/new.mako')


    @checkRole(('admin'))
    def create(self):
        instance = appointment_new_form.create_instance()

        cd = datetime.datetime.now()
        instance.set_value("start_date", cd.date())
        instance.set_value("start_time", cd.time())

        return self._render_new_form(instance, {})

    @checkRole(('admin'))
    @restrict_to_POST
    def create_action(self):
        instance, errors = appointment_new_form.validate_request(request.params)

        if errors:
            return self._render_new_form(instance, errors)

        form_result = instance.to_form_result()

        # Create appointment
        factory = GlobalAppointmentFactory()
        appointment = factory.createNew()
        appointment.setData(form_result)
        appointment.store()
        c.success_text = _(u"The appointment was successfully created.")
        return self.overview()

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def show(self, id):
        c.appointment = self._load_global_appointment(self._checkInt(id))
        return render('/appointments/show.mako')

    def _render_edit_form(self, instance, errors):
        c.rendered_form = appointment_edit_form.render(instance, errors)
        return render('/appointments/edit.mako')

    @checkRole(('admin'))
    def edit(self, id):
        app = self._load_global_appointment(self._checkInt(id))

        instance = appointment_edit_form.create_instance()

        instance.set_value('start_date', app.start_date.date())
        instance.set_value('start_time', app.start_date.time())
        instance.set_value('end_date', app.end_date.date())
        instance.set_value('end_time', app.end_date.time())
        instance.set_value('title', app.title)
        instance.set_value('id', app.id)
        instance.set_value('description', app.description)
        instance.set_value('trigger', app.getTriggerDays())
        instance.set_value('erledigt', app.erledigt)

        c.appointment_id = app.id
        return self._render_edit_form(instance, {})

    @checkRole(('admin'))
    @restrict_to_POST
    def edit_action(self):
        instance, errors = appointment_edit_form.validate_request(
                                                                 request.params)

        appointment_id = instance.get("id")

        if errors:
            c.appointment_id = appointment_id
            return self._render_edit_form(instance, errors)

        form_result = instance.to_form_result()

        factory = GlobalAppointmentFactory()

        appointment = self._load_global_appointment(appointment_id)
        appointment.setData(form_result)
        appointment.store()
        c.success_text = _(u"The appointment was successfully edited")
        return self.overview()

    @checkRole(('admin'))
    def delete(self, id, confirmed=False):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        if confirmed == 1:
            abort_unless_POST()
            appointment = self._load_global_appointment(id)
            appointment.delete()
            c.success_text = _(u"The appointment was successfully deleted.")
            return self.overview()
        else:
            c.context = "../main.mako"
            c.dialog_title = _(u'Delete appointment?')
            c.dialog_text = _(u"Do you really want to delete the appointment"
                              u"\nand irrevocably lose all the data?")
            c.url_yes = url(controller='appointment', action='delete',
                            id=id, confirmed=1)
            c.url_no = url(controller='appointment', action='overview')
            return render('/appointments/dialogs/confirm.mako')

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def save(self):
        return render('/appointments/new.mako')

    @restrict_to_POST
    @checkRole('cm')
    def set_done(self, id):
        rfactory = CaseReminderFactory()
        try:
            reminder = rfactory.loadById(self._checkInt(id))
            reminder.erledigt = True
            reminder.store()
        except MissingAppointmentError:
            abort(404)

        backlink = request.params.get("backlink")

        return self._handle_backlink(backlink)


    def _handle_backlink(self, backlink):
        """Handle backlink parameter for the set_done method.

        Note: The backlink parameter is an unvalidated request
        parameter! This method must handle it with care, here in derived
        classes that overwrite or extend this method.
        """
        session['success_text'] = _(u"The appointment was successfully set done.")
        session.save()
        redirect(url(controller='mpuls', action='index'))

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
