# -*- coding: utf-8 -*-
import logging

import re

import formencode

from datetime import datetime
from xml.sax import SAXParseException

from pylons import request, tmpl_context as c, app_globals as g, url
from pylons.controllers.util import redirect, abort

from webob.exc import HTTPNotFound, HTTPFound

from formed.instance.tree import SetDataException

from mpulsweb.controllers.case_bundle import build_case_list
from mpulsweb.controllers.formularpage import get_rendered_page
from mpulsweb.controllers.case_overview import CaseOverviewController
from mpulsweb.lib.base import DownloadController, render, session, h, \
    render_htmlfill, restrict_to_POST, abort_unless_POST
from mpulsweb.lib.db import PostgresDBInterface
from mpulsweb.lib.dialogs import checkPrivacyStatement, \
    render_privacy_statement_dialog_if_unsigned
from mpulsweb.lib.export import FormletterExport
from mpulsweb.lib.security import checkRole, check_feature, check_state
from mpulsweb.lib.translation import _, ungettext
from mpulsweb.lib.helpers import get_url_for_nc
from mpulsweb.lib.validators import SetStandinForm, SetEditorForm, NewCaseForm
from mpulsweb.model.agencysettings import Agency
from mpulsweb.model.case import ConsistenceCheckException, SessionCase, \
    MPulsXMLImporter, XMLImportException, \
    OPEN_STATE, CLOSED_STATE, OUTSTANDING_DELETION_STATE, \
    OUTSTANDING_ANONYMISATION_STATE, ANONYMISED_STATE
from mpulsweb.model.logbook import LogbookEntry
from mpulsweb.model.statement import PrivacyStatement
from mpulsweb.model.user import UserGroupList, UserListObject
from mpulsweb.lib.formedforms import FormedForm
from mpulsweb.controllers.logbook import LogbookEntryForm

log = logging.getLogger(__name__)

import_file_from = FormedForm("new_document.xml", "case", "import_xml_action")

keep_active_with_logbook_form = LogbookEntryForm("new_action")

class CaseController(DownloadController):

    FORMLETTER_FILE_NAME = "formletter"

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def index(self):
        case = session.get('case')
        if case:
            overview_url = url(controller='case_overview',
                                 action='index', anchor=case.id)
        else:
            overview_url = url(controller='case_overview',
                                 action='index')
        redirect(overview_url)

    def overview(self):
        return CaseOverviewController().index()

    @checkRole(('admin', 'cm', 'pt_dlr'))
    @checkPrivacyStatement()
    def select(self, id, confirmed=0):
        return self.digest(id)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def select_field(self, id, confirmed):
        # the parameter "confirmed" is actually the field name. It's
        # named "confirmed" so that the generic routing in mpuls works.
        field = confirmed
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        targeturl = get_url_for_nc(field, id)
        if targeturl is None:
            raise HTTPNotFound()
        c.targeturl = targeturl
        privacy_dialog = render_privacy_statement_dialog_if_unsigned(case)
        if privacy_dialog is not None:
            return privacy_dialog
        redirect(targeturl)

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def digest(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        if not check_state(case, {'admin': [OPEN_STATE, CLOSED_STATE,
                                            OUTSTANDING_DELETION_STATE,
                                            OUTSTANDING_ANONYMISATION_STATE,
                                            ANONYMISED_STATE],
                                  'cm': [OPEN_STATE, CLOSED_STATE]}):
            abort(401)

        sm = g.case_factory.get_state_machine()
        state = case.getState()
        c.case = case
        c.last_access = state.getAccessTime()
        c.daysToMaxSavetime = state.daysToMaxSavetime()
        c.maxsavetime = state.getMaxSavetime()
        c.finished = sm.is_end_state(state.getPhase())

        # Set success_text
        if 'success_text' in session:
            c.success_text = session['success_text']
            del session['success_text']
            session.save()

        return render('/casemanagement/digest_body.mako')

    #
    # Delete
    #

    @checkRole('cm')
    def markForDelete(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if not check_state(case, {'cm': [OPEN_STATE, CLOSED_STATE]}):
            abort(401)
        if confirmed == 1:
            abort_unless_POST()
            case.markDelete()
            case.checkValidity()
            c.success_text = _("Case document deleted! "
                               "The case document is given to the administrator"
                               " for deletion.")
            session['render_mode'] = 'ro'
            session.save()
            return self.overview()
        else:
            c.context = "../main.mako"
            c.dialog_title = ungettext('Delete case?', 'Delete cases?', 1)
            c.dialog_text = _("Do you really want to mark this case for"
                              " deletion?%s The case will not be editable"
                              " afterwards, and submited to the administration"
                              " for deletion.<br>") \
                % "\n".join(self._build_case_list([case]))

            c.url_yes = h.url(controller='case', action='markForDelete',
                              id=id, confirmed=1)
            c.url_no = h.url(controller='case_overview',
                             action='index', anchor=id)
            return render('/casemanagement/dialogs/confirm.mako')

    @checkRole('admin')
    def delete(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)

        case = self._loadCase(id)
        if not check_state(case, {'admin': [OUTSTANDING_DELETION_STATE]}):
            abort(401)
        if not confirmed:
            c.context = "../main.mako"
            c.dialog_title = ungettext('Delete case?', 'Delete cases?', 1)
            c.dialog_text = _("Warning! On deleting the case all data will"
                              " be erased irrecoverable!<br>"
                              "Only delete the case if this case is"
                              " <b>not needed for evaluations</b> anymore."
                              "<br><br>Do you want to delete the case "
                              "from the server?%s<br>") \
                % "\n".join(self._build_case_list([case]))
            c.url_yes = h.url(controller='case', action='delete',
                              id=id, confirmed=1)
            c.url_no = h.url(controller='case_overview',
                             action='index', anchor=id)
            return render('/casemanagement/dialogs/confirm_delete.mako')

        abort_unless_POST()
        try:
            return self._do_delete_case(case)
        except Exception:
            log.exception("Exception while trying to delete case %d", case.id)
            c.dialog_title = _("Case document not deleted!")
            c.dialog_text = _("The case can not be deleted because of an"
                              " unexpected problem.")
            c.url_ok = h.url(controller='case_overview',
                             action='index', anchor=id)
            return render('/casemanagement/dialogs/failed_delete.mako')

    def _do_delete_case(self, case):
        """Delete the case and return a HTML success or error message.
        Derived classes may override or extend this method if they need
        to handle specific exceptions raised by the case's delete method.
        """
        case.delete()
        c.success_text = _("Case document deleted!")
        return self.overview()

    #
    # Anonymize
    #
    @checkRole('cm')
    def markForAnonymizeFromBlackboard(self, id, confirmed):
        url = h.url('/')
        return self._markForAnonymize(id, confirmed, url)

    def markForAnonymizeFromReminderOverview(self, id, confirmed):
        url = h.url(controller='appointment', action='reminder_overview')
        return self._markForAnonymize(id, confirmed, url)

    @checkRole('cm')
    def markForAnonymizeFromOverview(self, id, confirmed):
        url = h.url(controller='case_overview', action='index', anchor=id)
        return self._markForAnonymize(id, confirmed, url)

    @checkRole('cm')
    def markForAnonymize(self, id, confirmed):
        check_feature('case-module', 'anonymize')
        url = h.url(controller='case', action='digest', id=id)
        return self._markForAnonymize(id, confirmed, url)

    def _markForAnonymize(self, id, confirmed, url):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if not check_state(case, {'cm': [OPEN_STATE, CLOSED_STATE]}):
            abort(401)
        # First check if the case is in a state which allows anonymization
        try:
            case.check_anonymizeability()
        except ConsistenceCheckException, e:
            c.dialog_title = _(u"""Anonymising not possible.""")
            c.dialog_text = h.literal(e.value)
            c.url_ok = url
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        # Show confirmation dialog if necessary.
        if not confirmed:
            # Get description of the phase which will be set after
            # anonymisation
            sm = g.case_factory.get_state_machine()
            ppid = sm.get_current_state_id(case)
            state = sm.get_end_state_for_state_id(ppid)
            phase_name = state.get_display_name()

            c.dialog_title = _(u"Do make anonymous?")
            c.dialog_text = (_(u"<p>Do you really want to mark the case for"
                               u" anonymization?%s<br>"
                               "The case can then no longer be"
                               u" edited and will given to the administrators"
                               u" for final anonymisation.</p>"
                               u"<p><b>The case document is going to be"
                               u" anonymised in phase </b>%s<b>!</b><br>""")
                             % ("\n".join(self._build_case_list([case])),
                                phase_name))
            c.url_yes = h.url(controller='case', action='markForAnonymize',
                              id=id, confirmed=1)
            c.url_no = url
            return render('/casemanagement/dialogs/confirm_markanonymize.mako')

        abort_unless_POST()
        # the action is confirmed, so call attempt to mark the case for
        # deletion.
        try:
            c.backlink = request.params.get("backlink", None)
            return self._do_mark_anonymize_case(case)
        except HTTPFound:
            # This is hackish: _do_mark_anonymize_case makes already
            # a redirect, but the HTTPFound linked to this is caught
            # by the general "except Exception". As we know the backlink,
            # we can reredirect ...
            redirect(c.backlink)
        except Exception:
            log.exception("Exception while trying to mark case %d for"
                          " anonymisation", case.id)
            c.dialog_title = _("Case document not marked for anonymisation!")
            c.dialog_text = _("The case can not be marked for anonymisation"
                              " because of an unexpected problem.")
            c.url_ok = h.url(controller='case_overview',
                             action='index', anchor=id)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

    def _do_mark_anonymize_case(self, case):
        """Mark the case for anonymization and handle SetDataException.
        Derived classes may override or extend this method if they need
        to handle more exceptions.

        The method returns a rendered HTML page with the success or
        error message.
        """
        try:
            case.markAnonymize()

            c.success_text = _(u"The case document has been marked for"
                              u" anonymisation and has been given to the"
                              u" administration for final anonymisation.")
            session['render_mode'] = 'ro'
            session.save()
            # if we have a backlink (which is not the digest):
            # Make the success_text available and do a redirect()
            if c.backlink and not re.search('digest', c.backlink):
                session['success_text'] = c.success_text
                session.save()
                redirect(c.backlink)
        except SetDataException, e:
            log.exception("SetDataException while trying to mark case %d for"
                          " anonymization", case.id)
            errors = []
            for k, values in e.errors.iteritems():
                for v in values:
                    errors.append(v.msg)
            c.dialog_title = _(u"Anonymisation not possible.")
            c.dialog_text = _(u"The case could not be anonymised for the"
                              u" following reason:</p>"
                              u"<p><i>%s</i></p>"
                              u"<p>Please click on OK to geht back to the"
                              u" overview.</p>""") % "<br>".join(errors)
            c.url_ok = h.url(controller='case', action='select',
                             id=case.id, confirmed=1)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        return self.overview()

    @checkRole('admin')
    def anonymize(self, id, confirmed):
        check_feature('case-module', 'anonymize')
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if not check_state(case, {'admin': [OUTSTANDING_ANONYMISATION_STATE]}):
            abort(401)

        # Check whether the case can actually be anonymized.
        try:
            case.check_anonymizeability()
        except ConsistenceCheckException, e:
            c.dialog_title = _(u"Anonymising not possible.")
            c.dialog_text = h.literal(e.value)
            c.url_ok = h.url(controller='case', action='digest', id=id)
            return render('/casemanagement/dialogs/failed_markanonymize.mako')

        # Show confirmation dialog if necessary.
        if not confirmed:
            c.context = "../main.mako"
            c.dialog_title = _(u"Do make anonymous?")
            c.dialog_text = _(u"Do you really want to anonymise the case and"
                              u" irrecovably lose the associated personal"
                              u" data? %s"
                              u" After it the case will not be available you."
                              u" The anonymised case will still be considered"
                              u" for the analysis.<br>") \
                % "\n".join(self._build_case_list([case]))
            c.url_yes = h.url(controller='case', action='anonymize',
                              id=id, confirmed=1)
            c.url_no = h.url(controller='case', action='digest', id=id)
            return render('/casemanagement/dialogs/confirm_anonymize.mako')

        abort_unless_POST()
        try:
            return self._do_anonymize_case(case)
        except Exception:
            log.exception("Exception while trying to anonymize case %s", id)
            c.dialog_title = _("Case document not anonymized!")
            c.dialog_text = _("The case can not be anonymized because of an"
                              " unexpected problem.")
            c.url_ok = h.url(controller='case_overview',
                             action='index', anchor=id)
            return render('/casemanagement/dialogs/failed_anonymize.mako')

    def _do_anonymize_case(self, case):
        """Anonymize the case and return a HTML success or error message.
        Derived classes may override or extend this method if they need
        to handle specific exceptions raised by the case's anonymize method.
        """
        case.anonymize()
        c.success_text = _(u"The personal data of the case document has been"
                           u" deleted and the case document is removed from"
                           u" the overview.")
        return self.overview()

    #
    # Restore
    #
    @checkRole('admin')
    def restore(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id)
        if not check_state(case, {'admin': [OUTSTANDING_ANONYMISATION_STATE,
                                            OUTSTANDING_DELETION_STATE]}):
            abort(401)

        if confirmed == 1:
            abort_unless_POST()
            case.restore()
            case.checkValidity()
            c.success_text = ungettext(
                u'Case restored!', u'Cases restored!', 1)
            return self.overview()
        else:
            c.context = "../main.mako"
            c.dialog_title = ungettext('Restore case?', 'Restore cases?', 1)
            c.dialog_text = _("Do you really want to restore the case?%s"
                              " The case will be editable afterwards again"
                              " for the users.<br>") \
                % "\n".join(self._build_case_list([case]))
            c.url_yes = h.url(controller='case', action='restore', id=id,
                              confirmed=1)
            c.url_no = h.url(controller='case_overview',
                             action='index', anchor=id)
            return render('/casemanagement/dialogs/confirm_restore.mako')

    #
    #
    #
    def _build_case_list(self, cases):
        """Format information about the list of cases.
        The return value is a list of strings with HTML.

        The default implementation simply calls
        mpulsweb.controllers.case_bundle.build_case_list. Derived
        classes may overwrite this method if they need different
        behavior.
        """
        return build_case_list(cases)


    #
    # Printing
    #
    @checkRole(('admin', 'cm', 'pt_dlr'))
    def printdigest(self, id):
        c.print_version = 1
        return self.digest(id)

    def _get_formpage_renderer(self):
        """Return a function for rendering a form page.
        The interface of that function is the same as that of the bound
        PageRenderer.get_rendered_page method from
        mpulsweb/controllers/formularpage.py

        The default implementation is exactly that method bound to a
        PageRenderer instantiated without arguments. Derived classes
        should overwrite this method if they need e.g. application
        specific behavior in their renderer.
        """
        return get_rendered_page

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def printAll(self, id, empty='0'):
        c.print_version = 1
        out = []
        case = self._loadCase(id, read_only=True)
        c.case = case
        c.statemachine = g.case_factory.get_state_machine()
        render_mode = session['render_mode']
        session['render_mode'] = 'ro'
        session.save()
        printed_headings = set()
        page_renderer = self._get_formpage_renderer()
        navigation = session.get('navigation.tree')
        if navigation:
            for item in navigation.getAllItems():
                # create the headings for all hierarchy elements that
                # have no page of their own.  We can do that by looking
                # at all the parents of the current item that have not
                # yet been used.
                headings = []
                parent = item.parent
                while parent is not None:
                    if parent not in printed_headings:
                        if parent.description:
                            # Ideally, we should be using different
                            # levels of hierarchies, but currently all
                            # headings are H1 headings
                            headings.append("<h3>%s</h3>\n"
                                            % parent.description)
                        printed_headings.add(parent)
                    parent = parent.parent
                out.extend(reversed(headings))

                out.append(page_renderer(item.key,
                                         case.get_formed_instance_tree()))
        c.formular = "".join(out)
        session['render_mode'] = render_mode
        session.save()
        return render('/formularpage/formular_print.mako')

    #
    # Create new case
    #

    def _createCase(self, form_result):
        case = g.case_factory.createNew(data=form_result)

        statement = case.getPrivacyStatement()
        # 3. Do we want to sign the statement?
        if form_result.get('statement_signed'):
            statement.sign()

        # Reload case to load navigation and set session correct. Loading
        # overhead here is not nice but not critical as creating new cases
        # should not happen so often
        return self._loadCase(case.id)

    @checkRole('cm')
    def new(self):
        session_case = SessionCase()
        session['case'] = session_case
        session.save()
        redirect(url('new2'))

    @checkRole('cm')
    def new2(self):
        session_case = session.get('case')
        if session_case is None:
            redirect(url('new'))
        else:
            return self._render_new(session_case)

    def _render_new(self, session_case, form_defaults=None, form_errors=None):
        if form_defaults is None:
            form_defaults = session_case.getData()
        if form_errors is None:
            form_errors = {}
        c.case = session_case
        form = render('/casemanagement/new.mako')
        return render_htmlfill(form, defaults=form_defaults,
                               auto_insert_errors=False, errors=form_errors)

    def _add_custom_fields(self):
        '''This function should be overwritten in specific applications to add
        custom fields like agencyname, and fkz to a new created case. On
        default this function returns an empty dictionary.'''
        return {}

    def _get_validator(self):
        '''This function should be overwritten in specific applications to add
        a custom validator'''
        return NewCaseForm()

    def _get_xmlimporter(self, xmlfile=None):
        '''This function should be overwritten in specific applications to add
        a custom XML Importer'''
        allow_documents = g.mpuls_config.is_enabled('case-module', 'documents')
        allow_logbook = g.mpuls_config.is_enabled('case-module', 'logbook')
        return MPulsXMLImporter(self._meta_tree_for_imports(),
                                allow_logbook=allow_logbook,
                                allow_documents=allow_documents)

    def _meta_tree_for_imports(self):
        """Return the meta tree for imports via self._get_xmlimporter.
        The default implementation simply returns g.formedTree. Derived
        classes may overwrite this to e.g. use a different set of rules
        when validating imports.
        """
        return g.formedTree

    def _convert_SetDataException2Formencode(self, err):
        """Convert a SetDataException into a formencode compatible dictionary"""

        # Expample SetDataException:
        # {'geburtsdatum:17': [type: 1, name: 'geburtsdatum:17', msg:
        # u"'Geburtsdatum' darf zeitlich nicht nach 'Datum Erstgespr\xe4ch'
        # liegen.", bad:  datetime.date(1990, 1, 1)], 'datum_ea_start:17':
        # [type: 1, name: 'datum_ea_start:17', msg:  u"'Geburtsdatum' darf
        # zeitlich nicht nach 'Datum Erstgespr\xe4ch' liegen.", bad:
        # datetime.date(1981, 12, 12)]}
        errors = {}
        for e in err.errors.keys():
            key = e.split(':')[0]
            errors[key] = err.errors.get(e)[0].msg
        return errors

    @checkRole('cm')
    @restrict_to_POST
    def newAction(self):
        # TODO: Use htmlfill here!
        session_case = SessionCase()
        session_case.setData(request.params)

        session['case'] = session_case
        session.save()

        # Do we want to print the statement?
        if "print" in request.params:
            c.print_version = 1
            statement = PrivacyStatement(session_case, Agency())
            c.content = statement.fillout()
            return render('privacy/default_statement.mako')

        # Or do we want to store the new case?
        elif "save" in request.params:
            validator = self._get_validator()
            try:
                form_result = validator.to_python(request.params)
                # Add custom fields to the form_result which are not part of
                # the new case formular
                form_result.update(self._add_custom_fields())
                # Create a new case
                case = self._createCase(form_result)
                # Call select to open the case, set confirmed to
                # suppress privacy check
                return self.select(str(case.id),
                         confirmed=\
                         g.mpuls_config.get_suppress_privacy_check_on_create()
                       )
            except SetDataException, err:
                errors = self._convert_SetDataException2Formencode(err)
                return self._render_new(session_case, form_result, errors)
            except formencode.Invalid, error:
                log.debug("Invalid form data: %r", error)
                return self._render_new(session_case, error.value,
                                        error.error_dict or {})

    @checkRole('cm')
    def copy(self, id, confirmed):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)
        case = self._loadCase(id, read_only=True)
        if confirmed:
            abort_unless_POST()
            case.copy()
            case.checkValidity()
            c.dialog_title = _('Case copied!')
            c.dialog_text = _('The case was succussfully copied and is now'
                              ' available in the case overview.'
                              ' Click on "OK" to open the case overview.')
            c.url_ok = h.url(controller='case_overview',
                             action='index', anchor=id)
            return render('/casemanagement/dialogs/success_copy.mako')
        else:
            c.dialog_title = _('Copy case?')
            c.dialog_text = _('Do you really want to copy this case?'
                              ' All data will be copied in the new case.')
            c.url_yes = h.url(controller='case', action='copy',
                              id=id, confirmed=1)
            c.url_no = h.url(controller='case', action='digest', id=id)
            return render('/casemanagement/dialogs/confirm_copy.mako')

    def _render_import_file(self, instance, errors):
        c.rendered_form = import_file_from.render(instance, errors)
        return render('/casemanagement/upload.mako')

    @checkRole("cm")
    def importCase(self):
        instance = import_file_from.create_instance()
        return self._render_import_file(instance, {})

    @checkRole('cm')
    def import_xml_action(self):
        try:
            xmlfile = request.POST['file']
            try:
                importer = self._get_xmlimporter(xmlfile.file)
                importer.import_xml(PostgresDBInterface(), xmlfile.file)
            finally:
                xmlfile.file.close()

            c.success_text = _("The file was imported successfully.")
            return self.overview()
        except Exception, e:
            log.exception("Exception during xml import")
            c.dialog_title = _("Case(s) not imported!")
            if isinstance(e, XMLImportException):
                c.dialog_text = h.escape(e.message)
            elif isinstance(e, SAXParseException):
                c.dialog_text = _("The import file could not be parsed."
                                  " Please check that the import file contains"
                                  " valid XML")
            else:
                c.dialog_text = _("An unknown error occurred during the import."
                                  " Click 'OK' to return to the import dialog.")
            c.url_ok = h.url(controller='case', action='importCase')
            return render('/casemanagement/dialogs/failed_xml_import.mako')

    #
    # ORGANISATION (standin and editor settings)
    #

    @checkRole(('admin', 'cm'))
    def organisation(self, id):
        id = self._checkInt(id)
        ulist = UserListObject()
        list = UserGroupList()
        c.case = case = self._loadCase(id, read_only=True)
        c.standin = c.case.getStandin()
        c.editor = c.case.getEditor()
        c.current_user = session['USER_AUTHORIZED']

        # group list for standin
        filter = [u.gid for u in ulist.getAdminList()]
        filter.append(c.editor.gid)
        filter2 = filter[:]
        filter.append(c.current_user.gid)
        c.group_list = list.getGroupList(filter)
        c.group_list2 = list.getGroupList(filter2)

        # userlist for editor
        filter3 = [u.id for u in ulist.getAdminList()]
        filter3.append(c.current_user.id)
        filter3.append(c.editor.id)
        c.user_list = ulist.getUserList(filter3)

        self._organisation_extra(case)

        return render('/casemanagement/organisation.mako')

    def _organisation_extra(self, case):
        """Perform actions related to casemanagement/organisation_extra.mako.

        The default implementation does nothing because the template
        does nothing. Applications may overwrite this method when they
        need to e.g. add more informations to the template context for
        the organisation_extra.mako template.
        """

    @checkRole('admin')
    @restrict_to_POST
    def setEditorAction(self):
        validator = SetEditorForm()
        try:
            form_result = validator.to_python(request.params)
            case = self._loadCase(form_result.get('case_id'))
            case.setEditor(form_result.get('editor'))
            c.success_text = h.status_message(_("Editor successfully set."))
            return self.organisation(form_result.get('case_id'))
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.status_msg = h.status_message(_("Error! Editor cannot be set."))
            return render('/casemanagement/organisation.mako')

    @checkRole(('admin', 'cm'))
    @restrict_to_POST
    def setStandinAction(self):
        validator = SetStandinForm()
        try:
            form_result = validator.to_python(request.params)
            case_id = form_result["case_id"]
            case = self._loadCase(case_id)
            standin = case.getStandin()
            standin.setGroups(form_result["standin"])
            c.success_text = h.status_message(_("Standin successfully set."))
            return self.organisation(case_id)
        except formencode.Invalid, error:
            c.form_result = error.value
            c.form_errors = error.error_dict or {}
            c.status_msg = h.status_message(_("Error! Standin is not set."))
            return render('/casemanagement/organisation.mako')

    @checkRole('cm')
    @checkPrivacyStatement('keepActive')
    def keepActive(self, id, confirmed):
        # If logbook is enabled let the user give the reason why he keeps the
        # case active.
        if g.mpuls_config.is_enabled('case-module', 'logbook'):
            return self._keep_active_with_logbook(id)
        # Just refresh last access date.
        else:
            return self._keep_active_without_logbook(id)

    def _keep_active_with_logbook(self, id):
        entry = LogbookEntry()
        entry.setDate(datetime.now())
        entry.setKind(-2)

        instance = keep_active_with_logbook_form.create_instance()
        instance.set_value("case_id", id)
        now = datetime.now()
        instance.set_value("date", now.date())
        instance.set_value("time", now.time())
        instance.set_value("art", entry.get_kinds())
        instance.set_value("short_notice", _('Case storage extended'))
        instance.set_value("notice", _('Please enter the reason for the case'
                                       ' storage extension.'))
        instance.set_value("bearbeiter",
                           session.get('USER_AUTHORIZED').shortenedName())
        instance.set_value("show_file_upload", True)

        c.entry = entry

        return self._render_new_entry_form(instance, {})

    def _render_new_entry_form(self, instance, errors):
        c.rendered_form = keep_active_with_logbook_form.render(instance, errors)
        c.conditional_form_info_url = url(controller='logbook', action='new_js')
        return render('/logbook/new.mako')

    def _keep_active_without_logbook(self, id):
        id = self._checkInt(id)
        case = self._loadCase(id)
        try:
            # Temporarily switch the state of the case to force
            # refreshing the last access date.
            state = case.getState().getState()
            case.getState().setState(2)
            case.getState().setState(state)
            c.dialog_title = _('Case storage extended')
            c.dialog_text = _('The max case storage has been extended.'
                              ' Please click on OK to continue')
            c.url_ok = h.url("/")
            return render('/casemanagement/dialogs/success_keepactive.mako')
        except:
            log.exception('Can not refresh case in'
                          ' "_keep_active_without_logbook"')
            raise

        pass

    # Export
    def get_formletter_file_name(self):
        return self.FORMLETTER_FILE_NAME

    def formletter(self, id, letterid=1):
        id = self._checkInt(id)
        case = self._loadCase(id, read_only=True)
        d, o = case.formletter(letterid)
        export = FormletterExport([d], o).export()
        return self._download(export, "application/msexcel", extension="xls",
                              name=self.get_formletter_file_name())


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
