# -*- coding: utf-8 -*-
# Copyright 2012 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of MPULS (CoMPUter-based case fiLeS).
#
# MPULS is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# MPULS is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with MPULS. If not, see <http://www.gnu.org/licenses/>.

import json
from datetime import datetime

import logging

import formencode
from formencode import Invalid
from formencode.validators import String, Int

from pylons import app_globals as g
from pylons.controllers.util import abort, redirect

from mpulsweb.lib.helpers import url
from mpulsweb.lib.base import c, render, request, session, BaseController, \
    render_htmlfill
from mpulsweb.lib.security import checkRole, check_feature
import mpulsweb.lib.sqlexpr as E
from mpulsweb.lib.searchexpr import And, Field, Between, Choice, \
     Status, Responsibilities, Needle, FederalState, Editor, Radio, Bool, \
     searchexpr_from_json, DayTimestamp, Updated
from mpulsweb.lib.helper.searchfield import dump_all_fields, nc_type_name, \
     extra_info, is_special_field, description_to_name, LOGBOOK_EXTRAS
from mpulsweb.lib.translation import _
from mpulsweb.lib.validators import FormedDateChecker
import mpulsweb.lib.helpers as h
import mpulsweb.lib.helper.filters as F
from mpulsweb.lib.customoverview import Column, Overview
from mpulsweb.model.user import UserListObject
from mpulsweb.model.case import MpulsCaseOverview
from mpulsweb.model.case import ExtendedCaseOverview
from mpulsweb.model.savedsearch import SavedSearch
from mpulsweb.model.customoverview import OverviewConfiguration


log = logging.getLogger(__name__)


def get_sort(params):
    '''Return sort field and sort order. First try to return a userdefined sort
    and order. If user does not provide one of there params, try to return the
    safed params in session. If session is also empty return default params.
    '''
    session_options = session.get('CASE_OVERVIEW_SEARCHOPTIONS', {})
    sort = params.get('sort_field',
                      session_options.get('sort_field',
                                          g.mpuls_config.get('search',
                                                             'default-sorting')
                                          ))
    order = params.get('sort_order',
                       session_options.get('sort_order',
                                           g.mpuls_config.get('search',
                                                              'default-order')
                                           ))
    return sort, order


def parseSearchOptions(options, extended):
    return g.case_search.convert_form_parameters(options,
                                                 session['USER_AUTHORIZED'],
                                                 extended=extended)


def search_options_from_session():
    options = session.get('CASE_OVERVIEW_SEARCHOPTIONS')
    if options:
        return options

    user = session['USER_AUTHORIZED']
    if user.isAdmin():
        return g.mpuls_config.get('search', 'admin_defaults')[0]
    else:
        return g.mpuls_config.get('search', 'cm_defaults')[0]


def get_session_search():
    return session.get("search.ast")


def put_session_search(expr):
    session["search.ast"] = expr
    session.save()


def is_flexible():
    ''' Returns true if the advacned-search is enabled '''
    return g.mpuls_config.is_enabled("case-module", "flexible-search")


class CaseOverviewController(BaseController):

    _predefined_overviews = {}

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def index(self):
        return self.overview()

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def overview_extended(self):
        extended = session.get('EXTENDED_SEARCH')
        if extended is not None:
            extended = not extended
        else:
            extended = True
        session['EXTENDED_SEARCH'] = extended
        session.save()
        return self.overview()

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def overview(self):
        # Set success_text
        if 'success_text' in session:
            c.success_text = session['success_text']
            del session['success_text']
            session.save()

        if 'case' in session:
            del session['case']

        if 'casebundle' in session:
            del session['casebundle']

        session.save()

        if is_flexible():
            return self._flexible_overview()
        return self._overview()

    @checkRole(('cm', 'admin'))
    def reset(self):
        """Reset the search parameters to the default"""
        check_feature("case-module", "flexible-search")
        put_session_search(self._get_default_search())
        redirect(url(controller='case_overview', action='overview'))

    @checkRole(('admin', 'cm', 'pt_dlr'))
    def search(self, search_options=None):
        if is_flexible():
            return self._flexible_search()
        else:
            return self._search(search_options)

    def _overview(self):
        # If reset parameter is set ti != 0 the delete all saved params and
        # cases on overview page
        reset = self._checkBool(request.params.get('reset', '0'))
        if reset:
            session['CASE_OVERVIEW_SEARCHOPTIONS'] = {}
            session['EXTENDED_SEARCH'] = False
            session.save()

        # Get saved search params from session.
        search_options = search_options_from_session()
        # Update search options with ordering
        sort, order = get_sort(request.params)
        search_options['sort_order'] = order
        search_options['sort_field'] = sort
        session['CASE_OVERVIEW_SEARCHOPTIONS'] = search_options
        session.save()
        return self.search(search_options)

    def _flexible_overview(self):
        check_feature("case-module", "flexible-search")
        if get_session_search() is None:
            put_session_search(self._get_default_search())

        search_options = {}
        sort, order = get_sort(request.params)
        search_options['sort_order'] = order
        search_options['sort_field'] = sort
        session['CASE_OVERVIEW_SEARCHOPTIONS'] = search_options
        session.save()

        search_ast = get_session_search()
        cases = self._getCaseListNg(search_ast)
        return self._render_flexible_overview(cases, search_ast, {})

    def _unknown_field(self, name):
        search_ast = get_session_search()
        cases = self._getCaseListNg(search_ast)
        msg = _("The field '%s' does not exist or is not searchable.") % name
        return self._render_flexible_overview(cases, search_ast, {}, msg)

    def _add_field(self, desc):
        """Add a field to the session with the current search parameters"""

        name = description_to_name(desc)
        if not name or name in self._get_field_filter():
            return self._unknown_field(desc)

        check_feature("case-module", "flexible-search")
        meta_tree = g.formedTree

        ast = get_session_search()

        if is_special_field(name):
            expr = ast.add_special_field(name)
        else:
            if name not in meta_tree.name2nc:
                abort(404)
            expr = ast.add_formed_field(meta_tree, name)

        put_session_search(expr)

        redirect(url(controller='case_overview', action='overview'))

    def _delete_field(self, name):
        """Delete a field from the current search parameters"""
        check_feature("case-module", "flexible-search")
        ast = get_session_search()
        put_session_search(ast.delete_form_field(name))

        redirect(url(controller='case_overview', action='overview'))

    def _flexible_search(self):
        '''
        Redirects to search overview to perform the search
        '''
        check_feature("case-module", "flexible-search")
        old_ast = get_session_search()
        if old_ast is None:
            # no old AST in session dictionary. Redirect to default case
            # overview. It's tempting to use a default AST (e.g. from
            # (self._get_default_search()) but that likely wouldn't
            # work, because we need some request parameters which likely
            # won't exist.
            redirect(url(controller='case_overview', action='overview'))

        (search_ast, ast_errors) = self._update_ast(old_ast)
        put_session_search(search_ast)

        if "perform_search" in request.params or \
            ("add_field" in request.params and
             not request.params.get("new_field_description")):
            cases = self._getCaseListNg(search_ast)
            return self._render_flexible_overview(cases, search_ast, ast_errors)
        elif "to_json" in request.params:
            self._save_ast(search_ast)
        elif "add_field" in request.params:
            desc = request.params.get("new_field_description")
            return self._add_field(desc)
        else:
            for param in request.params:
                if param.startswith("delete_field_"):
                    field = param[13:]
                    if field:
                        return self._delete_field(field)

        # we should not get here in normal use. If we do, the request is
        # missing some important information in the request parameters,
        # like the action to perform (perform_search, to_json, etc).
        abort(404)

    def select_overview(self, overview_id):
        overview_id = self._checkInt(overview_id)
        config = self._load_overview_configuration(overview_id)
        if config is None:
            overview_id = -1
        session["selected_overview"] = overview_id
        session.save()
        redirect(url(controller='case_overview', action='overview'))

    def _save_ast(self, search_ast):
        '''
        Save the current search.
        First jsonify it, then create a name for thes saved search,
        save it and redirect to the savedsearches list.
        '''
        name = datetime.now().strftime(_('Search %Y-%m-%d %H:%M'))
        saved_search = SavedSearch(None, name, g.formedTree.getVersion(),
                                   search_ast).new()
        redirect(url(controller='savedsearch', action='overview'))

    def _getCaseListNg(self, search_options):
        self._set_overview_columns()
        cases = ExtendedCaseOverview()
        cases.search(search_options,
                     sort_params=session['CASE_OVERVIEW_SEARCHOPTIONS'],
                     retrieve_fields=c.overview_columns)
        return cases

    def _getCaseList(self, search_options):
        cases = MpulsCaseOverview()
        log.debug('search options: %s' % search_options)
        if search_options:
            options = self.parseSearchOptions(
                search_options, extended=session.get("EXTENDED_SEARCH"))
            cases.search(options)
            if hasattr(self, "get_pages_count"):
                cases.set_count(self.get_pages_count(options))
        return cases

    def _validate_for_convert(self, field, field_type):
        map1 = {'date': DateConvertValidator,
                'choice': ChoiceConvertValidator,
                'int': IntConvertValidator,
                'bool': BoolConvertValidator,
                'radio': RadioConvertValidator,
                }
        validator_cls = map1.get(field_type, ConvertValidator)
        return validator_cls(field).validate()

    def _update_ast(self, searchexpr):
        fields = []
        errors = {}
        for expr in searchexpr.exprs:
            try:
                new_expr = self._update_ast_field(expr)
            except Invalid as e:
                new_expr = expr
                errors[expr.form_key()] = e
            fields.append(new_expr)

        return (And(fields), errors)

    _field_operator_map = {'EQ': E.EQ, 'LT': E.LT, 'LE': E.LE, 'GT': E.GT,
                           'GE': E.GE,
                           'Startwith': E.Startwith,
                           'Substring': E.Substring,
                           'IS_NULL': E.IS_NULL,
                           }

    def _update_ast_field(self, expr):
        if isinstance(expr, Needle):
            value = self._validate_for_convert("search_str", "text")[0]
            return Needle(['m.name', 'm.vorname'], value)
        elif isinstance(expr, Responsibilities):
            value = self._validate_for_convert("responsibilities", "choice")[0]
            if not value:
                return Responsibilities(False, False, False)
            own = 1 in value
            standin = 2 in value
            changed_editor = 3 in value
            return Responsibilities(own, standin, changed_editor)
        elif isinstance(expr, Status):
            return Status(self._validate_for_convert("status", "choice")[0])
        elif isinstance(expr, Editor):
            return Editor(self._validate_for_convert("bearbeiter_id",
                                                     "choice")[0])
        elif isinstance(expr, Updated):
            value, value2, type1 = self._validate_for_convert("updated", "date")
            return Updated(value, self._field_operator_map[type1])
        elif hasattr(expr, "fieldname"):
            # A generic field. Assume that expr.fieldname is the name
            # refers to a field in the formedtree.
            fieldname = expr.fieldname

            if fieldname in LOGBOOK_EXTRAS:
                value, value2, type1 = self._validate_for_convert(fieldname,
                    LOGBOOK_EXTRAS[fieldname]["type"])
                if fieldname == "datum" and not type1 == "Between":
                    # The datum logbook entry is searched like a date type
                    # but is a timestamp in the database.
                    return DayTimestamp(fieldname, value,
                                        self._field_operator_map[type1])
                elif fieldname == "art":
                    assert isinstance(expr, Choice), \
                           "logbook field art is expected to be a Choice instance"
                    return Choice(fieldname, value, multiselect=True)
            else:
                # It's a formed field or something sufficiently like it
                if fieldname == "id":
                    type_name = "int"
                else:
                    type_name = nc_type_name(g.formedTree.name2nc[fieldname])
                value, value2, type1 = self._validate_for_convert(
                    fieldname, type_name)

            if isinstance(expr, Choice):
                return Choice(fieldname, value)
            elif isinstance(expr, Radio):
                return Radio(fieldname, value)
            elif isinstance(expr, Bool):
                return Bool(fieldname, value)
            elif isinstance(expr, (Field, Between)):
                if type1 == "Between":
                    return Between(fieldname, value, value2)
                else:
                    return Field(fieldname, value,
                                 self._field_operator_map[type1])
            else:
                raise ValueError("Cannot handle searchexpr field: %r" %
                        (expr,))
        else:
            raise ValueError("Cannot handle searchexpr: %r" % (expr,))

    def _get_default_search(self):
        """Return the default search criteria."""

        user = session['USER_AUTHORIZED']
        if user.isAdmin():
            cfg = g.mpuls_config.get('search', 'admin_defaults')[0]
        else:
            cfg = g.mpuls_config.get('search', 'cm_defaults')[0]

        return searchexpr_from_json(json.dumps(cfg))


    def _prepare_overview(self, cases):
        """ Prepare the session / c object to render the overview """
        user = session['USER_AUTHORIZED']

        # Values used in the template
        c.extended_search = session.get('EXTENDED_SEARCH', False)
        ulo = UserListObject()
        filter1 = [u.id for u in ulo.getAdminList()]
        c.userlist = ulo.getUserList(filter1)
        c.statemachine = g.case_factory.get_state_machine()

        # count num of cases
        c.count_all = len([x for x in cases.getDatasets()])
        c.count_own = len([x for x in cases.getDatasets()
                           if x.user_is_editor(user.id)])
        c.count_standin = c.count_all - c.count_own
        c.count_anon = len([x for x in cases.getDatasets()
                            if x.getState().getState() == 5])
        c.count_markanon = len([x for x in cases.getDatasets()
                                if x.getState().getState() == 4])
        c.count_del = len([x for x in cases.getDatasets()
                           if x.getState().getState() == 3])

        # Hide evaluation option if someone did a search for a
        # particular editor
        c.hide_evaluation = False

        session["id_field_candidate"] = "m.id"
        c.all_ids = " ".join(unicode(case.id) for case in cases.getDatasets())

        c.cases = cases

    def _get_additional_fields(self):
        """Override this in a subclass to add new searchable fields.
        This function should return a list of the usual field description
        dictionaries. See mpulsweb.lib.helper.searchfield for examples.
        """
        return []

    def _get_field_filter(self):
        """Determine which search criteria should not be offered to the user.
        The return value is a list of search criteria names.

        Derived classes should extend this method by calling the base
        class method and adding new items to the list if they need to
        hide more criteria.
        """
        user = session['USER_AUTHORIZED']
        hide = []
        if not user.isAdmin():
            # only admins may search based on the editor of a case
            hide.append("bearbeiter_id")
        if user.isAdmin():
            # Admins shall not search in logbooks.
            hide.extend(LOGBOOK_EXTRAS.keys())
        if user.getPermissionRole() != "cm":
            # Searching for cases edited by the user or for which the
            # user is a standin ("responsibilities" in STATIC_EXTRAS) is
            # only useful for CM users, because other users cannot be
            # editor or standin.
            hide.append("responsibilities")
        return hide

    def _get_flexible_search_fields(self):
        """ Returns the fields that should be available for flexible
        search. For documentation on the format see dump_all_fields """
        return json.dumps(self._get_additional_fields() + \
                dump_all_fields(self._get_field_filter()))

    def _get_search_field_description(self, fieldname):
        return extra_info(g.formedTree, fieldname)

    def _get_flexible_search_exprs(self, search_ast, ast_errors):
        """ Returns a list of expr dict tuples of the search expressions
        in the search_ast with their respecitve extra info dicts.
        """
        exprs = []
        for expr in search_ast.exprs:
            key = expr.form_key()
            desc = self._get_search_field_description(key)
            exprs.append((expr, desc.with_error(ast_errors.get(key))))
        return exprs

    def _list_overview_configurations(self):
        return OverviewConfiguration.list_custom_overviews()

    def _load_overview_configuration(self, overview_id):
        config = OverviewConfiguration.get_custom_overview_by_id(overview_id)
        if config is not None:
            config = config.as_overview()
        return config

    def _set_overview_columns(self):
        overview_id = session.get("selected_overview")
        #overview = self._predefined_overviews.get(overview_id)
        overview = self._load_overview_configuration(overview_id)
        if overview is None:
            overview_id = -1
            columns = g.case.get_overview_fields_list()
            columns.extend([
                Column("bearbeiter_id", _('Editor'), "table_name",
                       ["admin"],
                       lambda case: "%s, %s" % (F.NA(case.getEditor().last_name),
                                                F.NA(case.getEditor().first_name[:1]))),
                Column("zugriff", _('Updated'), "table_name", ["cm"],
                       lambda case: h.format_date(case.state.getAccessTime())),
                Column("id", "ID", "table_width_mid_s", ["cm", "admin"],
                       lambda case: F.NA(case.id)),
                ])
        else:
            columns = overview.columns
        c.overview_columns = columns
        c.overview_id = overview_id
        overviews = self._list_overview_configurations()
        overviews.insert(0, OverviewConfiguration(-1, u"Standardübersicht", []))
        c.available_overviews = overviews

    def _render_flexible_overview(self, cases, search_ast, ast_errors,
                                  add_field_error=None):
        self._prepare_overview(cases)

        c.formed_fields = self._get_flexible_search_fields()
        c.search_exprs = self._get_flexible_search_exprs(search_ast, ast_errors)

        # Pass all saved searches to the template
        c.saved_searches = SavedSearch.list_saved_searches()

        # The usual htmlfill rendering breaks the scripted form building /
        # filling done in the flexible search. So we use this hack to
        # inject whats basically a validation error for the add_field action
        c.add_field_error = add_field_error

        self._set_overview_columns()

        # Hide evaluation option if someone did a search for a
        # particular editor
        for expr, extras in c.search_exprs:
            if isinstance(expr,Editor) and expr.values != None:
                c.hide_evaluation = True

        return render('/casemanagement/overview.mako')

    def _renderOverview(self, cases, form_results, form_defaults, form_errors):
        self._prepare_overview(cases)
        if form_results.get('editor') != -1:
            c.hide_evaluation = True

        c.sort_field = form_results.get("sort_field")
        c.sort_order = form_results.get("sort_order")

        session["id_field_candidate"] = "m.id"
        form_defaults["all_ids"] = c.all_ids
        overview = render('/casemanagement/overview.mako')
        return render_htmlfill(overview, defaults=form_defaults,
                               errors=form_errors)

    def _search(self, search_options=None):
        validator = g.case_search.get_form_validator()
        form_result = {}
        form_errors = {}
        form_defaults = None
        if search_options is None:
            search_options = {}
            form_result.update(search_options_from_session())
        else:
            form_result.update(search_options)

        if not search_options:
            try:
                form_result.update(validator.to_python(request.params))
                search_options.update(form_result)
                session['CASE_OVERVIEW_SEARCHOPTIONS'] = search_options
                session.save()
            except formencode.Invalid, error:
                log.exception("search: formencode error %s" % error)
                form_defaults = error.value
                form_errors = error.error_dict or {}

        if form_defaults is None:
            form_defaults = validator.from_python(form_result)

        cases = self._getCaseList(search_options)
        return self._renderOverview(cases, form_result, form_defaults,
                                    form_errors)

    def parseSearchOptions(self, options, extended):
        return parseSearchOptions(options, extended)


class ConvertValidator(object):

    def __init__(self, field):
        self.field = field

    def validate(self):
        return (self._validate_value(),
                self._validate_value2(),
                self._validate_type1())

    def _validate_value(self):
        return String().to_python(
            request.params.get('{0}_value'.format(self.field)))

    def _validate_value2(self):
        return None

    def _validate_type1(self):
        string_validator = String()
        return string_validator.to_python(
            request.params.get('{0}_type'.format(self.field)))


class IntConvertValidator(ConvertValidator):

    def _validate_value(self):
        integer_validator = Int()
        return integer_validator.to_python(
            request.params.get('{0}_value'.format(self.field)))


class BoolConvertValidator(IntConvertValidator):

    pass


class RadioConvertValidator(IntConvertValidator):

    pass


class DateConvertValidator(ConvertValidator):

    def _validate_value(self):
        str1 = String().to_python(
            request.params.get('{0}_value'.format(self.field)))
        return FormedDateChecker(allow_unknown=True).to_python(str1.split("-")[0])

    def _validate_value2(self):
        str1 = String().to_python(
            request.params.get('{0}_value'.format(self.field)))
        if len(str1.split("-")) == 2 and self._validate_type1() == "Between":
            return FormedDateChecker(allow_unknown=True).to_python(str1.split("-")[1])
        else:
            return None


class ChoiceConvertValidator(ConvertValidator):

    def _validate_value(self):
        values = request.params.getall("{0}_value".format(self.field))
        if values:
            integer_validator = Int()
            return list(set(integer_validator.to_python(v) for v in values))
        return None


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
