# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2022 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Roland Geider <roland.geider@intevation.de>
# Ludwig Reiter <ludwig.reiter@intevation.de>


import logging
import json

from formencode.validators import FancyValidator

from pylons.controllers.util import redirect, abort

from formed.renderer.helpers import escape
from formed.renderer.html import Item, tag

from mpulsweb.lib.security import checkRole
from mpulsweb.lib.base import BaseController, request, session, c, g, render, \
    restrict_to_POST
from mpulsweb.lib.translation import _
from mpulsweb.lib.helpers import url
from mpulsweb.model.customoverview import OverviewConfiguration, \
    ColumnConfiguration
from mpulsweb.lib.formedforms import FormedForm, Override, Button


log = logging.getLogger(__name__)


def search_columns_renderer(renderer, nc):
    html = []
    html.append(tag("div", class_="form-group col-form"))
    html.append(tag("label", for_="search_columns"))
    html.append(escape("Neue Spalte"))
    html.append("</label>")
    html.append(tag("div", class_="input-group input-group-sm"))
    html.append(tag("input", name="search_columns",
                    class_="form-control form-control-sm",
                    type="text", id="search_columns"))
    html.append(tag("div", class_="input-group-append"))
    html.append(tag("span", class_="input-group-text bg-transparent border-0"))
    html.append(tag("svg", class_="svg-icon", id="add-new-column"))
    html.append(tag("title"))
    html.append(escape(u"Spalte hinzufügen"))
    html.append("</title>")
    html.append('<use xlink:href="#plus-circle-solid"/>')
    html.append("</svg>")
    html.append("</span>")
    html.append("</div>")
    html.append("</div>")
    html.append("</div>")
    return Item("".join(html))

def columns_renderer(renderer, nc):
    html = []
    value = renderer.data.get_value(nc)

    html.append(tag("div", class_="form-group col-form"))
    html.append(tag("label", for_="search_columns"))
    html.append(escape("Vorhandene Spalten"))
    html.append("</label>")
    html.append(tag("ul", id="custom-overview-list",
                    class_="list-group"))
    for column in value:
        label = column.get_description()

        html.append(tag("li", class_="list-group-item d-flex justify-content-between"))
        html.append(escape(label))
        html.append(tag("input", type="hidden", name="columns",
                        value=column.get_specifier()))
        html.append(tag("svg", class_="svg-icon"))
        html.append(tag("title"))
        html.append(escape(u"Spalte entfernen"))
        html.append("</title>")
        html.append('<use xlink:href="#delete"/>')
        html.append("</svg>")
        html.append("</li>")
    html.append("</ul>")
    html.append("</div>")

    return Item("".join(html))


class ColumnsValidator(FancyValidator):

    def to_python(self, *args):
        validated = super(ColumnsValidator, self).to_python(*args)
        if not isinstance(validated, list):
            validated = [validated]
        candidates = g.case.get_overview_field_candidates()
        return [ColumnConfiguration.from_specifier(specifier)
                for specifier in validated if specifier in candidates]


class OverviewForm(FormedForm):

    def __init__(self, action):
        super(OverviewForm, self).__init__(
            "edit_overview_configuration.xml", "customoverview",
            action, modification_checks=True,
            overrides=[
                Override("search_columns", renderer=search_columns_renderer),
                Override("columns", renderer=columns_renderer,
                         validator=ColumnsValidator()),
                ],
            buttons=[
                Button(u"Speichern und Zurück", "save_and_back", primary=True),
                ])

    def create_instance(self):
        instance = super(OverviewForm, self).create_instance()
        instance.set_value("columns", [])
        return instance


overview_new_form = OverviewForm("new_action")
overview_edit_form = OverviewForm("edit_action")


class CustomoverviewController(BaseController):

    @checkRole(('cm', 'admin'))
    def overview(self):
        c.overviews = OverviewConfiguration.list_custom_overviews()
        return render('/customoverview/overview.mako')

    def delete(self, id, confirmed=False):
        id = self._checkInt(id)
        confirmed = self._checkBool(confirmed)

        if confirmed == 1:
            overview = OverviewConfiguration.get_custom_overview_by_id(id)
            if overview is not None:
                overview.delete()
            c.success_text = (_(u"The overview configuration was successfully deleted."))
            return self.overview()
        else:
            c.confirm_for = _(u"Delete overview configuration?")
            c.question = _(u"Do you want to delete this overview"
                           " configuration?\n"
                           " It will be irrevocably lost.")
            c.url_yes = url(controller='customoverview',
                            action='delete',
                            id=id,
                            confirmed=1)
            c.url_no = url(controller='customoverview', action='overview')
            return render('/customoverview/dialogs/confirm_delete.mako')

    def edit(self, id):
        id = self._checkInt(id)
        overview = OverviewConfiguration.get_custom_overview_by_id(id)
        if overview is None:
            redirect(url(controller='customoverview', action='overview'))
        else:
            instance = overview_edit_form.create_instance()
            instance.set_value("id", id)
            instance.set_value("name", overview.name)
            instance.set_value("columns", overview.columns)
            return self._render_edit_form(overview_edit_form, instance, {})

    def _render_edit_form(self, form, instance, errors, backlink=None):
        c.rendered_form = form.render(instance, errors)
        c.form_action = form.action
        column_candidates = g.case.get_overview_field_candidates().values()
        c.formed_fields = json.dumps(column_candidates)

        # Note: backlink may be an unvalidated request parameter. It
        # should therefore be handled very carefully here.
        if backlink == "case_overview":
            c.backlink_url = url(controller="case_overview", action="overview")
            c.backlink_title = _("Back to: Case document overview")
        else:
            c.backlink_url = url(controller="customoverview", action="overview")
            c.backlink_title = _("Back to: overview configuration")

        return render('/customoverview/edit.mako')

    @restrict_to_POST
    def edit_action(self):
        instance, errors = overview_edit_form.validate_request(request.params)

        if errors:
            return self._render_edit_form(overview_edit_form, instance, errors)

        overview_id = instance.get("id")
        overview = OverviewConfiguration.get_custom_overview_by_id(overview_id)
        if overview is None:
            abort(404)

        overview.name = instance.get("name")
        overview.columns = instance.get("columns")
        overview.save()
        c.success_text = (_(u"The overview configuration was successfully"
                            u" saved."))
        return self.overview()

    def new(self):
        instance = overview_new_form.create_instance()
        backlink = request.params.get("backlink")
        instance.set_value("backlink", backlink)
        return self._render_edit_form(overview_new_form, instance, {},
                                      backlink=backlink)

    def new_action(self):
        instance, errors = overview_new_form.validate_request(request.params)

        if errors:
            return self._render_edit_form(overview_new_form, instance, errors)

        overview = OverviewConfiguration(None,
                                         instance.get("name"),
                                         instance.get("columns"))
        overview.save()
        c.success_text = (_(u"The overview configuration was successfully"
                            u" saved."))

        # if the dialog was started from the case overview page,
        # redirect there.
        if instance.get("backlink") == "case_overview":
            session["selected_overview"] = overview.id
            redirect(url(controller='case_overview', action='overview'))

        return self.overview()
