# -*- coding: utf-8 -*-
import logging
import formencode

from pylons import request, tmpl_context as c, app_globals as g

from mpulsweb.lib.base import BaseController, render, render_htmlfill, \
    restrict_to_POST
from mpulsweb.lib.helpers import url
from mpulsweb.lib.security import checkRole
from mpulsweb.lib.translation import _, N_
from mpulsweb.lib.validators import EditSettingsForm, MaxMinInt
from mpulsweb.lib.formedforms import FormedForm, Override
from mpulsweb.model.agencysettings import Agency


log = logging.getLogger(__name__)

i18n_descriptions = {
    "group-settings": N_("Specification for agency"),
    "ka_name": N_("Name of the institution"),
    "ka_adresse": N_("Address of the institution"),

    "group-cases": N_("Specification for cases"),
    "max_speicherdauer": N_("Automatic review after days"),
    "anon_transfer": N_("Assign case document"),
    }


def render_reminder_info(renderer, nc):
    agency = Agency()
    return renderer.render_info_text(
        _("Automatic reminders are used to identify cases, which has not been"
          " edited for longer than the specified value."
          " Additionally there can be further conditions when a reminder is"
          " created. Further condition: %s")
          % (agency.getReminderConditionDesc(),))


class SettingsForm(FormedForm):

    def __init__(self):
        super(SettingsForm, self).__init__(
            "administration_settings.xml", "settings", "edit_action",
            overrides=[Override("info-max_speicherdauer",
                                renderer=render_reminder_info)])

    def load_formed_tree(self, formed_basename):
        tree = super(SettingsForm, self).load_formed_tree(formed_basename)
        for name, nc in tree.meta_tree.name2nc.items():
            untranslated = i18n_descriptions.get(name)
            if untranslated is not None:
                nc.setDescription(_(untranslated))


        max_speicherdauer = tree.meta_tree.name2nc["max_speicherdauer"]
        max_speicherdauer.setAttribute("minvalue",
                                       g.mpuls_config.get("case",
                                                          "remindermindays"))
        max_speicherdauer.setAttribute("maxvalue",
                                       g.mpuls_config.get("case",
                                                          "remindermaxdays"))

        info_transfer = tree.meta_tree.name2nc["info-anon_transfer"]
        info_transfer.setValue(_("If this option is activated, anonymized"
                                 " cases are transmitted for evaluation purpose"
                                 " to a central agency."
                                 " Anonymized cases do to not contain any"
                                 " personal data such as name or addresses."))
        return tree

    def to_form_result(self, instance):
        form_result = instance.to_form_result()
        form_result["anon_transfer"] = ("on" if instance.get("anon_transfer")
                                        else "off")
        return form_result


settings_form = SettingsForm()


class SettingsController(BaseController):

    @checkRole('admin')
    def index(self):
        return self.show()

    def _get_agency(self):
        return Agency()

    def show(self):
        c.form_errors = {}
        c.form_result = {}
        c.agency = self._get_agency()
        return render('/settings/show.mako')

    @checkRole('admin')
    def edit(self):
        agency = self._get_agency()
        instance = settings_form.create_instance()
        instance.set_value('ka_name', agency.getName())
        instance.set_value('ka_adresse', agency.getAddress())
        instance.set_value('max_speicherdauer', agency.getMaxSavetime())
        instance.set_value('anon_transfer', agency.isAnonTransfer())
        return self._render_settings_form(instance, {})

    def _render_settings_form(self, instance, errors):
        c.rendered_form = settings_form.render(instance, errors)
        return render('/settings/edit.mako')

    @checkRole('admin')
    @restrict_to_POST
    def edit_action(self):
        instance, errors = settings_form.validate_request(request.params)
        if errors:
            return self._render_settings_form(instance, errors)

        form_result = settings_form.to_form_result(instance)
        agency = self._get_agency()
        self.pre_save_hook(form_result)
        agency.store(form_result)

        c.success_text = (_('Configurations saved!'))
        return self.show()

    def pre_save_hook(self, form_data):
        """
        Customizable function, is called before saving the agency settings
        """
        pass


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
