# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import logging
from contextlib import contextmanager
import decimal

import psycopg2 as dbapi
import psycopg2.extensions

from paste.registry import StackedObjectProxy

from formed.instance.backends.database import DB as FormEdDBInterface

log = logging.getLogger(__name__)

UNKNOWN = -1


# Make some psycopg exceptions explicitly available, so that users of
# the db module can use them without having to also import psycopg2.
ProgrammingError = psycopg2.ProgrammingError


def get_version(version_string):
    version = version_string.split(" ")[0]
    version_tuple = version.split(".")
    major = int(version_tuple[0])
    minor = int(version_tuple[1])
    if len(version_tuple) > 2:
        try:
            patch = int(version_tuple[2])
        except ValueError:
            patch = UNKNOWN
    else:
        patch = 0
    return major, minor, patch

# Make psycopg2 return unicode objects for strings in query results.
# This is probably not the best way to deal with unicode and the DB-API.
# In particular, it's obviously not directly portable to other DB-API
# implementations.
psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY)

# Work around a subtle defect in psycopg2. When running two Python
# sub-interpreters in the same process the dictionary implementing the
# adaptation protocol in psycopg is shared between the interpreters.
# However, each interpreter has its own Decimal class which leads to
# adaptation failures when trying to use Decimal values in SQL queries.
# psycopg raises ProgrammingError with a message containing "can't
# adapt". We work around this problem by always registering the Decimal
# type of the current interpreter as described in the psycopg ticket for
# this defect. See also mpuls/issue3023.
# This defect has been fixed with psycopg 2.4.3
# psycopg ticket:
# http://psycopg.lighthouseapp.com/projects/62710/tickets/52-cant-adapt-type-decimal
major, minor, patch = get_version(psycopg2.__version__)
if major < 2 or (major == 2 and minor < 4) \
        or (major == 2 and minor == 4 and patch < 3):
    psycopg2.extensions.register_adapter(
        decimal.Decimal, psycopg2._psycopg.Decimal)

db = StackedObjectProxy(name="mpuls.db")


def enter(dbObject):
    db._push_object(dbObject)


def leave(dbObject=None):
    db._pop_object(dbObject)


def _quote_connect_parameter(s):
    """Return a quoted version of the string s for with psycopg2.connect."""
    return "'" + s.replace("\\", "\\\\").replace("'", "\\'") + "'"


if major < 2 or (major == 2 and minor < 4) \
        or (major == 2 and minor == 4 and patch < 1):
    # quote only for psycopg2 version < 2.4.1
    quote_connect_parameter = _quote_connect_parameter
else:
    quote_connect_parameter = lambda s: s


class DB(object):

    def __init__(self, database, user, password, host='localhost', port=5432):
        self.database = database
        self.user = user
        self.password = password
        self.host = host
        self.port = int(port)
        self.conn = None
        self.connections = []

    def getConnection(self):
        for used in self.connections:
            if not used[0]:
                used[0] = True
                return used[1]

        # quote all string values for the connect function to make sure
        # spaces, single quotes and backslashes don't lead to problems.
        # This is particularly important for passwords.  The port is an
        # int and so does not need to be quoted.  It seems to be a bug
        # in psycopg2 that it doesn't this kind of quoting automatically
        # when passing the connection parameters as keyword arguments.
        con = dbapi.connect(database=quote_connect_parameter(self.database),
                            host=quote_connect_parameter(self.host),
                            port=self.port,
                            user=quote_connect_parameter(self.user),
                            password=quote_connect_parameter(self.password))
        used = [True, con]
        self.connections.append(used)
        return con

    def recycleConnection(self, con, cur=None):
        if cur:
            try:
                cur.close()
            except:
                log.exception("Error ignored while closing cursor")
        if con:
            for used in self.connections:
                if used[1] == con:
                    used[0] = False
                    return
            log.warning("WARNING: tried to recycle unused connection")

    def closeConnections(self):
        cons = self.connections
        self.connections = []
        for used in cons:
            if used[0]:
                log.warning("WARNING: Closing an unrecycled connection")
            try:
                used[1].close()
            except:
                log.exception("Error ignored while closing connection")

    @contextmanager
    def transaction(self):
        """Return a context manager for performing a block within a transaction.
        Example usage:
            with db.transaction() as cur:
                cur.execute("UPDATE foo ...")
                ...

        The object bound to cur in the example is a database cursor that
        can be used to execute database queries in the usual way. The
        context manager will commit the changes once control flow leaves
        the block if no exception is raised. If an exception is raised,
        the transaction will be rolled back automatically.
        """
        has_committed = False
        cur = None
        con = self.getConnection()
        try:
            cur = con.cursor()
            yield cur
            con.commit()
            has_committed = True
        finally:
            if not has_committed:
                con.rollback()
            self.recycleConnection(con, cur)

    @contextmanager
    def cursor(self, *args, **kw):
        """Return a context manager for performing queries on a cursor
        Example usage:
            with db.cursor() as cur:
                cur.execute("SELECT bar FROM ...")
                ...

        The context manager will automatically recycle the received connection.

        Any parameters passed to this method are passed on to the
        connection's cursor method.
        """
        cur = None
        con = self.getConnection()
        try:
            cur = con.cursor(*args, **kw)
            yield cur
        finally:
            self.recycleConnection(con, cur)

    def execute(self, statement, *args):
        """Execute an SQL statement in a transaction.

        This method is intended for the common case that a single SQL
        statement has to be executed in a transaction.  This method gets
        a connection with getConnection(), creates a cursor object and
        calls its execute method with the SQL statement and any further
        arguments as arguments.  If the statement is executed
        successfully, the transaction is committed.  If an error occurs
        the transaction is rolled back.  This method does not catch any
        exceptions.

        Any results retrieved from the database are discarded.
        """
        with self.transaction() as cur:
            cur.execute(statement, *args)

    def query(self, statement, *args):
        """Execute a single query and return the first result tuple.
        Since this method simply returns the first result, the query
        should produce exactly one result tuple.
        """
        with self.cursor() as cur:
            cur.execute(statement, *args)
            return cur.fetchone()


class PostgresDBInterface(FormEdDBInterface):

    def __init__(self):
        FormEdDBInterface.__init__(self)
        self.db = db._current_obj()
        self.connection = None

    def acquire_cursor(self):
        if self.connection is None:
            self.connection = self.db.getConnection()
        return self.connection.cursor()

    def release_cursor(self, cur):
        if self.transaction_level == 0:
            self.db.recycleConnection(self.connection, cur)
            self.connection = None
        else:
            cur.close()

    def do_rollback(self):
        # if self.connection is still None, no cursor was acquired and
        # there's nothing to do, actually.
        if self.connection is not None:
            self.connection.rollback()

    def do_commit(self):
        # if self.connection is still None, no cursor was acquired and
        # there's nothing to do, actually.
        if self.connection is not None:
            self.connection.commit()


# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
