# -*- coding: utf-8 -*-
#
# Copyright 2014, 2020 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls base
#
# mpuls base is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls base is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls base. If not, see <http://www.gnu.org/licenses/>.

"""
Adapt libformed to render more forms in mpulsweb.

Besides the main case form, mpulsweb has many other forms all of which
should look and behave in the same way. This module allows us to use
libformed for form rendering and validation for those other forms as
well.

The main interface for this functionality is the FormedForm class.
"""

from collections import defaultdict
from functools import partial
import logging
from cgi import escape

import formencode

from pylons import app_globals as g
from pylons.i18n.translation import get_lang

from formed.meta.rulegen import RuleGenerator
from formed.meta.semantic import SemanticError, checkAndConvert, ErrorItem
from formed.meta.conditionals import FormInfo
import formed.meta.data as FD
from formed.renderer.helpers import tag, NA
from formed.renderer.html import apply_hidden_booleans, apply_hidden_multiselect, \
     Item, ErrorRenderer
from formed.instance.simplified import StructureInstanceTree, \
     StructureFormData, StructureVariableProvider
from formed.instance.backends.common import evaluate, rule_available

from mpulsweb.lib.formedrenderer import BaseViewRenderer, \
     UrlFactory as BaseUrlFactory
from mpulsweb.lib.helpers import url, literal
from mpulsweb.lib.translation import _


log = logging.getLogger(__name__)


def formed_form_data(params):
    """Prepare raw HTML form data for use with libformed.

    This function takes a dictionary with the raw form data and returns
    a new dictionary where all items that are not part of the actual
    form are removed.

    More specifically, the forms generated by libformed contain various
    buttons, e.g. for adding/removing repeat group instances and saving
    the form, as well as other inputs like the hidden input for the page
    name that are necessary to process user input but are not part of
    the actual form and which therefore cannot be handled by e.g. the
    intance tree's setData method. This function removes those non-form
    items.
    """
    return {key: value
            for key, value in params.items()
            if not key.startswith(('save', 'savenext', 'page', 'add_rg',
                                   'del_rg', 'mpuls-toggle-all-'))}


def convert_error_items2(error_items):
    errors = {}
    defaults = {}
    for key, items in error_items.items():
        for e in items:
            errors[e.name] = e.msg
            defaults[e.name] = e.bad
    return errors, defaults


def update_errors(err1, err2):
    errors = err1.copy()
    for key, value in err2.items():
        if isinstance(value, basestring):
            value = [value]
        errors[key] = value
    return errors


class UrlFactory(BaseUrlFactory):

    def __init__(self, controller, action):
        self.overrides = {("formularpage", "save"): (controller, action)}

    def url(self, controller, action, id=None):
        pair = (controller, action)
        controller, action = self.overrides.get(pair, pair)
        url_args = {"controller": controller, "action": action}
        if id is not None:
            url_args["id"] = id
        return url(**url_args)


class FormFormData(StructureFormData):

    def evaluate_rules(self, meta, warnings_only=True):
        return super(FormFormData, self).evaluate_rules(meta, warnings_only)


class SimplifiedRenderer(BaseViewRenderer):


    """Variant of the BaseViewRenderer for FormedForm.

    Besides influencing the rendering by e.g. not rendering the submit
    buttons at the top of the form, its main feature is that it supports
    overrides, which the FormedForm class will automatically pass to
    this class when rendering a form. It will also be instantiated with
    the buttons defined in the FormedForm and render those instead of
    the default submit buttons below the form.

    This class uses the following aspects of the overrides:

      - render_nc method will call the override's render_nc method if an
        override exists for the field.

      - render_choice_children will call the override's
        render_choice_children method if an override exists for the
        field.
    """

    def __init__(self, *args, **kw):
        overrides = kw.pop("overrides", None)
        self.overrides = overrides if overrides is not None else {}
        self.buttons = kw.pop("buttons", ())
        super(SimplifiedRenderer, self).__init__(*args, **kw)

    def server_side_only(self, nc):
        return True

    def render_submit_buttons_top(self, show_next=True):
        return []

    def render_submit_buttons_bottom(self, show_next=True):
        if self.buttons:
            return [button.render(self) for button in self.buttons]
        else:
            return super(SimplifiedRenderer,
                         self).render_submit_buttons_bottom(show_next=show_next)

    def render_nc(self, nc):
        override = self.overrides.get(nc.getName())
        if override is None:
            render = super(SimplifiedRenderer, self).render_nc
        else:
            render = partial(override.render_nc, self)
        return render(nc)

    def render_choice_children(self, nc, value):
        override = self.overrides.get(nc.getName())
        if override is None:
            render = super(SimplifiedRenderer, self).render_choice_children
        else:
            render = partial(override.render_choice_children, self)
        return render(nc, value)


class FormInstanceTree(StructureInstanceTree):

    """Instance tree for use with FormedForm

    This class extends the base class with support for field-specific
    overrides. The FormedForm instance will automatically instantiate
    this class so that it uses the overrides defined in the FormedForm.

    This class uses the following aspects of the overrides:

      - The to_form_result attribute is used to convert values in
        to_form_result and to_hidden_inputs.

        Note: It's not good design, and in fact was not deliberately
        designed at all, that the to_form_result attribute is used for
        both of these methods because their purposes differ too much.
        Luckily there's no field which is used by both to_form_result
        and to_hidden_inputs so there shouldn't be a problem in
        practice, for now.
    """

    def __init__(self, *args, **kw):
        self.overrides = kw.pop("overrides", {})
        return super(FormInstanceTree, self).__init__(*args, **kw)

    def to_form_result(self):
        """Return the contents of the 'form' page as a dictionary.

        This is intended to be sufficiently close to what the formencode
        validator would have returned, so that the controller code can
        use it in the same way when using the model code to actually
        store the data.
        """
        return self._dict_for_page("form")

    def to_hidden_inputs(self):
        """Return the contents of the 'hidden-inputs' page as a dictionary.

        This return value should be usable as value of the hidden_fields
        parameter of the renderer's render method.
        """
        return self._dict_for_page("hidden-inputs")

    def _dict_for_page(self, page_name):
        def convert(name, value):
            meta = self.meta_structure.meta_tree.findByName(name)
            override = self.overrides.get(name)
            if override is not None and override.to_form_result:
                return override.to_form_result(meta, value)

            if isinstance(meta, FD.BoolLeaf):
                return bool(value)
            else:
                return value

        name2page = self.meta_structure.meta_tree.name2page
        return {key: value
                for key, value in self.root.to_dict(convert).items()
                if name2page[key].getName() == page_name}


class FormedValidator(object):

    def __init__(self, instance, page_name, overrides=None):
        self.instance = instance
        self.meta_tree = instance.meta_structure.meta_tree
        self.page_name = page_name
        self.overrides = overrides if overrides is not None else {}

    def _to_formed(self, form_data, page_name):
        """Transform the form_data into a form suitable for formed."""
        form_data = dict(form_data.items())
        apply_hidden_booleans(form_data)
        apply_hidden_multiselect(form_data)

        parsed_params = {}
        for key, value in form_data.iteritems():
            parsed_params[StructureFormData.parse_form_name(key)] = value

        return parsed_params

    def validate(self, form_data):
        params = self._to_formed(form_data, self.page_name)
        values, errors = self._check_semantic_errors(params)
        values, errors = self._check_rules(values, errors)

        return values, errors

    def _check_semantic_errors(self, structure_data):
        values = dict()
        errors = defaultdict(list)

        for key, value in structure_data.items():
            widget = self.meta_tree.name2nc[key[-1]]
            is_file_upload = hasattr(value, "filename")
            if (isinstance(value, basestring) and not value.strip()
                and not isinstance(widget, (FD.TextLeaf, FD.TextAreaLeaf))):
                values[key] = None
            else:
                try:
                    values[key] = self._check_and_convert(widget, value,
                                                          structure_data)
                except SemanticError, err:
                    log.debug('Semantic error (key %s): %s' % (key, err.value))
                    if is_file_upload:
                        value = None
                    errors[key].append(ErrorItem(ErrorItem.SEMANTIC, key,
                                                 err.value, value))
        return values, errors

    def _check_and_convert(self, widget, value, structure_data):
        # TODO: structure_data doesn't seem to be needed
        override = self.overrides.get(widget.getName())
        if override is None:
            return checkAndConvert(widget, value)
        else:
            return override.check_and_convert(widget, value)

    def _check_rules(self, checked_values, all_errors):
        changes, eval_errors = evaluate(checked_values, self.meta_tree,
                                        self._rule_available,
                                        self._create_eval_context,
                                        lambda k: k[-1],
                                        find_best_changes=False)
        for key, errors in eval_errors.iteritems():
            all_errors[key].extend(errors)
        return changes, all_errors

    def _rule_available(self, rule, eval_context):
        return rule_available(self.meta_tree, rule, eval_context)

    def _create_eval_context(self, key, shadow):
        return StructureVariableProvider(self.instance, key, shadow)


class FormedFormRuleGenerator(RuleGenerator):

    def __init__(self, document):
        super(FormedFormRuleGenerator, self).__init__(document)
        self.generate_unknown_rules()

    def expression_for_required_rule(self, widget):
        x = "$" + widget.getName()
        if isinstance(widget, (FD.TextLeaf, FD.TextAreaLeaf)):
            return "%s isset %s '' != and" % (x, x)
        return super(FormedFormRuleGenerator,
                     self).expression_for_required_rule(widget)

    def required_rule_mark(self, widget, parsed_flags):
        return "error:" + widget.getName()

    def generate_unknown_rules(self):
        for widget in self.document.findAllByClass((FD.DateLeaf, FD.IntLeaf,
                                                    FD.NumericLeaf)):
            rule = FD.RuleLeaf()
            rule.setName("unknown-rule-%s" % widget.getName())
            rule.setExpr("%(ref)s isset not %(ref)s known or"
                         % {"ref": "$" + widget.getName()})
            rule.setValue("Wert darf nicht Unbekannt sein")
            rule.setDescription("Wert darf nicht Unbekannt sein")
            rule.setMark("error:%s" % widget.getName())
            widget.getParent().addChild(rule)


class RenderedForm(object):

    def __init__(self, rendered_form, rendered_errors):
        self.rendered_form = rendered_form
        self.rendered_errors = rendered_errors


class Override(object):

    """Override behavior of a specific field in a FormedForm.
    """

    def __init__(self, name, validator=None, checker=None,
                 render_choices=None, renderer=None, to_form_result=None):
        """Initialize the Override object.

        Parameters:
            name -- The name of the form field

            validator -- Optional formencode validator to use for this
                field instead of libformed's default validation. If
                given, the object must have a to_python method which
                raises formencode.Invalid exceptions when the input is
                invalid.

            checker -- Optional libformed compatible validator. If
                given, it will be called with the node-component for the
                field and the value and should raise a SemanticError
                just like the checkers in formed/meta/semantic.py do.

            render_choices -- Optional renderer for the choices of a
                choice field. If given it will be called with the HTML
                renderer, the field's node-component and current value
                of the field. It should return the rendered HTML
                fragment as a string.

            renderer -- Optional renderer for the field. If given it
                will be called with the HTML renderer and the field's
                node-component. It should return the rendered HTML
                fragment as a string.

            to_form_result -- Optional converter for the field for
                FormInstanceTree's to_form_result and to_hidden_inputs
                methods.
        """
        self.name = name
        self.formencode_validator = validator
        self.checker = checker
        self.render_choices = render_choices
        self.renderer = renderer
        self.to_form_result = to_form_result

    def check_and_convert(self, nc, value):
        if self.formencode_validator is not None:
            try:
                return self.formencode_validator.to_python(value)
            except formencode.Invalid, e:
                raise SemanticError(unicode(e))
        if self.checker is not None:
            return self.checker(nc, value)
        return checkAndConvert(nc, value)

    def render_choice_children(self, renderer, nc, value):
        render = self.render_choices
        if render is None:
            render = super(renderer.__class__, renderer).render_choice_children
        else:
            render = partial(render, renderer)
        return render(nc, value)

    def render_nc(self, renderer, nc):
        render = self.renderer
        if render is None:
            render = super(renderer.__class__, renderer).render_nc
        else:
            render = partial(render, renderer)
        return render(nc)


class Button(object):

    def __init__(self, label, name, primary=False):
        self.label = label
        self.name = name
        self.primary = primary

    def render(self, renderer):
        return renderer.render_submit_button(self.label, self.name,
                                             "btn-primary"
                                             if self.primary else
                                             "btn-outline-primary")


class FormedForm(object):

    """
    A form rendered and validated with libformed.

    As with the main case form, the form is described by an XML file.
    This file has the same structure as the main formedtree.xml file.
    The FormedForm class and related code makes some assumptions about
    this structure:

     - The actual form is described by a page named "form".

     - Optionally, there may another page named "hidden-inputs"

       The fields defined on the "hidden-inputs" page are rendered into
       the HTML form as hidden input fields and can be used for non-user
       editable fields like e.g. IDs of objects the users can edit with
       the form or fields that influence the rendering of the form with
       conditionals.

    There may be other pages as well. For instance, some forms have a
    hidden page called "parameters" for fields that influence the form
    (e.g. via conditionals) but which are not actually form fields and
    do not need to be known on the client side.

    The content of the form is represented by an instance of
    FormInstanceTree.

    The forms we need to support here sometimes have requirements that
    are not easily met with libformed itself. For instance, some choice
    fields have options that need to be determined dynamically at
    runtime, or fields that require validation that libformed cannot
    perform like the HTML validation required for the logbook comments.
    This behavior can be implemented with overrides which allow
    field-specific behavior in the renderer, the validator and other
    places.
    """

    def __init__(self, formed_basename, controller, action,
                 modification_checks=False, overrides=None, buttons=()):
        """Initialize the FormedForm.

        Parameters:

          formed_basename (string) -- The basename, e.g. "myform.xml" of
              the formedtree XML file. The file will be loaded from the
              mpulsweb/public/forms/ directory.

          controller, action -- The name of the controller and
              corresponding action for the URL used in the HTML form
              element.

          modification_checks (bool) -- Whether to show a warning dialog
              to the user when clicking on a link that would leave the
              current web-page when are are unsaved changes in the form.
              This is standard behavior for the main case form but is
              usually not needed for the forms rendered with this class.
              The default is therefore False.

          override (list of Override) -- Optional. Each Override
              instance defines behaviour of a field that cannot be
              expressed in the XML file. See Override for details.

          buttons (list of Button) -- Optional. If the standard submit
              button is not suitable for the form, this list defines the
              buttons that should be rendered instead.
        """
        self.meta_structure = self.load_formed_tree(formed_basename)
        self.page_name = "form"
        self.controller = controller
        self.action = action
        self.modification_checks = modification_checks
        self.overrides = self.overrides_as_dict(overrides)
        self.buttons = buttons
        self.form_info = FormInfo(self.meta_structure.meta_tree)
        self.form_info_js = self.form_info.as_js()

    def load_formed_tree(self, formed_basename):
        return g.load_formed_form(formed_basename,
                                  rule_generator=FormedFormRuleGenerator)

    def overrides_as_dict(self, overrides):
        return {o.name: o
                for o in (overrides if overrides is not None else ())}

    def create_instance(self):
        """Return a new instance tree with default values"""
        return FormInstanceTree(self.meta_structure, overrides=self.overrides)

    def get_renderer(self, instance, error_items):
        errors, shadow = convert_error_items2(error_items)
        meta_tree = self.meta_structure.meta_tree
        return SimplifiedRenderer(FormFormData(instance, shadow, errors),
                                  meta_tree.findByName(self.page_name),
                                  UrlFactory(self.controller, self.action),
                                  read_only=False, form_info=self.form_info,
                                  overrides=self.overrides,
                                  buttons=self.buttons,
                                  modification_checks=self.modification_checks,
                                  languages=get_lang())

    def get_error_renderer(self, instance, form_renderer, errors):
        errors, shadow = convert_error_items2(errors)
        errors = update_errors(form_renderer.errors, errors)
        return ErrorRenderer(FormFormData(instance, shadow, errors),
                             self.meta_structure.meta_tree, errors,
                             form_renderer.warnings, form_renderer.notices,
                             form_renderer.widget_order,
                             UrlFactory(self.controller, self.action),
                             disabled_fields=form_renderer.disabled_fields,
                             show_error_undo=False,
                             languages=get_lang())


    def render(self, instance, errors, hidden_fields=()):
        """Render the form for the given instance tree and existing errors.

        The errors parameter should be a dictionary usable as the errors
        parameter for libformed's ErrorRenderer.
        """
        hidden = instance.to_hidden_inputs()
        for key, value in hidden_fields:
            hidden[key] = value
        hidden["page"] = self.page_name
        renderer = self.get_renderer(instance, errors)
        rendered_form = literal(renderer.render(show_next=False,
                                                hidden_fields=hidden.items()))
        error_renderer = self.get_error_renderer(instance, renderer, errors)
        rendered_errors = literal(error_renderer.render())
        return RenderedForm(rendered_form, rendered_errors)

    def validate_request(self, request_params):
        """Validate request parameters and return an instance tree and errors.

        This method always returns an instance tree, which contains the
        values that have been determined as valid and defaults
        otherwise. The errors value is a dictionary that can be e.g.
        passed to the render method as the errors value.
        """
        instance = self.create_instance()
        validator = FormedValidator(instance, self.page_name,
                                    overrides=self.overrides)
        decoded = formencode.variabledecode.variable_decode(request_params)
        values, errors = validator.validate(formed_form_data(decoded))
        for key, value in values:
            try:
                instance.set_value(key, value)
            except ValueError, e:
                log.warn("Value error %s when setting value %r for %r"
                         % (e, value, key))

        def add_error(key, value, msg):
            errors[key].append(ErrorItem(ErrorItem.SEMANTIC, (key,), msg, value))

        self.validate_instance(instance, add_error)

        return instance, errors

    def validate_instance(self, instance, add_error):
        """Perform instance level validation.

        This method is called by validate_request after it has validated
        the request data and create the instance. The default
        implementation of this method does nothing. Derived classes may
        overwrite this method in order to perform additional validation
        that cannot be done on a per-field bases (hence "instance level
        validation").

        The add_error parameter is a function that can be called with
        the name, value and error message to add more errors to the
        errors dictionary returned by validate_request.
        """


class UserSetOverride(Override):

    """Override for user lists"""

    def __init__(self, name, checker=None, renderer=None, **kw):
        if checker is None:
            checker = int_set_checker
        if renderer is None:
            renderer = user_set_renderer
        super(UserSetOverride, self).__init__(name, checker=checker,
                                              renderer=renderer, **kw)

def int_set_checker(nc, value):
    """Checker for a set of ints rendered as a bunch of check-boxes

    Due to the way libformed and HTML forms work, if more than one of
    those check-boxes have been checked, value is a list of strings. If
    only one of those check-boxes has been checked, value is a string.
    If no check-box has been checked, this checker is not called at all.

    The return value is a set of ints.
    """
    # We need to return a set of ints, so we convert to a list of
    # strings first, if necessary, and then convert the strings and
    # build the set.
    if not isinstance(value, list):
        value = [value]
    return set([int(user_id) for user_id in value])


def user_set_renderer(renderer, nc):
    from mpulsweb.model.user import UserListObject
    userlist = UserListObject()
    candidates = userlist.getUserList(filter=[u.id
                                              for u in userlist.getAdminList()])
    current_members = renderer.data.get_value(nc)

    fieldname = nc.getName()
    toggle_all_id = "mpuls-toggle-all-%s" % (fieldname,)

    # checkbox to (de-)select all users
    out = [tag("div", class_="form-check mb-3"),
           tag("input", type="checkbox", class_="form-check-input", value="1",
               name=toggle_all_id, id=toggle_all_id),
           tag("label", class_="form-check-label", for_=toggle_all_id),
           escape(_('Select/deselect all users')),
           "</label>",
           "</div>"]

    # checkboxes for individual users.
    for user in candidates:
        input_id = "%s-%d" % (fieldname, user.id)
        out.extend([tag("div", class_="form-check"),
                    tag("input", type="checkbox", name=fieldname,
                        class_="form-check-input", value=user.id, id=input_id,
                        checked=user.id in current_members),
                    tag("label", class_="form-check-label", for_=input_id),
                    escape("%s, %s"
                           % (NA(user.last_name), NA(user.first_name))),
                    "</label>",
                    "</div>"])
    return Item("\n".join(out))


def file_size_checker(nc, value):
    max_size=g.mpuls_config.get("case", "max-document-size")
    max_size_text=g.mpuls_config.get("case", "max-document-size-text")

    uploads = checkAndConvert(nc, value)
    for upload in uploads:
        if upload["size"] > max_size:
            raise SemanticError(u"Die Datei %s ist größer als die maximal"
                                u" zulässige Größe von %s."
                                % (upload["filename"], max_size_text))
    return uploads


class DocumentNewForm(FormedForm):

    def __init__(self, controller, action):
        super(DocumentNewForm, self).__init__(
            "new_document.xml", controller, action,
            overrides=[Override("file", checker=file_size_checker)])

    def create_instance(self):
        instance = super(DocumentNewForm, self).create_instance()
        instance.set_value("maxupload",
                           g.mpuls_config.get("case", "max-document-size-text"))
        return instance


class DocumentRenameForm(FormedForm):

    def __init__(self, controller, action):
        super(DocumentRenameForm, self).__init__(
            "rename_document.xml", controller, action)

    def get_api(self, instance):
        """Return the document API object.
        The parameter is an instance of the form and may be needed to
        instantiate the API object.

        The default implementation raises NotImplementedError. Derived
        classes must overwrite this method.
        """
        raise NotImplementedError

    def from_object(self, doc):
        """Create a new instance tree from a document object."""
        instance = self.create_instance()
        instance.set_value("name", doc.getName())
        instance.set_value("id", doc.id)
        return instance

    def validate_instance(self, instance, add_error):
        api = self.get_api(instance)
        document_id = instance.get("id")
        name = instance.get("name")
        for doc in api.used_names_db(name=name):
            if doc.id != document_id:
                add_error('name', name,
                          _("A file with that name already exists"))
