# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008, 2011 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Bernhard Herzog <bernhard.herzog@intevation.de>
#

from pylons import app_globals as g

from formed.renderer.html import HtmlRenderer, UrlFactory as BaseUrlFactory
from formed.renderer.html import Item
from formed.renderer.helpers import tag, escape
from formed.meta.conditionals import UnsupportedFeatureError

from mpulsweb.lib.helpers import url
from mpulsweb.lib.security import current_csrf_token


class UrlFactory(BaseUrlFactory):

    def url(self, controller, action, id=None):
        url_args = {"controller": controller, "action": action}
        if id is not None:
            url_args["id"] = id
        return url(**url_args)

    def icon(self, name):
        """Return the url for the icon given by name."""
        return url("/images/icons/" + name)


class BaseViewRenderer(HtmlRenderer):

    def __init__(self, *args, **kw):
        self.form_info = kw.pop("form_info", g.conditional_form_info)
        modification_checks = kw.pop("modification_checks", True)
        if modification_checks:
            kw["set_modification_attrs"] = {"onchange": "setModification();"}
        super(BaseViewRenderer, self).__init__(*args, **kw)

    def render_form_tag(self):
        html = super(BaseViewRenderer, self).render_form_tag()
        if g.mpuls_config.is_enabled('security', 'require-csrf-tokens'):
            html += tag("input", type="hidden", name="csrftoken",
                        value=current_csrf_token())
        return html

    def error_info(self, nc):
        css, errors = super(BaseViewRenderer, self).error_info(nc)
        if errors:
            return ("error", errors)

        messages = self.get_warnings(nc)
        if messages:
            return ("warning", messages)

        return ("", [])

    def render_page(self, nc):
        parts = super(BaseViewRenderer, self).render_page(nc)
        page_info = self.form_info.pages[nc.getName()]
        for field in page_info.foreign_fields:
            converter = self.form_info.fields[field].converter

            # use the instance item because self.data.get_id returns the
            # wrong ID work correctly when referencing a master_tbl
            # field when rendering a page in a repeat group (the RG-ID
            # is included in the item's ID). The getIdentifier method of
            # the instance item gives the correct ID, though.
            instance_item = self.data.instance.getItem(self.data.get_id(field))

            value = instance_item.getValue()
            if value is None:
                formatted = ""
            elif converter == "choice":
                if instance_item.getMeta().getMultiple():
                    raise UnsupportedFeatureError(
                        "Unsupported: choice converter for multi-select fields")
                formatted = str(value)
            elif converter in ("int", "bool", "radio"):
                formatted = str(value)
            elif converter in ("date",):
                formatted = self.format_date(instance_item.getMeta(), value)
            else:
                raise UnsupportedFeatureError("Unsupported converter: %r"
                                              % (converter,))
            parts.append(tag("input", type="hidden",
                             name=instance_item.getIdentifier(),
                             value=formatted))
        return parts

    def server_side_only(self, nc):
        """Return whether the conditial can be handled completely by the server.

        Conditionals that do not reference any fields on the current
        page do not have to be handled by the client. This method
        determines that by examining the form_info passed to the
        constructor. Only conditionals that reference fields on the same
        page are contained in the page info's conditionals dictionary.
        """
        page = self.form_info.pages[self.view_root.getName()]
        return page.conditionals.get(nc.getName()) is None

    def render_save(self, nc):
        """Overwrite the inherited method so that save-elements are now shown
        """
        return Item("")


class ViewRenderer(BaseViewRenderer):

    def render_submit_buttons_top(self, show_next=True):
        """Render invisible submit button as default action for the enter Key.

        The normal submit buttons for the main form are outside the form
        element. However, there are some submit buttons in the form
        element for adding and deleting repeat group instances. Top make
        sure pressing the enter key does not invoke one of those
        buttons, we add an invisible default button at the top of the
        form.
        """
        return [tag("button", name="save", style="display: none;"),
                "</button>"]

    def render_submit_buttons_bottom(self, show_next=True):
        return []



class ViewPrintRenderer(ViewRenderer):

    def render_group(self, nc):
        description = nc.getDescription()

        out = []
        if nc.isInvisible():
            if description:
                out.append("<!-- %s -->\n" % escape(description))
        else:
            out.append(tag("div", class_="print", id=self.data.get_id(nc)))
            if description:
                out.append('<legend>%s %s</legend>'
                           % ("".join(self.render_label(nc,
                                                        self.data.get_id(nc))),
                              "".join(self.render_annotation_links(nc))))
            out.append(self.render_children(nc))
            out.append("</div>")
        return Item("".join(out))

    def render_inline_repeat_group(self, nc):
        description = nc.getDescription()
        out = []
        out.append('<div>')
        out.append('<legend>%s %s</legend>'
                   % (escape(description),
                      "".join(self.render_annotation_links(nc))))
        out.append('<div class="repeatgroup">')
        out.extend(self.render_repeat_list(nc))
        out.append('</div>')
        out.append('</div>')
        return Item("".join(out))
