# -*- coding: utf-8 -*-
#
# Copyright 2022 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of MPULS (CoMPUter-based case fiLeS).
#
# MPULS is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# MPULS is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with MPULS. If not, see <http://www.gnu.org/licenses/>.

import logging
import json
import types

from pylons import app_globals as g

from mpulsweb.lib.db import db
from mpulsweb.lib.customoverview import Overview, Column
import mpulsweb.lib.helper.filters as F
import mpulsweb.lib.helpers as h
from mpulsweb.lib.translation import _


log = logging.getLogger(__name__)


LOAD_CUSTOM_OVERVIEWS = """
SELECT id, name, overview
  FROM custom_overview_tbl_view
"""


LOAD_CUSTOM_OVERVIEW = """
SELECT id, name, overview
  FROM custom_overview_tbl_view
 WHERE id = %(id)s
 ORDER BY name
"""

SAVE_CUSTOM_OVERVIEW = """
UPDATE custom_overview_tbl_view
   SET name = %(name)s,
       overview = %(overview)s
 WHERE id = %(id)s
"""


DELETE_CUSTOM_OVERVIEW = """
SELECT delete_custom_overview_ds(%(id)s);
"""

CREATE_CUSTOM_OVERVIEW = """
SELECT create_custom_overview_ds(%(name)s, %(overview)s);
"""


class CustomOverviewError(Exception):

    pass


class ColumnConfiguration(object):

    @classmethod
    def from_specifier(cls, specifier):
        category, sep, name = specifier.partition(":")
        if category in ("formed", "mpuls"):
            return cls(category, name)
        else:
            raise CustomOverviewError("Column specifier has unknown category:"
                                      " %r" % (specifier,))

    def __init__(self, category, name):
        self.category = category
        self.name = name

    def get_specifier(self):
        return self.category + ":" + self.name

    def get_description(self):
        return g.case.get_overview_field_candidates()[self.get_specifier()]["description"]


class OverviewConfiguration(object):

    @classmethod
    def get_custom_overview_by_id(cls, id1):
        result = None
        with db.cursor() as cur:
            cur.execute(LOAD_CUSTOM_OVERVIEW, {'id': id1})
            row = cur.fetchone()
            if row is not None:
                result = cls(row[0], row[1], row[2])
            else:
                result = None
        return result

    @classmethod
    def list_custom_overviews(cls):
        result = []
        with db.cursor() as cur:
            cur.execute(LOAD_CUSTOM_OVERVIEWS)
            for row in cur.fetchall():
                result.append(cls(row[0], row[1], row[2]))
        return result

    def __init__(self, id, name, columns):
        self.id = id
        self.name = name
        if isinstance(columns, types.StringTypes):
            self.columns = [ColumnConfiguration.from_specifier(specifier)
                            for specifier in json.loads(columns)]
        else:
            self.columns = columns

    def save(self):
        params = {"name": self.name,
                  "overview": json.dumps([col.get_specifier()
                                          for col in self.columns])}
        if self.id is not None:
            query = SAVE_CUSTOM_OVERVIEW
            params["id"] = self.id
        else:
            query = CREATE_CUSTOM_OVERVIEW
        with db.transaction() as cur:
            cur.execute(query, params)
            if self.id is None:
                self.id = cur.fetchone()[0]

    def delete(self):
        with db.transaction() as cur:
            cur.execute(DELETE_CUSTOM_OVERVIEW, {"id": self.id})

    def as_overview(self):
        fields = []
        for column in self.columns:
            if column.category == "formed":
                fields.append(g.case.new_overview_field(column.name))
            elif column.category == "mpuls":
                if column.name == "id":
                    fields.append(Column("id", "ID", "table_width_mid_s", ["cm", "admin"],
                                         lambda case: F.NA(case.id)))
                elif column.name == "zugriff":
                    fields.append(Column("zugriff", _('Updated'), "table_name", ["cm"],
                                         lambda case: h.format_date(case.state.getAccessTime()))),
            else:
                raise RuntimeError("Unexpected column category: %r"
                                   % (column.category))
        return Overview(self.name, fields)
