# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
# Bernhard Herzog <bernhard.herzog@intevation.de>
#

from __future__ import with_statement

import os.path
import base64
from collections import defaultdict
from mimetypes import guess_type
import xml.etree.ElementTree as ET

import psycopg2 as dbapi

from formed.meta.data import RepeatNode, TextLeaf, DateLeaf
from formed.meta.structure import StructureNode
from formed.instance.backends.xmlimport import ComponentFactory
from formed.instance.backends.xmlexport import add_leaf
from formed.instance.simplified import StructureInstanceNode, \
    StructureInstanceTree

import mpulsweb.lib.helpers as h
from mpulsweb.lib.base import g
from mpulsweb.lib.db import db
from mpulsweb.lib.translation import _
from mpulsweb.lib.exceptions import MpulsException


MAX_SIZE = 10 * 1024 * 1024
BLOCK_SIZE = 4 * 1024


LOAD_DOCUMENT_TEMPLATE = """
SELECT master_id, groesse, mime,
       CASE WHEN %(load_contents)s THEN daten ELSE NULL END AS daten,
       name, datum
FROM ka_{parent}_dokumente_tbl_view
WHERE id = %(id)s
"""


class DocumentException(MpulsException):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class DocumentNotFoundException(DocumentException):

    """Raised when trying to load a non-existing document."""


class DocumentAPI(object):

    def list_db(self, name=None):
        with db.cursor() as cursor:
            return self.list(cursor, name)

    def list(self, cursor, name=None):
        """Return a list describing all documents.
        The return value is a list of Document objects.
        """
        query_args = {}
        condition = "TRUE"
        if name is not None:
            query_args["name"] = name
            condition = "name = %(name)s"
        query, extra_args = self._get_list_query()
        query_args.update(extra_args)
        cursor.execute(query.format(condition=condition), query_args)
        return [self._document_from_list_row(row) for row in cursor.fetchall()]

    @classmethod
    def _document_from_list_row(cls, row):
        id, case, size, mime, name, insert_date = row
        return Document(id, case, size, mime, name, insert_date)

    def _get_list_query(self):
        raise NotImplementedError

    def list_as_dict(self, cursor, name=None):
        return dict((doc.name, doc) for doc in self.list(cursor, name))

    def used_names_db(self, name=None):
        """The same as used_names, but with the default db connection."""
        with db.cursor() as cursor:
            return self.used_names(cursor, name)

    def used_names(self, cursor, name=None):
        """Like the list method, but intended for name collision checks.

        The default implementation simply delegates to the list method,
        but derived classes may override this method to e.g. consider a
        broader set of documents.
        """
        return self.list(cursor, name=name)

    def used_names_as_dict(self, cursor, name=None):
        """Calls used_names and turns the result into a dict with names as keys.
        """
        return dict((doc.name, doc) for doc in self.used_names(cursor, name))

    def load_db(self, document_id, load_contents_immediately):
        with db.cursor() as cursor:
            return self.load(cursor, document_id, load_contents_immediately)

    def load(self, cursor, document_id, load_contents_immediately):
        cursor.execute(self._get_load_query(),
                       {"load_contents": bool(load_contents_immediately),
                        "id": document_id})
        row = cursor.fetchone()
        if not row:
            raise DocumentNotFoundException(_(u"The requested document can't be"
                                              " found."))
        return self._document_from_load_row((document_id,) + row)

    @classmethod
    def _document_from_load_row(cls, row):
        document_id, case_id, size, mime, data, name, insert_date = row
        return Document(document_id, case_id, size, mime, name, insert_date,
                        data)

    def _get_load_query(self):
        raise NotImplementedError

    def load_data(self, cursor, document_id):
        cursor.execute(self._get_load_data_query(),
                       {"id": document_id})
        row = cursor.fetchone()
        if not row:
            raise DocumentNotFoundException(_(u"The requested document can't be"
                                              " found."))
        return row[0]

    def _get_load_data_query(self):
        raise NotImplementedError

    def create_transaction(name, data, uuid=None):
        with db.transaction() as cur:
            self.create(cursor, name, data, uuid=uuid)

    def create(self, cursor, name, data, uuid=None, mime=None,
               insert_date='TODAY'):
        name = name.strip().rsplit('\\', 1)[-1].rsplit('/', 1)[-1]
        if not name:
            raise DocumentException(_(u"No document available"))

        if mime is None:
            mime_guess = guess_type(name)
            mime = mime_guess[0] if mime_guess else "application/octet-stream"

        create_args = {"name": name}
        create_stmt, extra_create_args = self._get_create_query(uuid)
        create_args.update(extra_create_args)
        cursor.execute(create_stmt, create_args)
        row = cursor.fetchone()
        if row is None:
            raise DocumentException(_(u"Could not create new document"))
        document_id = row[0]

        update_args = {'bytes': dbapi.Binary(data), 'size': len(data),
                       'name': name, 'mime': mime, 'id': document_id,
                       'datum': insert_date}
        update_stmt, extra_update_args = self._get_update_query()
        update_args.update(extra_update_args)
        cursor.execute(update_stmt, update_args)

        return Document(document_id, create_args.get("id"), update_args["size"],
                        mime, name, insert_date, data)

    def _get_create_query(self, uuid):
        raise NotImplementedError

    def _get_update_query(self):
        raise NotImplementedError

    def delete_transaction(self, id):
        with db.transaction() as cursor:
            self.delete(cursor, id)

    def rename_transaction(self, id, name):
        with db.transaction() as cursor:
            self.rename(cursor, id, name)

    def rename(self, cursor, id, name):
        cursor.execute(self._get_rename_query(), {'id': id, 'name': name})

    def _get_rename_query(self):
        raise NotImplementedError

    def delete(self, cursor, id):
        cursor.execute(self._get_delete_query(), {'id': id})

    def _get_delete_query(self):
        raise NotImplementedError

    def upload_documents_transaction(self, uploads):
        with db.transaction() as cursor:
            self.upload_documents(cursor, uploads)

    def upload_documents(self, cursor, uploads):
        for entry in uploads:
            self.create(cursor, self.unique_name(cursor, entry['filename']),
                        entry['content'], uuid=None)

    def unique_name(self, cursor, name):
        used_names = self.used_names_as_dict(cursor)

        def possible_filename(name):
            yield name

            base, ext = os.path.splitext(name)
            template = base + "-{}" + ext
            for number in range(1, 100):
                yield template.format(number)

        for filename in possible_filename(name):
            if filename not in used_names:
                return filename

        raise ValueError(u"100 filenames with some name, sth is wrong {}"
                         .format(name))



class GlobalDocumentAPI(DocumentAPI):

    def _get_list_query(self):
        return ("""SELECT id, master_id, groesse, mime, name, datum
                     FROM ka_global_dokumente_tbl_view
                    WHERE {condition}
                """,
                {})

    def _get_load_query(self):
        return LOAD_DOCUMENT_TEMPLATE.format(parent="global")

    def _get_load_data_query(self):
        return """SELECT daten
                    FROM ka_global_dokumente_tbl_view
                   WHERE id = %(id)s"""

    def _get_create_query(self, uuid):
        return ("SELECT * FROM create_global_dokument_ds(%(name)s)",
                {})

    def _get_update_query(self):
        return ("""UPDATE ka_global_dokumente_tbl_view
                      SET daten = %(bytes)s, groesse = %(size)s,
                          name = %(name)s, mime = %(mime)s, datum = %(datum)s
                    WHERE id = %(id)s""",
                {})

    def _get_rename_query(self):
        return """UPDATE ka_global_dokumente_tbl_view
                     SET name = %(name)s
                   WHERE id = %(id)s"""

    def _get_delete_query(self):
        return "SELECT delete_global_dokument_ds(%(id)s)"


class CaseDocumentAPI(DocumentAPI):

    def __init__(self, case_id):
        self.case_id = case_id

    def _get_list_query(self):
        return ("""SELECT id, master_id, groesse, mime, name, datum
                     FROM ka_fall_dokumente_tbl_view
                    WHERE master_id = %(master_id)s
                      AND {condition}
                 ORDER BY datum DESC, name""",
                {"master_id": self.case_id})

    def _get_load_query(self):
        return LOAD_DOCUMENT_TEMPLATE.format(parent="fall")

    def _get_load_data_query(self):
        return """SELECT daten
                    FROM ka_fall_dokumente_tbl_view
                   WHERE id = %(id)s"""

    def _get_create_query(self, uuid):
        return ("SELECT *"
                " FROM create_fall_dokument_ds(%(id)s, %(uuid)s, %(name)s)",
                {"id": self.case_id, "uuid": uuid})

    def _get_update_query(self):
        return ("""UPDATE ka_fall_dokumente_tbl_view
                      SET daten = %(bytes)s, groesse = %(size)s,
                          name = %(name)s, mime = %(mime)s, datum = %(datum)s
                    WHERE id = %(id)s
                      AND master_id = %(case)s""",
                {"case": self.case_id})

    def _get_rename_query(self):
        return """UPDATE ka_fall_dokumente_tbl_view
                     SET name = %(name)s
                   WHERE id = %(id)s"""

    def _get_delete_query(self):
        return "SELECT delete_fall_dokument_ds(%(id)s)"


class UserDocumentAPI(DocumentAPI):

    def _get_list_query(self):
        return ("""SELECT id, NULL as master_id, groesse, mime, name, datum
                     FROM ka_benutzer_dokumente_tbl_view
                    WHERE {condition}
                 ORDER BY datum DESC, name""",
                {})

    def _get_load_query(self):
        return """SELECT NULL as master_id, groesse, mime,
                         CASE WHEN %(load_contents)s THEN daten ELSE NULL
                         END AS daten,
                         name, datum
                    FROM ka_benutzer_dokumente_tbl_view
                   WHERE id = %(id)s"""

    def _get_load_data_query(self):
        return """SELECT daten
                    FROM ka_benutzer_dokumente_tbl_view
                   WHERE id = %(id)s"""

    def _get_create_query(self, uuid):
        return ("SELECT * FROM create_benutzer_dokument_ds(%(name)s)",
                {})

    def _get_update_query(self):
        return ("""UPDATE ka_benutzer_dokumente_tbl_view
                      SET daten = %(bytes)s, groesse = %(size)s,
                          name = %(name)s, mime = %(mime)s, datum = %(datum)s
                    WHERE id = %(id)s""",
                {})

    def _get_rename_query(self):
        return """UPDATE ka_benutzer_dokumente_tbl_view
                     SET name = %(name)s
                   WHERE id = %(id)s"""

    def _get_delete_query(self):
        return "SELECT delete_benutzer_dokument_ds(%(id)s)"


class LogbookDocumentAPI(DocumentAPI):

    def __init__(self, logbook_id):
        self.logbook_id = logbook_id

    @classmethod
    def list_case_logbook_documents(cls, master_id):
        """List all documents associated with the logbook of the given case.
        """
        with db.cursor() as cursor:
            query = """SELECT id, master_id, logbuch_id, groesse, mime, name,
                              datum
                         FROM ka_logbuch_dokumente_tbl_view
                        WHERE master_id = %(master_id)s
                     ORDER BY datum DESC, name"""
            cursor.execute(query, {"master_id": master_id})
            return [cls._document_from_list_row(row)
                    for row in cursor.fetchall()]

    def _get_list_query(self):
        return ("""SELECT id, master_id, logbuch_id, groesse, mime, name, datum
                     FROM ka_logbuch_dokumente_tbl_view
                    WHERE logbuch_id = %(logbuch_id)s
                      AND {condition}
                 ORDER BY datum DESC, name""",
                {"logbuch_id": self.logbook_id})

    @classmethod
    def _document_from_list_row(cls, row):
        id, case, logbuch_id, size, mime, name, insert_date = row
        return Document(id, case, size, mime, name, insert_date,
                        logbook_id=logbuch_id)

    def used_names(self, cursor, name=None):
        query_args = {}
        condition = "TRUE"
        if name is not None:
            query_args["name"] = name
            condition = "name = %(name)s"
        query = """SELECT id, master_id, logbuch_id, groesse, mime, name, datum
                     FROM ka_logbuch_dokumente_tbl_view
                    WHERE master_id = (SELECT master_id
                                         FROM ka_logbuch_tbl_view
                                        WHERE id = %(logbuch_id)s)
                      AND {condition}
                 ORDER BY datum DESC, name"""
        query_args["logbuch_id"] = self.logbook_id
        cursor.execute(query.format(condition=condition), query_args)
        return [self._document_from_list_row(row) for row in cursor.fetchall()]

    def _get_load_query(self):
        return """SELECT master_id, groesse, mime,
                         CASE WHEN %(load_contents)s THEN daten ELSE NULL
                         END AS daten,
                         name, datum, logbuch_id
                    FROM ka_logbuch_dokumente_tbl_view
                   WHERE id = %(id)s"""

    @classmethod
    def _document_from_load_row(cls, row):
        doc_id, case_id, size, mime, data, name, insert_date, logbook_id = row
        return Document(doc_id, case_id, size, mime, name, insert_date,
                        data, logbook_id)

    def _get_load_data_query(self):
        return """SELECT daten
                    FROM ka_logbuch_dokumente_tbl_view
                   WHERE id = %(id)s"""

    def _get_create_query(self, uuid):
        return ("SELECT *"
                " FROM create_logbuch_dokument_ds(%(id)s, %(uuid)s, %(name)s)",
                {"id": self.logbook_id, "uuid": uuid})

    def _get_update_query(self):
        return ("""UPDATE ka_logbuch_dokumente_tbl_view
                      SET daten = %(bytes)s, groesse = %(size)s,
                          name = %(name)s, mime = %(mime)s, datum = %(datum)s
                    WHERE id = %(id)s
                      AND logbuch_id = %(logbuch_id)s""",
                {"logbuch_id": self.logbook_id})

    def _get_rename_query(self):
        return """UPDATE ka_logbuch_dokumente_tbl_view
                     SET name = %(name)s
                   WHERE id = %(id)s"""

    def _get_delete_query(self):
        return "SELECT delete_logbuch_dokument_ds(%(id)s)"



def get_api(case_id, logbook_id=None):
    if logbook_id is not None:
        return LogbookDocumentAPI(logbook_id)
    elif case_id is not None:
        return CaseDocumentAPI(case_id)
    else:
        return GlobalDocumentAPI()


class Document(object):

    def __init__(self, id=None, case=None, size=None, mime=None, name=None,
                 insert_date=None, data=None, logbook_id=None):
        self.id = id
        self.case = case
        self.size = size
        self.mime = mime
        self.name = name
        self.data = data
        self.insert_date = insert_date
        self.logbook_id = logbook_id

    def getName(self):
        name = self.name
        if name is None:
            return None
        return h.ensure_unicode(name)

    def create(self, name, src, case=None, uuid=None):
        with db.transaction() as cur:
            self.create_with_cursor(cursor, name, src, case=case, uuid=uuid)

    def create_with_cursor(self, cursor, name, src, case=None, uuid=None):
        if case is not None:
            api = CaseDocumentAPI(case)
        else:
            api = GlobalDocumentAPI()

        doc = api.create(cursor, name, src, uuid=uuid)
        self.name = doc.name
        self.mime = doc.mime
        self.case = doc.case
        self.size = doc.size
        self.id = doc.id

    def load(self, asCase=False, load_contents_immediately=True):
        if self.id is None:
            raise DocumentException(_(u"An ID is needed"))

        if asCase:
            api = CaseDocumentAPI(None)
        else:
            api = GlobalDocumentAPI()

        with db.cursor() as cursor:
            doc = api.load(cursor, self.id, load_contents_immediately)
            self.case = doc.case
            self.size = doc.size
            self.mime = doc.mime
            self.data = doc.data
            self.name = doc.name
            self.insert_date = doc.insert_date

    def load_data(self):
        with db.cursor() as cur:
            self.load_data_with_cursor(cur)

    def load_data_with_cursor(self, cursor):
        if self.id is None:
            raise DocumentException(_(u"An ID is needed"))

        api = get_api(self.case, self.logbook_id)
        self.data = api.load_data(cursor, self.id)

    def get_data(self):
        """
        Loads and returns the data of this document
        """
        if self.data is None:
            self.load_data()
        return self.data


def export_case_documents(case, case_id, cursor, selection, anonymize):
    """Add the documents of the case with id case_id to the elementtree case"""
    export_documents_from_api(ET.SubElement(case, "anlagen"),
                              CaseDocumentAPI(case_id), cursor)


def export_logbook_documents(et_parent, logbook_id, cursor, selection,
                             anonymize):
    """Add the documents of a logbook entry to et_parent.

    The logbook entry is given by its ID in logbook_id. The documents
    will be added to the elementtree object et_parent.
    """
    export_documents_from_api(et_parent, LogbookDocumentAPI(logbook_id), cursor)


def export_documents_from_api(et_parent, api, cursor):
    for document in api.list(cursor):
        document.load_data_with_cursor(cursor)
        document_element = ET.SubElement(et_parent, "anlage")
        add_leaf(document_element, "anlage-mime", document.mime)
        add_leaf(document_element, "anlage-name", document.name)
        add_leaf(document_element, "anlage-daten",
                 base64.b64encode(document.data))
        # description is here, because for now the XSD file specifies it
        # as minOccurs="1" for historical reasons
        add_leaf(document_element, "anlage-beschreibung", "")
        add_leaf(document_element, "anlage-datum", document.insert_date)


def new_node(cls, name):
    node = cls()
    node.setName(name)
    return node


document_structure_node = \
    StructureNode(new_node(RepeatNode, "anlage"),
                  [new_node(TextLeaf, "anlage-mime"),
                   new_node(TextLeaf, "anlage-name"),
                   new_node(TextLeaf, "anlage-daten"),
                   new_node(TextLeaf, "anlage-beschreibung"),
                   new_node(DateLeaf, "anlage-datum")
                   ],
                  [])

document_component_factory = \
    ComponentFactory(None,
                     StructureNode(None, [], [document_structure_node]),
                     StructureInstanceNode,
                     StructureInstanceTree)


def import_case_documents(case_tree, cursor, attachment_tree):
    """Import the documents"""
    api = CaseDocumentAPI(case_tree.getRootNode().getIdentifier())
    import_documents_from_api(cursor, api, attachment_tree.root)


def import_documents_from_api(cursor, api, attachment_node):
    documents = attachment_node.get_repeat_groups("anlage")
    if documents:
        known = api.list_as_dict(cursor)
        for document in documents.itervalues():
            name = document.get_leaf("anlage-name")
            if name in known:
                del known[name]
            import_document_from_api(cursor, api, document)
        # delete documents which aren't imported again
        for doc in known.itervalues():
            api.delete(cursor, doc.id)


def import_document_from_api(cursor, api, document):
    name = document.get_leaf("anlage-name")
    mime = document.get_leaf("anlage-mime")
    data = document.get_leaf("anlage-daten")
    if name is None or mime is None or data is None:
        raise DocumentException(_("The document seems to be corrupted. Invalid "
                                  "anlage in document. Missing fields "
                                  "anlage-name, anlage-mime or anlage-daten."))
    content = base64.b64decode(data)
    insert_date = document.get_leaf("anlage-datum")
    api.create(cursor, name, content, mime=mime, insert_date=insert_date)

# vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8:
