# -*- coding: utf-8 -*-
#
# Copyright 2014 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls base
#
# mpuls base is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls base is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls base. If not, see <http://www.gnu.org/licenses/>.

import logging

from pylons import app_globals as g

from mpulsweb.lib.base import render
from mpulsweb.lib.helpers import get_formular_url, tag, get_field_identifier, \
    escape
from mpulsweb.lib.translation import _
from mpulsweb.model.phase import PhaseList, StartPhasePart

log = logging.getLogger(__name__)

EMPTY_VALUES = ["", None, []]

UNBEKANNT = -1
KEINE_ANGABE = -3


def is_set(value):
    return value not in EMPTY_VALUES


def is_not_set(value):
    return value in EMPTY_VALUES


def choice_is_set(value):
    return is_set(value) and value not in [UNBEKANNT, KEINE_ANGABE]


class Condition(object):

    """ A base class for transistion specific derived conditions """

    def check(self, case):
        return True

    def __call__(self, case):
        return self.check(case)


class BaseFieldsCondition(Condition):

    """
    A base class for field checks
    """

    def __init__(self, fields):
        """
        Fields must be a iterable
        """
        self.fields = fields

    def check(self, case):
        for field_name in self.get_fields():
            if not self.check_field(case, field_name):
                return False
        return True

    def get_value(self, case, field_name):
        return case.get_value(field_name)

    def check_field(self, case, field_name):
        value = self.get_value(case, field_name)
        return self.check_value(value)

    def check_value(self, value, req_value=None):
        raise NotImplementedError()

    def get_fields(self):
        """
        Returns the list of field names evaluated in this condition
        """
        return self.fields


class CheckFieldsCondition(BaseFieldsCondition):

    """
    A Condition class for checking values of fields in a case

    Examples:
        CheckFieldsCondition(
            {"firstname": is_set,
            "lastname": is_set,
            "status": 1,
            "value1": [1, 2, 3],
            "value2": lambda x: x != 5,
            })
    """

    def __init__(self, fields):
        """
        Fields must be a dictionary
        """
        self.fields = fields

    def check_field(self, case, field_name):
        value = self.get_value(case, field_name)
        req_value = self.fields[field_name]
        return self.check_value(value, req_value)

    def check_value(self, value, req_value=None):
        if isinstance(req_value, (tuple, list)):
            if value not in req_value:
                return False
        elif callable(req_value):
            if not req_value(value):
                return False
        else:
            if value != req_value:
                return False
        return True

    def get_fields(self):
        return self.fields.keys()


class RequiredFieldsCondition(BaseFieldsCondition):

    """
    A Condition class to check wether the values of all fields are set
    """

    def check_value(self, value, req_value=None):
        """ Returns False if value is not set """
        return is_set(value)


class RequiredChoicesCondition(BaseFieldsCondition):

    """
    A Condition class to check wether the values of all choice fields are set
    and if their values are not UNBEKANNT or KEINE_ANGABE
    """

    def check_value(self, value, req_value=None):
        """ Returns True if the value is set and not UNBEKANNT
        or KEINE_ANGABE
        """
        return choice_is_set(value)


class FieldNotSetCondition(BaseFieldsCondition):

    """
    A Condition class to check wether the value of one of the fields is not set
    """

    def check_value(self, value, req_value=None):
        """ Returns True if value is not set """
        return is_not_set(value)


class ChoiceNotSetCondition(BaseFieldsCondition):

    """
    A Condition class to check wether the value of one of the choice fields is
    not set, is UNBEKANNT or KEINE_ANGABE
    """

    def check_value(self, value, req_value=None):
        """ Returns True if value is not set or UNBEKANNT or
        KEINE ANGABE
        """
        return is_not_set(value) or value in [UNBEKANNT, KEINE_ANGABE]


class RequiredFieldsTransitionDigestRenderer(object):

    def render(self, case, state, transition):
        fields = []
        for con in transition.conditions:
            if isinstance(con, BaseFieldsCondition):
                for field in con.get_fields():
                    valid = con.check_field(case, field)
                    fields.append((valid, field))

        meta_tree = case.get_formed_instance_tree().getFormEd()

        out = []
        for valid, field_name in sorted(fields, key=lambda val: val[1]):
            out.extend(self.get_anchor(case, meta_tree, field_name, valid))
        return "".join(out)

    def get_field_identifier(self, case, field_name):
        return get_field_identifier(case.id, field_name,
                                    case.get_formed_instance_tree())

    def get_formular_url(self, case, meta_tree, field_name, identifier):
        return get_formular_url(meta_tree, identifier, field_name)

    def get_anchor(self, case, meta_tree, field_name, valid):
        identifier = self.get_field_identifier(case, field_name)
        if identifier is None:
            return []

        nc = meta_tree.findByName(field_name)
        return [tag("a", class_=self.get_style(case, nc, field_name, valid),
                    href=self.get_formular_url(case, meta_tree, field_name,
                                               identifier),
                    title_=self.get_title(case, nc, field_name, valid)),
                escape(self.get_label(case, nc, field_name, valid)),
                "</a>\n"]

    def get_label(self, case, meta, field_name, valid):
        return meta.getDescription()

    def get_style(self, case, meta, field_name, valid):
        if valid:
            return "ok"
        return "error"

    def get_title(self, case, meta, field_name, valid):
        if valid:
            return _("Field is OK and requires no further editing")
        return _("Field is not OK and requires further editing")


class Transition(object):

    """
    A transition between two states

    Set conditions to check the case for specific rules that must be required
    for following the transtition.
    """

    def __init__(self, to_id, conditions=None,
                 digest_renderer_cls=RequiredFieldsTransitionDigestRenderer):
        self.to_id = to_id
        self.conditions = conditions or []
        self.digest_renderer_cls = digest_renderer_cls

    def check_conditions(self, case):
        return all(cond(case) for cond in self.conditions)

    def get_to_state_id(self):
        return self.to_id

    def get_digest_renderer(self):
        return self.digest_renderer_cls()

    def render_digest(self, case, state):
        if self.digest_renderer_cls is None:
            return ""
        return self.get_digest_renderer().render(case, state, self)

    def __repr__(self):
        return "<%s to_state_id=%r id=0x%x>" % (
            self.__class__.__name__, self.get_to_state_id(), id(self))


class AutomaticTransition(Transition):

    """ A transition which may be passed without any user interaction """


class PhaseState(object):

    """ A state class for mpuls phases """

    def __init__(self, state_id, name, transitions=None, display_name=None,
                 icon_name=None, digest_template=None, sort_value=None):
        self.state_id = state_id
        self.name = name
        self.icon_name = icon_name or name[0].lower()
        self.transitions = transitions or []
        if display_name is None:
            self.display_name = name
        else:
            self.display_name = display_name
        self.digest_template = digest_template
        if sort_value is None:
            self.sort_value = state_id
        else:
            self.sort_value = sort_value

    def get_transitions(self):
        return self.transitions

    def get_name(self):
        return self.name

    def get_id(self):
        return self.state_id

    def get_display_name(self):
        return self.display_name

    def get_icon_name(self):
        return self.icon_name

    def get_digest_template(self):
        return self.digest_template

    def get_sort_value(self):
        return self.sort_value

    def __eq__(self, other):
        if not isinstance(other, PhaseState):
            return False
        return self.get_id() == other.get_id()

    def __ne__(self, other):
        if not isinstance(other, PhaseState):
            raise True
        return self.get_id() != other.get_id()

    def __repr__(self):
        return "<%s state_id=%r name=%r display_name=%r id=0x%x" % (
            self.__class__.__name__, self.get_id(), self.get_name(),
            self.get_display_name(), id(self))


class StateHelper(object):

    """ A tiny empty helper class for StateMachine

    The class is used to get the states by id and name on a StateMachine
    """


class StateDigestRenderer(object):

    """ A class to render a StateDigest

    The digest is StateMachine dependend. Therefore the rendering is abstracted
    to an own class. For adjusting the state digest either the state/digest.mako
    template must be overridden or an own StateDigestRenderer class must be
    implemented.
    """

    template_name = "/state/digest.mako"

    def __init__(self, template_name=None):
        self.template_name = template_name or self.template_name

    def get_template_name(self, state):
        if state and state.get_digest_template():
            return state.get_digest_template()
        return self.template_name

    def render(self, case, state, extra_context=None):
        context = {
            "case": case,
            "state": state,
        }
        context.update(extra_context)
        template_name = self.get_template_name(state)
        return render(template_name, extra_vars=context)


class BaseStateMachine(object):

    TRANSITION_MODE_MANUAL = 0
    TRANSITION_MODE_HALFAUTO = 1
    TRANSITION_MODE_FULLAUTO = 2

    def __init__(self, transition_mode=TRANSITION_MODE_FULLAUTO,
                 state_digest_renderer_cls=StateDigestRenderer):
        self.transition_mode = transition_mode
        self.state_digest_renderer_cls = state_digest_renderer_cls

    def is_transition_mode_auto(self):
        """
        Returns True if the current transition mode is an auto mode

        An auto mode may change the state without explicit user interaction.
        """
        return self.transition_mode in [self.TRANSITION_MODE_FULLAUTO,
                                        self.TRANSITION_MODE_HALFAUTO]

    def get_transition_mode(self):
        """
        Returns the current transition mode of the StateMachine
        """
        return self.transition_mode

    def get_next_state(self, case):
        """
        Returns the next PhaseState which can be set automatically

        It returns the current state if the state can't be changed.
        """
        raise NotImplementedError()

    def get_current_state(self, case):
        """
        Returns the current PhaseState of the case
        """
        raise NotImplementedError()

    def get_current_state_id(self, case):
        """
        Returns the current state id of the case
        """
        raise NotImplementedError()

    def is_reachable(self, case, state_id):
        """
        Returns True if the state_id can be set as the next state

        In most cases this method is used to check wether it is allowed to set
        a state for a case.
        """
        raise NotImplementedError()

    def is_end_state(self, state_id):
        """
        Returns True if the state_id belongs to an end state

        The meaning of an end state belongs to the implementation of the
        StateMachine.
        """
        raise NotImplementedError()

    def is_active_state(self, state_id):
        """
        Returns True if the state_id belongs to an active state
        """
        raise NotImplementedError()

    def is_anonymizable_state(self, state_id):
        """
        Returns True if the state_id belongs to an anonymizeable state
        """
        raise NotImplementedError()

    def is_running_state(self, state_id):
        """
        Returns True if the state_id belong to a running state
        """
        raise NotImplementedError()

    def get_state_legend(self):
        """
        Returns a list of (ID, Description) tuples for displaying a state legend
        """
        raise NotImplementedError()

    def __getitem__(self, key):
        """
        Returns the State with key as state_id from this StateMachine or a
        KeyError if not found
        """
        raise NotImplementedError()

    def get(self, key, default=None):
        """
        Returns the State with key as state_id from this StateMachine or default
        if not found.
        """
        raise NotImplementedError()

    def items(self):
        """
        Returns state_id, PhaseState pairs for all states in the StateMachine
        """
        raise NotImplementedError()

    def values(self):
        """
        Returns PhaseStates for all states in the StateMachine
        """
        raise NotImplementedError()

    def get_search_items(self):
        """
        Returns item_id, description pairs for using in searches

        The item_id MAY NOT reflect the a state_id.
        """
        raise NotImplementedError()

    def get_search_item_ids(self):
        """
        Returns all item_ids for search

        Usually this method returns all item ids which should be searched for
        per default
        """
        raise NotImplementedError()

    def get_state_ids_from_search_item(self, item_id):
        """
        Returns a list of state_ids for the corresponding item_id or an empty
        list if no state_id could be found for the item_id
        """
        raise NotImplementedError()

    def get_initial_state_ids(self):
        """
        Returns a list of state_ids for initial/start/unknown states
        """
        raise NotImplementedError()

    def get_end_state_ids(self):
        """Returns a list of state_ids for end states.
        The return value should be equivalent to a list consisting of
        all ids returned by get_search_item_ids for which is_end_state
        returns true.
        """
        raise NotImplementedError()

    def get_active_state_ids(self):
        """Returns a list of state_ids for active, i.e. non-end, states.
        The set of ids returned by this method is effectively the
        complement of the one returned by get_end_state_ids(), thus the
        return value should be equivalent to a list consisting of all
        ids returned by get_search_item_ids for which is_end_state
        returns false.
        """
        raise NotImplementedError()

    def get_state_digest_renderer(self):
        """
        Returns the used StateDigestRenderer class for this StateMachine
        """
        return self.state_digest_renderer_cls()

    def render_state_digest(self, case, **kwargs):
        """
        Renders the state digest for this StateMachine and case
        """
        kwargs["statemachine"] = self
        renderer = self.get_state_digest_renderer()
        return renderer.render(case, self.get_current_state(case),
                               extra_context=kwargs)

    def get_all_state_ids(self):
        """
        Returns all state ids in this StateMachine
        """
        raise NotImplementedError()

    def get_sorted_states(self):
        """
        Returns all states in a sorted order
        """
        raise NotImplementedError()

    def get_successor_state_ids(self, item_id):
        """
        Returns all state ids of the successors for the item
        """
        raise NotImplementedError()

    def get_state_datefield_names(self, item_id):
        """
        Returns the names of the date fields for a item_id
        """
        return NotImplementedError()

    def get_end_state_for_state_id(self, state_id):
        """
        Returns the end PhaseState for a state id
        """
        return NotImplementedError()

    def get_end_state_id_for_state_id(self, state_id):
        """
        Returns the state_id for the end state of a state_id
        """
        return NotImplementedError()


class StateMachine(BaseStateMachine):

    """ A state machine for implementing a state diagram in python """

    def __init__(self, states,
                 state_digest_renderer_cls=StateDigestRenderer,
                 transition_mode=BaseStateMachine.TRANSITION_MODE_FULLAUTO,
                 end_state_ids=None):
        super(StateMachine, self).__init__(
            transition_mode, state_digest_renderer_cls=state_digest_renderer_cls
        )
        self.states = states
        self.end_state_ids = end_state_ids or []
        self.build_mapping()
        self.check_consistency()

    def _get_automatic_transitions(self, transitions):
        return [trans for trans in transitions
                if isinstance(trans, AutomaticTransition)]

    def _get_next_auto_state(self, next_state, case):
        transitions = next_state.get_transitions()

        for trans in self._get_automatic_transitions(transitions):
            if trans.check_conditions(case):
                next_state = self.get(trans.get_to_state_id())
                return self._get_next_auto_state(next_state, case)

        return next_state

    def _is_reachable_state_id(self, case, start_id, end_id):
        for trans in self[start_id].get_transitions():
            if trans.check_conditions(case):
                state_id = trans.get_to_state_id()
                if state_id == end_id:
                    return True
                return self._is_reachable_state_id(case, state_id, end_id)
        return False

    def get_next_state(self, case):
        if self.is_transition_mode_auto():
            return self._get_next_auto_state(self.get_current_state(case), case)
        else:
            return None

    def get_current_state(self, case):
        return self[self.get_current_state_id(case)]

    def get_current_state_id(self, case):
        return case.getPhase()

    def build_mapping(self):
        self.values = StateHelper()

        mapping = {}
        for s in self.states:
            state_id = s.get_id()
            name = s.get_name()
            if name in mapping:
                raise ValueError("Duplicate state name %r" % name)
            if state_id in mapping:
                raise ValueError("Duplicate state id %r" % state_id)
            mapping[name] = mapping[state_id] = s
            setattr(self.values, name, state_id)

        self.mapping = mapping

    def check_consistency(self):
        for s in self.states:
            for t in s.get_transitions():
                if t.get_to_state_id() not in self.mapping:
                    raise ValueError("State %r defines a transition to unknown"
                                     " state %r" % (s.get_state_id(),
                                                    t.get_to_state_id()))

    def is_reachable(self, case, state_id):
        current_id = self.get_current_state_id(case)
        return self._is_reachable_state_id(case, current_id, state_id)

    def is_end_state(self, state_id):
        return state_id in self.end_state_ids

    def __getitem__(self, key):
        return self.mapping[key]

    def get(self, key, default=None):
        return self.mapping.get(key, default)

    def items(self):
        return [(state.get_id(), state) for state in self.states]

    def values(self):
        return self.states

    def get_search_items(self):
        return [(state.get_id(), state.get_display_name()) for state in
                self.get_sorted_states()]

    def get_search_item_ids(self):
        return self.get_all_state_ids()

    def get_state_ids_from_search_item(self, item_id):
        state = self.get(item_id)
        if state is not None:
            return [state.get_id()]
        return []

    def get_all_state_ids(self):
        return [state.get_id() for state in self.states]

    def get_state_legend(self):
        return [(state.get_icon_name().upper(), state.get_display_name())
                for state in self.get_sorted_states()]

    def get_initial_state_ids(self):
        # per default the first state should be the initial state
        return [self.states[0]]

    def get_end_state_ids(self):
        return self.end_state_ids

    def get_active_state_ids(self):
        return [state_id for state_id in self.get_search_item_ids()
                if not self.is_end_state(state_id)]

    def get_state_digest_renderer(self):
        return self.state_digest_renderer_cls("/state/digest_statemachine.mako")

    def get_sorted_states(self):
        return sorted(self.states, key=lambda s: s.get_sort_value())


class PhaseStateMachine(BaseStateMachine):

    """ A StateMachine that uses the legacy Phase config mode """

    UNKNOWN_STATE_ID = -1

    def __init__(self):
        transition_mode = int(g.mpuls_config.get('phases', 'changemode'))
        super(PhaseStateMachine, self).__init__(transition_mode)
        self._phaselist = None
        self._current_loaded_case_id = None

    @property
    def phaselist(self):
        """
        Lazy loading of phaselist
        """
        if self._phaselist is None:
            self._phaselist = PhaseList()
        return self._phaselist

    def _phase_part_id_to_phase_state(self, phase_id):
        phasepart = self._get_phase_part_from_id(phase_id)
        if not phasepart:
            raise ValueError("Unkown phase part for id %r" % phase_id)
        return self._phase_part_to_phase_state(phasepart)

    def _phase_part_to_phase_state(self, phasepart):
        desc = phasepart.getDescription()
        icon = phasepart.getIcon()
        phase_id = phasepart.id
        name = phase_id
        return PhaseState(phase_id, name, display_name=desc, icon_name=icon)

    def _get_phase_part_from_id(self, phase_id):
        return self.phaselist.get_phase(phase_id)

    def _load_case(self, case):
        if case.id != self._current_loaded_case_id:
            self.phaselist._load(case.id)
            self._current_loaded_case_id = case.id

    def get_current_state_id(self, case):
        return case.getPhase()

    def get_current_state(self, case):
        phase_id = self.get_current_state_id(case)
        return self._phase_part_id_to_phase_state(phase_id)

    def get_next_state(self, case):
        self._load_case(case)
        phase_id = self.phaselist.get_max_phase_id()
        return self._phase_part_id_to_phase_state(phase_id)

    def __getitem__(self, key):
        phasepart = self._get_phase_part_from_id(key)
        if not phasepart:
            raise KeyError("Unknown phase part id %s" % key)
        return self._phase_part_to_phase_state(phasepart)

    def get(self, key, default=None):
        phasepart = self._get_phase_part_from_id(key)
        if not phasepart:
            return default
        return self._phase_part_to_phase_state(phasepart)

    def is_reachable(self, case, state_id):
        self._load_case(case)
        return self.phaselist.is_on_reachable_path(state_id)

    def is_end_state(self, state_id):
        return str(state_id) in g.mpuls_config.get('phases', 'finished')

    def get_state_legend(self):
        legend = []
        descriptions = g.mpuls_config.get('phases', 'shortdescription')[0]

        for p in self.phaselist:
            legend.append((descriptions[str(p.id)], p.getDescription()))

        return legend

    def items(self):
        items = []
        for p in self.phaselist:
            items.append((p.start.id,
                          self._phase_part_id_to_phase_state(p.start.id)))
            items.append((p.end.id,
                          self._phase_part_id_to_phase_state(p.end.id)))
        return items

    def values(self):
        items = []
        for p in self.phaselist:
            items.append(self._phase_part_id_to_phase_state(p.start.id))
            items.append(self._phase_part_id_to_phase_state(p.end.id))
        return items

    def get_search_items(self):
        return [(p.id, p.getDescription()) for p in self.phaselist]

    def get_search_item_ids(self):
        items = []
        for p in self.phaselist:
            items.append(p.start.id)
            items.append(p.end.id)
        return items

    def get_state_ids_from_search_item(self, item_id):
        p = self.phaselist.get(item_id)
        if p:
            return [p.start.id, p.end.id]
        return []

    def get_initial_state_ids(self):
        return [self.UNKNOWN_STATE_ID]

    def get_end_state_ids(self):
        return [int(i) for i in g.mpuls_config.get('phases', 'finished')]

    def get_active_state_ids(self):
        return [int(i) for i in g.mpuls_config.get('phases', 'active')]

    def render_state_digest(self, case, **kwargs):
        self._load_case(case)
        kwargs["phases"] = self.phaselist
        return super(PhaseStateMachine, self).render_state_digest(case,
                                                                  **kwargs)

    def get_all_state_ids(self):
        return [p.id for p in self.phaselist]

    def get_state_digest_renderer(self):
        return self.state_digest_renderer_cls(
            "/state/digest_phasestatemachine.mako")

    def get_successor_state_ids(self, phase_id):
        """The argument phase_id should be an Integer"""
        suc = []
        phase = self.phaselist.get(phase_id)
        if not phase:
            log.warn("Don't find phase %s" % str(phase_id))
            return suc

        suc.append(phase.start.id)
        suc.append(phase.end.id)
        for ppid in phase.end.suc:
            pp = self._get_phase_part_from_id(ppid)
            suc.extend(self.get_successor_state_ids(int(pp.phase)))
        return suc

    def get_state_datefield_names(self, phase_id):
        """
        Returns the start and end field names for the phase id
        """
        phase = self.phaselist.get(phase_id)
        return phase.start.datefield, phase.end.datefield

    def get_end_state_id_for_state_id(self, state_id):
        """
        Returns the phase part id for the end phase part of a state_id

        The state_id is also a phase part id.
        """
        if state_id % 2:
            end_phase = state_id
        elif state_id == 0:
            end_phase = 1
        else:
            end_phase = state_id + 1
        return end_phase

    def get_end_state_for_state_id(self, state_id):
        """
        Returns the PhaseState for the end phase part of a state_id

        The state_id is a phase part id.
        """
        end_phase = self.get_end_state_id_for_state_id(state_id)
        return self._phase_part_id_to_phase_state(end_phase)

    def is_anonymizable_state(self, state_id):
        return str(state_id) not in g.mpuls_config.get('phases',
                                                       'notanonymizable')

    def is_active_state(self, state_id):
        return str(state_id) in g.mpuls_config.get('phases', 'active')

    def is_running_state(self, state_id):
        return isinstance(self._get_phase_part_from_id(state_id),
                          StartPhasePart)
