# -*- coding: latin-1 -*-
#
# Copyright 2007, 2008, 2013 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

import sys
import os
import re
import logging

from datetime import date, time, timedelta
from decimal import Decimal

from formed.meta.nodecomponents import Node, Leaf, CaseNode, \
    RootNode, NodeComponent, VISIT_IGNORE_CHILDREN, VISIT_CONTINUE
from formed.meta.exprtree import Expr


log = logging.getLogger(__name__)


ORDER_RE = re.compile("([^;]*);(.*)")


def int_or_none(string_value):
    """Convert the string parameter to int. If the string is empty, return None.
    """
    if not string_value:
        value = None
    else:
        value = int(string_value)
    return value


def float_or_none(string_value):
    """Convert the string parameter to flat. If the string is empty, return None.
    """
    if not string_value:
        value = None
    else:
        value = float(string_value)
    return value


def decimal_or_none(string_value):
    """
    Convert the string parameter to Decimal. If the string is empty, None is
    returned.
    """

    if not string_value:
        value = None
    else:
        value = Decimal(string_value)
    return value


class RelevantVisitor(object):

    def __init__(self, callback=None):
        self.callback = callback

    def visitor(self, nc):
        if isinstance(nc, (RadioNode, ChoiceNode)):
            self.callback(nc)
            return VISIT_IGNORE_CHILDREN
        if isinstance(nc, LogbookNode):
            # eat it!
            return VISIT_IGNORE_CHILDREN
        if isinstance(nc, (CaseNode, RootNode)):
            return VISIT_CONTINUE
        self.callback(nc)
        return VISIT_CONTINUE


class WidgetCollector(object):

    def __init__(self, root=None):
        self.widgets = []
        self.root = root

    def visitor(self, nc):
        if isinstance(nc, Node):
            if isinstance(nc, (RadioNode, ChoiceNode)):
                self.widgets.append(nc)
                return VISIT_IGNORE_CHILDREN
            if isinstance(nc, RepeatNode):
                return self.handleRepeatNode(nc)
        elif (isinstance(nc, Leaf)
              and not isinstance(nc, (RuleLeaf, ExternalChoiceListLeaf,
                                      InfoLeaf, SaveLeaf, ExprLeaf))):
            self.widgets.append(nc)

        return VISIT_CONTINUE

    def handleRepeatNode(self, repeat):
        if repeat != self.root:
            return VISIT_IGNORE_CHILDREN
        return VISIT_CONTINUE


class ConditionalWidgetCollector(object):

    def __init__(self, evaluator=lambda expr: True):
        self.widgets = []
        self.evaluator = evaluator

    def visitor(self, nc):
        if isinstance(nc, Node):
            if isinstance(nc, ConditionalNode):
                expr = nc.getExpr()
                if expr and not self.evaluator(expr):
                    return VISIT_IGNORE_CHILDREN
            elif isinstance(nc, (RadioNode, ChoiceNode)):
                self.widgets.append(nc)
                return VISIT_IGNORE_CHILDREN
            elif isinstance(nc, GroupNode) and nc.isRepeat() and nc != self:
                return VISIT_IGNORE_CHILDREN
        elif (isinstance(nc, Leaf)
              and not isinstance(nc, (RuleLeaf, ExternalChoiceListLeaf,
                                      InfoLeaf, SaveLeaf))):
            self.widgets.append(nc)

        return VISIT_CONTINUE


class FilteredNodeComponentProxy(object):

    def __init__(self, params):
        self.params = params

    def set(self, nc):
        self.nc = nc

    def getName(self):
        if not self.nc:
            return ""
        return self.nc.getName()

    def allWidgets(self):
        if self.nc:
            print >> sys.stderr, self.nc
            return [w for w in self.nc.allWidgets()
                    if w.getName() in self.params]
        return []


class PageNode(Node):

    def __init__(self):
        super(PageNode, self).__init__()
        self.attributes["invisible"] = ""

    def __repr__(self):
        return "<PageNode 0x%x name=%r>" % (id(self), self.getName())

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def allWidgets(self):
        collector = WidgetCollector()
        self.visit(collector.visitor)
        return collector.widgets


class LogbookNode(Node):

    pass


class GroupNode(Node):

    def __init__(self):
        super(GroupNode, self).__init__()
        self.attributes["repeat"] = ""
        self.attributes["invisible"] = ""
        self.attributes["folding"] = ""

    def __repr__(self):
        return "<GroupNode 0x%x name=%r>" % (id(self), self.getName())

    def isRepeat(self):
        return self.getAttribute("repeat") and True or False

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def isFolding(self):
        return self.getAttribute("folding") and True or False

    def allWidgets(self):
        collector = WidgetCollector()
        self.visit(collector.visitor)
        return collector.widgets


class RowNode(Node):

    pass


class RepeatNode(Node):

    # The unique flag can be used to impose a UNIQUE constraint on the
    # instances of the repeat group. Among all instances with the same
    # parent ID, no two instances must have the same values for all of
    # the fields given by the unuique flag.
    #
    # The 'unique' flag is optional, but when given must have at least
    # one field name as parameter
    #
    # Example:
    #   <repeat ... flags="unique:rg_koop_2_institution_art" ...>

    default_flags = Node.default_flags.copy()
    default_flags["unique"] = None

    flag_splitters = Node.flag_splitters.copy()
    flag_splitters["unique"] = lambda x: x.split(",")

    def __init__(self):
        super(RepeatNode, self).__init__()
        self.attributes["min"] = ""
        self.attributes["max"] = ""
        self.attributes["initial"] = ""
        self.attributes["digest"] = ""
        self.attributes["orderby"] = ""
        self.attributes["reverse"] = ""
        self.attributes["showcontrols"] = None
        self.attributes["insidefolding"] = ""
        self.show_controls = None

    def __repr__(self):
        return ("<RepeatNode 0x%x name=%r min=%r max=%r>"
                % (id(self), self.getName(), self.getMin(), self.getMax()))

    def getMin(self):
        """Return the minimal number of instances of the repeat group.
        The return value is an integer.  An empty string as the value of
        the XML attribute is interpreted as 0.
        """
        min_str = self.getAttribute("min")
        if min_str:
            return int(min_str)
        return 0

    def getInitial(self):
        """Return the initial number of instances of the repeat group.
        The return value is the maximum of the values specified by the
        'initial' XML attribute and the return value of the getMin()
        method. An empty string as the value of the XML attribute
        'initial' is interpreted as 0.

        The return value is an integer.
        """
        initial_str = self.getAttribute("initial")
        if initial_str:
            initial = int(initial_str)
        else:
            initial = 0
        return max(initial, self.getMin())

    def getMax(self):
        """Return the maximal number of instances of the repeat group.
        The return value is an integer if the XML attriute has been set
        and is not empty, otherwise it is None, indicating that the
        number of instances is unlimited.
        """
        max_str = self.getAttribute("max")
        if max_str:
            return int(max_str)
        return None

    def getDigest(self):
        """Return a list with the names of the fields to show in the digest"""
        return self.getAttribute("digest").split()

    def getOrderBy(self):
        """Return the parsed contents of the orderby attribute.
        The return value is a list of (FIELDNAME, DIRECTION) pairs,
        where
        - FIELDNAME is the name of a field in the repeat group and
        - DIRECTION is either
          - one of the strings 'asc' and 'desc' specifying whether to
            order in ascending or descending order, respectively or
          - a list of values the field is expected to contain.
            The order of values is used for ordering the repeat groups.
            The list is derived from a colon separated tuple, e.g.
            (1:3:2)

        For backwards compatibility, the attribute 'reverse' is also
        taken into account in once specific circumstance. If the orderby
        attribute consists of exactly one fieldname and no direction
        specification, the implicit direction is 'desc' instead of 'asc'
        if the reverse attribute is not empty (as determined by the
        getReverse() method).
        """
        raw = self.getAttribute("orderby")
        parsed = []
        if not raw:
            digest = self.getDigest()
            if digest:
                raw = digest[0]
            else:
                return parsed

        for field_spec in raw.split(","):
            split = field_spec.split()
            if len(split) == 1:
                field = split[0]
                direction = "asc"
            elif len(split) == 2:
                field, direction = split
                direction = direction.lower()
                if direction not in ("asc", "desc"):
                    if direction[0] == '(':
                        direction = map(int, direction.strip('()').split(':'))
                    else:
                        raise ValueError("the direction for an orderby field"
                                     " must be 'asc', 'desc' or a"
                                     " colon separated tuple, found %r"
                                     % (direction,))
            else:
                raise ValueError("Each part of the orderby attribute must"
                                 " have one or two words.")
            parsed.append((field, direction))

        if self.getReverse():
            if len(parsed) != 1:
                raise ValueError("Attribute 'reverse' set but more than one"
                                 " field in 'orderby' attribute")
            if parsed[0][1] != "asc":
                raise ValueError("Attribute 'reverse' set but 'orderby'"
                                 " attribute already specifies 'desc' for"
                                 " the one field it contains")
            parsed = [(parsed[0][0], "desc")]

        return parsed

    def getReverse(self):
        return bool(self.getAttribute("reverse"))

    def setAttribute(self, attrib, value):
        if attrib == "showcontrols":
            newExpr = Expr(value)
            newExpr.compile()
            self.show_controls = newExpr
        super(RepeatNode, self).setAttribute(attrib, value)

    def getShowControls(self):
        return self.show_controls

    def setShowControls(self, value):
        self.setAttribute("showcontrols", value)

    def isInsideFolding(self):
        return self.getAttribute("insidefolding") and True or False


class ConditionalNode(Node):

    def __init__(self):
        super(ConditionalNode, self).__init__()
        self.attributes["expr"] = ""
        self.attributes["invisible"] = ""
        self.expr = None

    def __repr__(self):
        return ("<ConditionalNode 0x%x name=%r expr=%r invisible=%r>"
                % (id(self), self.getName(), self.expr.expr,
                   self.isInvisible()))

    def setAttribute(self, attrib, value):
        if attrib == "expr":
            newExpr = Expr(value)
            newExpr.compile()
            self.expr = newExpr
        super(ConditionalNode, self).setAttribute(attrib, value)

    def setExpr(self, expr):
        self.setAttribute('expr', expr)

    def getExpr(self):
        return self.expr

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False


class MatrixNode(GroupNode):

    def __repr__(self):
        return "<MatrixNode 0x%x name=%r>" % (id(self), self.getName())


class EnumeratedNode(Node):

    def __init__(self):
        super(EnumeratedNode, self).__init__()
        self.options = None
        self.option_desc_mapping = None
        self.option_nodes = None
        self.option_mapping = None
        self.option_name_mapping = None

    def firstFindDefault(self):
        value = None
        if self.children:
            for child in self.children:
                if isinstance(child, ExternalChoiceListLeaf):
                    found, v = child.find_default()
                    if found:
                        value = v
                        break
                elif isinstance(child, BoolLeaf):
                    if child.getChecked():
                        v = child.getValue()
                        if v is not None:
                            value = v
                        break

        self.found_default = value
        return super(EnumeratedNode, self).firstFindDefault()

    def getValueMapping(self):
        """Returns a dictionary mapping the values of each option in the choice
        list to its description"""
        if self.option_desc_mapping is None:
            self.option_desc_mapping = dict(option
                                            for option in self.get_options())
        return self.option_desc_mapping

    def get_option(self, value, default=None):
        """Return the node (BoolLeaf) for the value."""
        if self.option_mapping is None:
            self.option_mapping = dict((node.getValue(), node)
                                       for node in self.get_option_nodes())
        return self.option_mapping.get(value, default)

    def get_option_by_name(self, name, default=None):
        """Return the node (BoolLeaf) for the value."""
        if self.option_name_mapping is None:
            self.option_name_mapping = dict((node.getName(), node)
                                            for node in self.get_option_nodes())
        return self.option_name_mapping.get(name, default)

    def get_options(self):
        """Return a list of (value, description) tuples with all options.
        The order of the list items is the order that will also be used
        in the renderer form.
        """
        if self.options is None:
            self.options = [(node.getValue(), node.getDescription())
                            for node in self.get_option_nodes()]
        return self.options

    def get_option_nodes(self):
        """Return a list of all the options, i.e. BoolLeaf instances.
        The order of the list items is the order that will also be used
        in the rendered form.
        """
        if self.option_nodes is None:
            nodes = []
            for c in self.walk():
                if isinstance(c, BoolLeaf):
                    nodes.append(c)
                elif isinstance(c, ExternalChoiceListLeaf):
                    nodes.extend(c.getChildren())
            self.option_nodes = nodes
        return self.option_nodes

    def get_table_name(self):
        table_name = self.getAttribute("table_name")
        if table_name is None:
            table_name = self.getName() + "_tbl"
        return table_name

    def getMultiple(self):
        """Return the value of the multiple attribute if the node can have one.

        The default implementation always returns an empty string.
        Subclasses that have a multiple attribute should overwrite this
        method.

        This method is implemented in EnumeratedNode so that code
        working with the formed tree can more easily treat all
        enumerated nodes the same way.
        """
        return ""



class RadioNode(EnumeratedNode):

    def __init__(self):
        super(RadioNode, self).__init__()
        self.attributes["selected"] = ""
        self.attributes["layout"] = "vertical"

    def __repr__(self):
        return ("<RadioNode 0x%x name=%r selected=%r>"
                % (id(self), self.getName(), self.getSelected()))

    def getSelected(self):
        return self.getAttribute("selected")

    def getLayout(self):
        return self.getAttribute("layout")


class BoolLeaf(Leaf):

    def __init__(self):
        super(BoolLeaf, self).__init__()
        self.attributes["checked"] = ""
        self.attributes["invisible"] = ""

    def __repr__(self):
        return ("<BoolLeaf 0x%x name=%r value=%r checked=%r>"
                % (id(self), self.getName(), self.getValue(),
                   self.getChecked()))

    def getChecked(self):
        return self.getAttribute("checked")

    def getValue(self):
        """Return the value associated with the bool leaf as an int.
        If no value has been set, return None.
        """
        return int_or_none(super(BoolLeaf, self).getValue())

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def firstFindDefault(self):
        """Find default value for BoolLeaf as checkbox.
        This method isn't used for BoolLeaf instances used as ChoiceNode
        options.
        """
        self.found_default = bool(self.getChecked())
        return super(BoolLeaf, self).firstFindDefault()


class ExternalChoiceListLeaf(Leaf):

    def __init__(self):
        super(ExternalChoiceListLeaf, self).__init__()
        self.choice = None
        self.path = None

    def __repr__(self):
        return ("<ExternalChoiceListLeaf 0x%x name=%r value=%r>"
                % (id(self), self.getName(), self.getValue()))

    def setLoadContext(self, path):
        self.path = os.path.dirname(path)

    def find_default(self):
        found = False, None
        children = self.getChildren()
        if children:
            for child in children:
                if isinstance(child, BoolLeaf):
                    if child.getChecked():
                        v = child.getValue()
                        if v:
                            found = (True, int(v))
                        break
        return found

    def getChildren(self):
        if not self.choice:
            fname = self.getValue()
            if fname:
                if not os.path.isabs(fname) and self.path:
                    fname = os.path.join(self.path, fname)
                # TODO: This loading code should be else where!
                from formed.meta.io.document import openDocument
                document = openDocument(fname, validate_tree=False)
                self.choice = document.findByClassAndName(ChoiceNode)
                if not self.choice:
                    raise RuntimeError("No ChoiceNode in %s" % fname)
        return self.choice.children


class InfoLeaf(Leaf):

    def __init__(self):
        super(InfoLeaf, self).__init__()
        self.attributes["invisible"] = ""
        self.children = []

    def __repr__(self):
        return "<InfoLeaf 0x%x name=%r>" % (id(self), self.getName())

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False

    def addChild(self, child, idx=-1):
        if idx == -1:
            idx = len(self.children)
        self.children.insert(idx, child)
        if isinstance(child, NodeComponent):
            child.parent = self

    def getChildren(self):
        return self.children

    def get_content(self):
        if self.children:
            return self.children
        return [self.getValue()]


class SaveLeaf(Leaf):

    def __init__(self):
        super(SaveLeaf, self).__init__()
        self.attributes["invisible"] = ""

    def __repr__(self):
        return "<SaveLeaf 0x%x name=%r>" % (id(self), self.getName())

    def isInvisible(self):
        return self.getAttribute("invisible") and True or False


class LeafWithPlaceholder(Leaf):

    def __init__(self):
        super(LeafWithPlaceholder, self).__init__()
        self.attributes["placeholder"] = ""

    def getPlaceholder(self):
        return self.getAttribute("placeholder")


class TextAreaLeaf(LeafWithPlaceholder):

    def __init__(self):
        super(TextAreaLeaf, self).__init__()
        self.attributes["cols"] = ""
        self.attributes["rows"] = ""

    def __repr__(self):
        return ("<TextAreaLeaf 0x%x name=%r cols=%r rows=%r>"
                % (id(self), self.getName(), self.getCols(), self.getRows()))

    def firstFindDefault(self):
        value = self.getValue()
        if not value:
            value = None
        self.found_default = value
        return super(TextAreaLeaf, self).firstFindDefault()

    def getCols(self):
        return self.getAttribute("cols")

    def getRows(self):
        return self.getAttribute("rows")


class TextLeaf(LeafWithPlaceholder):

    def __init__(self):
        super(TextLeaf, self).__init__()
        self.attributes["size"] = ""
        self.attributes["maxlength"] = ""

    def __repr__(self):
        return ("<TextLeaf 0x%x name=%r size=%r maxlength=%r>"
                % (id(self), self.getName(), self.getSize(),
                   self.getMaxLength()))

    def firstFindDefault(self):
        value = self.getValue()
        if not value:
            value = None
        self.found_default = value
        return super(TextLeaf, self).firstFindDefault()

    def getSize(self):
        return self.getAttribute("size")

    def getMaxLength(self):
        """Return the maximum length of the field in characters.
        If the length is unlimited, return None.
        """
        return int_or_none(self.getAttribute("maxlength"))


class ChoiceNode(EnumeratedNode):

    def __init__(self):
        super(ChoiceNode, self).__init__()
        self.attributes["size"] = "1"
        self.attributes["multiple"] = ""
        self.attributes["input"] = ""

    def __repr__(self):
        return "<ChoiceNode 0x%x name=%r>" % (id(self), self.getName())

    def firstFindDefault(self):
        value = super(ChoiceNode, self).firstFindDefault()
        if self.getMultiple() and not isinstance(value, list):
            if value is None:
                value = []
            else:
                value = [value]
            self.found_default = value
        return value

    def getSize(self):
        return self.getAttribute("size")

    def getMultiple(self):
        return self.getAttribute("multiple")

    def getSelected(self):
        return self.getAttribute("selected")

    def getInput(self):
        return self.getAttribute("input")


class IntLeaf(LeafWithPlaceholder):

    def __init__(self):
        super(IntLeaf, self).__init__()
        self.attributes["minvalue"] = ""
        self.attributes["maxvalue"] = ""

    def __repr__(self):
        return ("<IntLeaf 0x%x name=%r minvalue=%r maxvalue=%r>"
                % (id(self), self.getName(), self.getMinValue(),
                   self.getMaxValue()))

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = int(value)
            except ValueError:
                log.warning("Broken default for %r", self.getName())
                value = None
        else:
            value = None
        self.found_default = value
        return super(IntLeaf, self).firstFindDefault()

    def getMinValue(self):
        return int_or_none(self.getAttribute("minvalue"))

    def getMaxValue(self):
        return int_or_none(self.getAttribute("maxvalue"))


class NumericLeaf(LeafWithPlaceholder):

    def __init__(self):
        super(NumericLeaf, self).__init__()
        self.attributes["minvalue"] = ""
        self.attributes["maxvalue"] = ""
        self.attributes["decimals"] = "2"
        self.attributes["stripzeros"] = ""
        self.attributes["step"] = ""

    def __repr__(self):
        return ("<NumericLeaf 0x%x name=%r minvalue=%r maxvalue=%r>"
                % (id(self), self.getName(), self.getMinValue(),
                   self.getMaxValue()))

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = Decimal(value)
            except ValueError:
                log.warning("Broken default for %r", self.getName())
                value = None
        else:
            value = None
        self.found_default = value
        return super(NumericLeaf, self).firstFindDefault()

    def getMinValue(self):
        return decimal_or_none(self.getAttribute("minvalue"))

    def getMaxValue(self):
        return decimal_or_none(self.getAttribute("maxvalue"))

    def getDecimals(self):
        return int(self.getAttribute("decimals"))

    def getStripZeros(self):
        return self.getAttribute("stripzeros")

    def getStep(self):
        return decimal_or_none(self.getAttribute("step"))


class DateLeaf(LeafWithPlaceholder):

    def __init__(self):
        super(DateLeaf, self).__init__()
        self.attributes["order"] = ""

    def __repr__(self):
        return "<DateLeaf 0x%x name=%r>" % (id(self), self.getName())

    def getOrder(self):
        return self.getAttribute("order")

    def get_parsed_order(self):
        match = re.match(ORDER_RE, self.getOrder())
        if match:
            preds = self._parse_order_field_list(match.group(1))
            succs = self._parse_order_field_list(match.group(2))
        else:
            preds = []
            succs = []
        return preds, succs

    def _parse_order_field_list(self, fieldlist):
        fields = []
        for item in fieldlist.split(","):
            item = item.strip()
            if not item:
                continue
            fields.append(item)
        return fields

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = date(*[int(x) for x in value.split("-")])
            except (TypeError, ValueError):
                log.warning("Broken default for %r", self.getName())
                value = None
        else:
            value = None
        self.found_default = value
        return super(DateLeaf, self).firstFindDefault()


class TimeLeaf(LeafWithPlaceholder):

    def __init__(self):
        super(TimeLeaf, self).__init__()

    def __repr__(self):
        return "<TimeLeaf 0x%x name=%r>" % (id(self), self.getName())

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = time(*[int(x) for x in value.split(":")])
            except (TypeError, ValueError):
                log.warning("Broken default for %r", self.getName())
                value = None
        else:
            value = None
        self.found_default = value
        return super(TimeLeaf, self).firstFindDefault()


class DurationLeaf(LeafWithPlaceholder):

    def __init__(self):
        super(DurationLeaf, self).__init__()

    def __repr__(self):
        return "<DurationLeaf 0x%x name=%r>" % (id(self), self.getName())

    def firstFindDefault(self):
        value = self.getValue().strip()
        if value:
            try:
                value = timedelta(int(value))
            except (TypeError, ValueError):
                log.warning("Broken default for %r", self.getName())
                value = None
        else:
            value = None
        self.found_default = value
        return super(DurationLeaf, self).firstFindDefault()


class RuleLeaf(Leaf):

    def __init__(self):
        super(RuleLeaf, self).__init__()
        self.attributes["expr"] = ""
        self.attributes["mark"] = ""
        self.attributes["tag"] = ""
        self.attributes["alwaysavailable"] = None
        self.expr = None

    def __repr__(self):
        return ("<RuleLeaf 0x%x name=%r expr=%r mark=%r>"
                % (id(self), self.getName(), self.expr.expr,
                   self.getMark()))

    def shallowCopy(self):
        r = Leaf.shallowCopy(self)
        if self.expr:
            r.expr = self.expr.clone()
        else:
            r.expr = None
        return r

    def setAttribute(self, attrib, value):
        if attrib == "expr":
            newExpr = Expr(value)
            newExpr.compile()
            self.expr = newExpr
        super(RuleLeaf, self).setAttribute(attrib, value)

    def getExpr(self):
        return self.expr

    def setExpr(self, value):
        self.setAttribute("expr", value)

    def getTag(self):
        return self.getAttribute("tag")

    def getMark(self):
        return self.getAttribute("mark")

    def setMark(self, mark):
        self.setAttribute("mark", mark)

    def isAlwaysAvailable(self):
        """Return whether the rule is always available.
        If a rule is always available, it should be evaluated even if
        the fields it depends on belong to conditionals that are false.

        A rule is always available if it is set to be available or if the
        alwaysavailable attribute is not None and it is an error rule as
        determined by the isError method.
        """
        avail = self.getAttribute("alwaysavailable")
        if avail is None:
            return self.isError()

        return avail

    def isError(self):
        """
        Returns True if the rule is an error rule
        """
        mark = self.getMark()
        return mark and mark.startswith("error")

    def isWarning(self):
        """
        Returns True if the rule is a warning or notice rule
        """
        mark = self.getMark()
        return mark and (mark.startswith("warning")
                         or mark.startswith("notice"))

    def isNotice(self):
        """
        Returns True if the rule is a notice rule
        """
        mark = self.getMark()
        return mark and mark.startswith("notice")

    def getMarkedName(self):
        mark = self.getMark()
        if mark:
            split_mark = mark.split(":")
            if len(split_mark) == 2:
                return split_mark[-1]
            else:
                raise ValueError("There must be exactly one colon in a"
                                 " mark value")
        return None

    def setAlwaysAvailable(self, value):
        self.setAttribute("alwaysavailable", value)


class ReferenceNode(NodeComponent):

    def __init__(self):
        super(ReferenceNode, self).__init__()
        self.attributes["ref"] = ""

    def setReference(self, ref):
        self.setAttribute("ref", ref)

    def getReference(self):
        return self.getAttribute("ref")

    def __repr__(self):
        return "<ReferenceNode 0x%x name=%r ref=%r>" % (
            id(self), self.getName(), self.getReference())


class BaseExprLeaf(Leaf):

    def __init__(self):
        super(BaseExprLeaf, self).__init__()
        self.attributes["expr"] = ""
        self.expr = None

    def getExpr(self):
        return self.expr

    def setExpr(self, value):
        self.setAttribute("expr", value)

    def setAttribute(self, attrib, value):
        if attrib == "expr":
            newExpr = Expr(value)
            newExpr.compile()
            self.expr = newExpr
        super(BaseExprLeaf, self).setAttribute(attrib, value)


class ExprLeaf(BaseExprLeaf):

    def __repr__(self):
        return "<ExprLeaf 0x%x expr=%r>" % (id(self), self.expr.expr)


class VirtualLeaf(BaseExprLeaf):

    def __repr__(self):
        return ("<VirtualLeaf 0x%x name=%r expr=%r>"
                % (id(self), self.getName(), self.expr.expr))


class FileLeaf(Leaf):

    def __init__(self):
        super(FileLeaf, self).__init__()
        self.attributes["multiple"] = ""

    def __repr__(self):
        return ("<FileLeaf 0x%x name=%r>"
                % (id(self), self.getName()))

    def is_multiple(self):
        return self.getAttribute("multiple") and True or False


# vim:set ts=4 sw=4 si et sta sts=4:
