# -*- coding: utf-8 -*-
#
# Copyright 2007, 2008 Intevation GmbH, Germany, <info@intevation.de>
#
# This file is part of mpuls WASKA (CoMPUter-based case fiLeS -
# Web-Anwendungs-Server fuer Kompetenzagenturen).
#
# mpuls WASKA is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# mpuls WASKA is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public
# License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with mpuls WASKA. If not, see <http://www.gnu.org/licenses/>.
#
# mpuls WASKA has been developed on behalf of the
# Projekttraeger im Deutschen Zentrum fuer Luft- und Raumfahrt e.V. (PT-DLR)
# within the programme Kompetenzagenturen (Durchfuehrungsphase) funded by
# the Bundesministerium fuer Familie, Senioren, Frauen und Jugend and
# European Social Fund resources.
#
# Authors:
# Sascha L. Teichmann <teichmann@intevation.de>
#

from sys import maxint
import re
from datetime import date, time, timedelta
from decimal import Decimal, InvalidOperation

import data as data
from formed.meta import UNKNOWN_INT, UNKNOWN_NUMERIC, UNKNOWN_DATE


def ensure_unicode(s):
    """Convert s to a unicode object.
    If it's alredy a unicode object, s is returned as is.  Otherwise
    it's converted to a byte-string which is assumed to be in utf-8 and
    then converted to unicode."""
    if isinstance(s, unicode):
        return s
    return unicode(str(s), 'utf-8')


# Regular expression matching unicode characters that are invalid in XML
# 1.0.
rx_invalidxmlchar = re.compile(u"[^"
                               u"\u0009\u000A\u000D"
                               u"\u0020-\uD7FF\uE000-\uFFFD"
                               u"\U00010000-\U0010FFFF"
                               u"]")


def strip_non_xml_chars(s):
    """Remove all characters from s that are invalid in XML 1.0.

    The definition in XML 1.0 for valid characters is

     #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]

    See http://www.w3.org/TR/2008/REC-xml-20081126/#charsets
    """
    return rx_invalidxmlchar.sub("", s)


class ErrorItem(object):

    SEMANTIC = 0
    RULE = 1

    def __init__(self, type=SEMANTIC, name=None, msg="", bad=None):
        self.type = type
        self.name = name
        self.msg = msg
        self.bad = bad

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return ", ".join(["type: %s" % repr(self.type),
                          "name: %s" % repr(self.name),
                          "msg:  %s" % repr(self.msg),
                          "bad:  %s" % repr(self.bad)])


class SemanticError(Exception):

    def __init__(self, value):
        Exception.__init__(self)
        self.value = value

    def __str__(self):
        return repr(self.value)

    def __repr__(self):
        return "SemanticError(%r)" % (self.value,)


class ValueChecker(object):

    def convert(self, value, nc):
        return value

    def check(self, value, nc):
        return value

    def checkAndConvert(self, value, nc):
        return self.check(self.convert(value, nc), nc)


class BaseTextChecker(ValueChecker):

    def convert(self, value, nc):
        """Convert value to unicode if necessary and remove invalid characters.
        'Invalid' means characters that must not be used in XML 1.0. We
        strip those characters because we usually want to be able to
        export data as XML and the characters that are invalid --
        basically just control characters other than TAB, LF and CR as
        well as surrogate pairs -- are not very useful in normal text
        anyway.
        """
        return strip_non_xml_chars(ensure_unicode(value))


class TextChecker(BaseTextChecker):

    def check(self, value, nc):
        ml = nc.getMaxLength()
        if ml is not None:
            value = value[:ml]
        return value


class TextAreaChecker(BaseTextChecker):

    pass


class IntChecker(ValueChecker):

    def convert(self, value, nc):
        value = value.strip()

        if value[:2].lower() == u"un":
            return UNKNOWN_INT

        try:
            return int(float(value))
        except (ValueError, OverflowError):
            raise SemanticError("'%s' ist keine ganze Zahl." % value)

    def check(self, value, nc):
        # UNKNOWN_INT is always valid, even if it's not in the range
        # required by nc
        if value == UNKNOWN_INT:
            return value

        minv = nc.getMinValue()
        if minv is None:
            minv = -maxint - 1
        maxv = nc.getMaxValue()
        if maxv is None:
            maxv = maxint

        minv, maxv = min(minv, maxv), max(minv, maxv)

        if value < minv:
            raise SemanticError(u"%d ist kleiner als der zulässige"
                                u" Minimalwert %d." % (value, minv))

        if value > maxv:
            raise SemanticError(u"%d ist größer als der zulässige"
                                u" Maximalwert %d." % (value, maxv))

        return value


class NumericChecker(ValueChecker):

    def convert(self, value, nc):
        value = value.strip()

        if value[:2].lower() == u"un":
            return UNKNOWN_NUMERIC

        try:
            t = value.replace(',', '.')
            return Decimal(t)
        except InvalidOperation, ValueError:
            raise SemanticError("'%s' ist keine Zahl." % value)

    def check(self, value, nc):
        minv = nc.getMinValue()
        if minv is None:
            minv = -maxint - 1
        maxv = nc.getMaxValue()
        if maxv is None:
            maxv = maxint

        minv, maxv = min(minv, maxv), max(minv, maxv)

        if value < minv:
            raise SemanticError(u"%s ist kleiner als der zulässige"
                                u" Minimalwert %.2f." % (value, minv))

        if value > maxv:
            raise SemanticError(u"%s ist größer als der zulässige"
                                u" Maximalwert %.2f." % (value, maxv))

        check_decimal_step(value, nc.getStep())

        return value


def check_decimal_step(value, step):
    """Check whether value is a multiple of step.

    If the check fails, a SemanticError is raised.

    The value of step may be None, in which case this function always
    succeeds. None is allowed so that the function can be called with
    the value returned by the node component's getStep() method.
    """
    if step and value % step:
        lower_value = value - value % step
        upper_value = lower_value + step
        raise SemanticError(u"Bitte wählen sie einen gültigen Wert."
                            u" Die zwei nähesten gültigen Werte"
                            u" sind %.2f und %.2f"
                            % (lower_value, upper_value))


def expand_two_digit_year(year):
    """If year was only given with two digits convert it to the intended
    year.  The parameter year should be an int.  If year is less than
    100, the year was specified with only two digits and is actually
    meant to refer to a year in the 20th or 21st centuries.  If the year
    is less than 70 it's assumed to actually refer to 2000 + year,
    otherwise to 1900 + year.
    """
    if year < 100:
        if year < 70:
            year += 2000
        else:
            year += 1900
    return year


class DateChecker(ValueChecker):

    DATE = re.compile(r"(?P<day>[0-9]{1,2})\.(?P<month>[0-9]{1,2})"
                      r"\.(?P<year>[0-9]{4})$")
    DATE_2 = re.compile(r"(?P<day>[0-9]{1,2})\.(?P<month>[0-9]{1,2})"
                        r"\.(?P<year>[0-9]{2})$")
    DATE_ISO = re.compile(r"(?P<year>[0-9]{4})-(?P<month>[0-9]{1,2})"
                          r"-(?P<day>[0-9]{1,2})$")
    DATE_ISO_2 = re.compile(r"(?P<year>[0-9]{2})-(?P<month>[0-9]{1,2})"
                            r"-(?P<day>[0-9]{1,2})$")
    DATE_YEAR_ONLY = re.compile(r"(?P<year>[0-9]{4})$")

    def convert(self, value, nc):
        value = value.strip()

        if value[:2].lower() == u"un":
            return UNKNOWN_DATE

        for reg in (DateChecker.DATE, DateChecker.DATE_ISO,
                    DateChecker.DATE_YEAR_ONLY, DateChecker.DATE_2,
                    DateChecker.DATE_ISO_2):
            m = reg.match(value)
            if not m:
                continue
            try:
                day = int(m.group("day"))
            except IndexError:
                day = 1
            try:
                month = int(m.group("month"))
            except IndexError:
                month = 1
            year = expand_two_digit_year(int(m.group("year")))
            try:
                return date(year, month, day)
            except ValueError:
                raise SemanticError(u"'%s' ist kein gültiges Datum"
                                    u" (Format: TT.MM.JJJJ)." % value)

        raise SemanticError(u"'%s' ist kein gültiges Datum"
                            u" (Format: TT.MM.JJJJ)." % value)


class TimeChecker(ValueChecker):

    TIME = re.compile(r"(?P<hour>[0-9]{1,2}):(?P<minute>[0-9]{1,2})$")

    def convert(self, value, nc):
        value = value.strip()

        m = self.TIME.match(value)
        if not m:
            raise SemanticError(u"'%s' ist keine gültige Uhrzeit"
                                u" (Format: HH:MM)." % value)
        hour = int(m.group("hour"))
        minute = int(m.group("minute"))
        try:
            return time(hour, minute)
        except ValueError:
            raise SemanticError(u"'%s' ist keine gültige Uhrzeit"
                                u" (Format: HH:MM)." % value)


class DurationChecker(ValueChecker):

    TIME = re.compile(r"(?P<hour>[0-9]{1,2}):(?P<minute>[0-9]{1,2})$")

    def convert(self, value, nc):
        value = value.strip()
        try:
            duration = int(value)
        except:
            raise SemanticError(u"Bitte geben Sie die Dauer in Minuten als"
                                u" Zahl an.")
        try:
            return timedelta(minutes=duration)
        except:
            raise SemanticError(u"Bitte geben Sie eine gültige Dauer an.")


class ChoiceChecker(ValueChecker):

    def convert(self, value, nc):
        if nc.getMultiple():
            if not isinstance(value, list):
                value = [value]
            return [int(item) for item in value]
        else:
            return int(value)

    def check(self, value, nc):
        whitelist = nc.getValueMapping()
        if nc.getMultiple():
            unknown = [item for item in value if item not in whitelist]
            if unknown:
                SemanticError("Invalid values for '%r': %s"
                              % (nc.getName(),
                                 ", ".join(str(item) for item in uunknown)))
        else:
            if value not in nc.getValueMapping():
                raise SemanticError("'%s' is not a valid value for '%s'"
                                    % (value, nc.getName()))
        return value



class RadioChecker(ValueChecker):

    def convert(self, value, nc):
        return int(value)

    def check(self, value, nc):
        for c in nc.children:
            if isinstance(c, data.BoolLeaf):
                if value == c.getValue():
                    return value
        raise SemanticError(u"'%s' ist kein valider wert für '%s'"
                            % (value, nc.getName()))


class BoolChecker(ValueChecker):

    def convert(self, value, nc):
        value = value.strip()
        return len(value) and 1 or 0


class FileChecker(ValueChecker):

    def convert(self, value, nc):
        # Assume that something like formencode's variable_decode was
        # used to turn a multi-file upload into a list.
        if nc.is_multiple():
            if not isinstance(value, list):
                value = [value]
            return [self._convert_file(item) for item in value]
        else:
            return self._convert_file(value)

    def _convert_file(self, value):
        return {"filename": value.filename,
                "size": len(value.value),
                "content": value.value}


TYPES = {
    data.IntLeaf: IntChecker(),
    data.NumericLeaf: NumericChecker(),
    data.DateLeaf: DateChecker(),
    data.TimeLeaf: TimeChecker(),
    data.DurationLeaf: DurationChecker(),
    data.TextLeaf: TextChecker(),
    data.TextAreaLeaf: TextAreaChecker(),
    data.ChoiceNode: ChoiceChecker(),
    data.BoolLeaf: BoolChecker(),
    data.RadioNode: RadioChecker(),
    data.FileLeaf: FileChecker(),
}


def checkAndConvert(nc, value):
    try:
        checker = TYPES[nc.__class__]
    except KeyError:
        raise SemanticError("no type check found for '%s'" % str(nc.__class__))
    return checker.checkAndConvert(value, nc)


def check(nc, value):
    """Perform the semantic checks defined by nc on value.
    The nc parameter should be the meta-Node for the value.  The value
    must already be and instance of the native python type for the node.
    Raises SemanticError if the semantic checks fail.
    """
    try:
        checker = TYPES[nc.__class__]
    except KeyError:
        raise SemanticError("no type check found for '%s'" % str(nc.__class__))
    return checker.check(value, nc)


def convert(nc, value):
    """Convert user input into the native python type for the nc.
    """
    try:
        checker = TYPES[nc.__class__]
    except KeyError:
        raise SemanticError("no type check found for '%s'" % str(nc.__class__))
    return checker.convert(value, nc)


# vim:set ts=4 sw=4 si et sta sts=4:
