"""Basic HTML rendering support"""

import os
import logging
import gettext
import datetime
from collections import defaultdict

from formed.meta import UNKNOWN_INT, UNKNOWN_NUMERIC, UNKNOWN_DATE
from formed.meta.nodecomponents import NodeComponent
from formed.meta.data import PageNode, GroupNode, RepeatNode, ConditionalNode, \
    RadioNode, ChoiceNode, ExternalChoiceListLeaf, MatrixNode, RowNode, \
    TextLeaf, TextAreaLeaf, DateLeaf, NumericLeaf, IntLeaf, BoolLeaf, \
    InfoLeaf, SaveLeaf, RuleLeaf, ReferenceNode, ExprLeaf, VirtualLeaf, \
    FileLeaf, TimeLeaf, DurationLeaf
from formed.meta.document import conditional_parents, containing_repeat_group
from formed.instance.formdata import FormData, as_name
from formed.instance.backends.common import RuleEvalContext
from formed.renderer.helpers import escape, NA, tag, folding_helper_functions, \
    folding_init


log = logging.getLogger(__name__)


i18n_base_dir = None


def set_i18n_basedir(basedir):
    """Set the base directory for libformed's translation files.
    This function is only useful in special circumstances, such as when
    libformed is packaged via py2exe.
    """
    global i18n_base_dir
    i18n_base_dir = basedir


def load_catalog(languages):
    """Return message catalog for the languages given by the list languages"""
    global i18n_base_dir
    if not i18n_base_dir:
        i18n_path = os.path.join(os.path.dirname(__file__), "..", "i18n")
        if os.path.isdir(i18n_path):
            i18n_base_dir = i18n_path
    return gettext.translation("libformed", i18n_base_dir, languages,
                               fallback=True)


def hidden_bool_tag(id):
    """Return the hidden HTML input tag for the plain boolean field given by id.

    This hidden input field will have a name derived from id by
    prepending the id with 'bool-'.  The hidden field can be used to
    determine which booleans are present to detect which have been
    unckecked by the user.
    """
    return tag("input", type="hidden", name="bool-" + id)


def apply_hidden_booleans(form_result, value=""):
    """Set the boolean values in form results using the hidden fields.

    In HTML-Forms, check boxes that are not checked when the user
    submits the form are not included in the form results.  They need to
    be included the form results for mpuls, though, so that their values
    are updated properly.  This function uses the hidden fields created
    by hidden_bool_tag to determine which booleans are present in the
    form and adds all booleans which have a hidden field but are not
    included in form_result to form_result with the value given by the
    optional value parameter (default: empty string).  The hidden fields
    are removed from form_result.

    The form_result parameter should be a dictionary mapping HTML form
    field names to values.
    """
    for key in [key for key in form_result.keys() if key.startswith("bool-")]:
        del form_result[key]
        fieldid = key[5:]
        if fieldid not in form_result:
            form_result[fieldid] = value


def hidden_multiselect_tag(id):
    """Return the hidden HTML input tag for the multiselect field given by id.

    This hidden input field will have a name derived from id by
    prepending the id with 'multiselect-'. The hidden field can be used
    to determine which multiselect fields are present to detect for
    which the user has deselected all options.
    """
    return tag("input", type="hidden", name="multiselect-" + id)


def apply_hidden_multiselect(form_result):
    """Set the multiselect values in form results using the hidden fields.

    When a user deselects all options from a multiselect field, the
    browser does not send information about that field to the server.
    This functions adds these fields with an empty list as value for
    those fields which have been added to the form with a hidden input
    field generated by the hidden_multiselect_tag function. The hidden
    fields are removed from form_result.

    The form_result parameter should be a dictionary mapping HTML form
    field names. This function assumes that for multiselect fields the
    values have already been converted to lists.
    """
    for key in [key for key in form_result.keys()
                if key.startswith("multiselect-")]:
        del form_result[key]
        fieldid = key[12:]
        if fieldid not in form_result:
            form_result[fieldid] = []


def extract_hidden_folding(form_result):
    """Extract (hidden) input variables covering folding

    The folding feature uses hidden input fields to store and hand over
    the state of folding groups. These fields are not part of the form
    definition and have to be stripped from the form_result before these are
    stored to the instance tree.

    The function return a list of input fields (key,value)-tuples
    starting with 'folded.', which in turn are deleted from the form_result.
    """

    hidden_folding = []
    for key in [key for key in form_result.keys() if key.startswith("folded.")]:
        hidden_folding.append((key, form_result[key]))
        del form_result[key]
    return hidden_folding


def is_evaluation_relevant(flags):
    return flags.find("evaluation:") >= 0


def group_nesting_level(group):
    """Detemine group nesting level within a page"""
    level = 0
    for nc in reversed(group.pathToRoot()):
        if isinstance(nc, GroupNode):
            level += 1
        elif isinstance(nc, PageNode):
            break
    return level


class Text:

    def __init__(self, txt, htmlid=None, cssclass=None):
        self.txt = txt
        self.htmlid = htmlid
        self.cssclass = cssclass

    def render(self):
        return self.txt


class Item(Text):

    def render(self):
        return self.txt


class Col(Text):

    def render(self, span=None):
        classes = [self.cssclass if self.cssclass is not None else "form-group"]
        classes.append("col-form col-%d" % (span,) if span is not None else "")
        return "".join([tag("div", class_=" ".join(classes)),
                        self.txt,
                        "</div>"])


def repeat_id(full_id):
    """Helper function for TreeFormData"""
    return full_id.split(":", 1)[-1]


class TreeFormData(FormData):

    """FormData class for the full instance tree"""

    def __init__(self, instance, shadow=None, errors=None, node=None):
        """Initialize the form data object.

        The parameter instance is the instance tree.  The parameter
        shadow is a dictionary mapping full instance item names to
        values which will be consulted before the instance tree when
        data is fetched, so that the values in the shadow dict shadow
        the ones from the instance tree.  The parameter errors is a
        dictionary mapping item ids to error messages associated with
        the item.

        The parameter node is used as the initial node on the node
        stack. It should be either a node of the instance tree, or a
        meta node which is then used to find the corresponding instance
        node, or None in which case the root node will be used. The
        default is None.
        """
        super(TreeFormData, self).__init__(instance.getFormEd())
        self.instance = instance
        self.shadow = shadow if shadow is not None else dict()
        self.errors = errors if errors is not None else dict()
        if node is None:
            node = instance.getRootNode()
        elif isinstance(node, NodeComponent):
            case_id = instance.getRootNode().getIdentifier()
            node = instance.getItem(as_name(node) + ":" + unicode(case_id))
        self.node_stack = [node]

    def push_node(self, node):
        self.node_stack.append(node)

    def pop_node(self):
        self.node_stack.pop()

    def get_value(self, meta):
        full_name = self.get_id(meta)
        if full_name in self.shadow:
            return self.shadow[full_name]
        return self.instance.getItem(full_name).getValue()

    def get_repeat_group(self, meta):
        return self.instance.getItem(self.get_id(meta)).getChildren()

    def get_id(self, meta):
        return (as_name(meta) + ":"
                + repeat_id(unicode(self.node_stack[-1].getIdentifier())))

    def get_rg_id(self, rg):
        return rg.getIdentifier()

    def has_errors(self, meta):
        return self.get_id(meta) in self.errors

    def get_errors(self, meta):
        message = self.errors.get(self.get_id(meta))
        if message:
            return (message,)
        return ()

    def get_meta_node_from_item_name(self, item_name):
        return self.get_meta_node(item_name.split(":")[0])

    def full_item_name(self, name):
        return self.get_id(name)

    def form_name_from_full_name(self, full_name):
        return full_name

    def get_variable_provider(self):
        reference = self.node_stack[-1].getIdentifier()
        if isinstance(reference, int):
            # in most cases, the return value of getIdentifier is a
            # string. The only case where it's an int, is if the node is
            # the root node in which case getIdentifier returns the case
            # id as an int. When used as a reference item string, it
            # needs to be prefixed with a colon, so that the code in
            # formed/instance/backends/common.py will parse it
            # correctly.
            reference = ":" + unicode(reference)
        return RuleEvalContext(self.instance, reference)

    def rule_available(self, formed, rule, context):
        """Extend inherited method for required fields in conditionals.

        With client-side conditionals, rules for warnings about required
        fields should always be evaluated so that the client can update
        the warnings when conditionals change. Other rules are delegated
        to the parent class.
        """
        if rule.isWarning() and rule.getName().startswith("required-"):
            return True
        return super(TreeFormData, self).rule_available(formed, rule, context)


class UrlFactory(object):

    def url(self, controller, action, id=None):
        """Return the URL for the given controller, action and id.
        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError

    def icon(self, name):
        """Return the URL for the icon indicated by name.
        This method must be overwritten in derived classes.  The default
        implementation always raises NotImplementedError.
        """
        raise NotImplementedError


class GermanFormatter(object):

    """Format values using the conventions used in Germany.

    Ideally, the formatting should be based on the current locale, or be
    set depending on the preferences of the user (which is not the same
    in a web-application), but the german conventions are used in many
    places in libformed, especially the user input parsers in
    formed.meta.semantic, and the applications using it, and therefore
    this is the default formatter used in the renderer.
    """

    def format_numeric(self, meta, value):
        """Format the decimal object value with comma as decimal separator."""
        formatted = (u"%.*f" % (meta.getDecimals(), value)).replace(".", ",")
        if meta.getStripZeros():
            formatted = formatted.rstrip("0,")
        return formatted

    def format_date(self, meta, date):
        """Format the date object value using the format 'DD.MM.YYYY'."""
        return u"%02d.%02d.%d" % (date.day, date.month, date.year)

    def format_time(self, meta, time):
        """Format the time object value using the format 'HH:MM'."""
        return time.strftime("%H:%M")


class BaseRenderer(object):

    """Base class for renderers that use a FormData object and a UrlFactory.
    """

    def __init__(self, data, view_root, url_factory, read_only=True,
                 languages=None):
        self.data = data
        self.view_root = view_root
        self.url_factory = url_factory
        self.formatter = GermanFormatter()
        self.read_only = read_only
        self.translation = load_catalog(languages)
        self.conditional_css_class = None

    def url(self, controller, action, id=None):
        return self.url_factory.url(controller, action, id=id)

    def icon(self, name):
        return self.url_factory.icon(name)

    def gettext(self, text):
        return self.translation.ugettext(text)

    def ungettext(self, singular, plural, n):
        return self.translation.ungettext(singular, plural, n)

    def format_numeric(self, nc, value):
        """Return the numeric value as a formatted string.
        If value is None, the method returns an empty string. If the
        value is already a string (should only happen if the user tried
        to input values that cannot be parsed as a float) the method
        returns that string. If the value is UNKNOWN_NUMERIC the method
        returns the localized "unknown" string. Otherwise the value
        should be a decimal object and the method formats the value with
        the format_numeric method of the renderer's formatter instance.
        """
        if value is None:
            formatted = ""
        elif isinstance(value, basestring):
            formatted = value
        elif value == UNKNOWN_NUMERIC:
            formatted = self.gettext("unknown")
        else:
            formatted = self.formatter.format_numeric(nc, value)
        return formatted

    def format_date(self, nc, value):
        """Return the date value as a formatted string.
        If value is None, the method returns an empty string. If the
        value is already a string (should only happen if the user tried
        to input values that cannot be parsed as a date) the method
        returns that string. If the value is UNKNOWN_DATE the method
        returns the localized "unknown" string. Otherwise the value
        should be a date object and the method formats the date with the
        format_date method of the renderer's formatter instance.
        """
        if value is None:
            formatted = ""
        elif isinstance(value, basestring):
            formatted = value
        elif value == UNKNOWN_DATE:
            formatted = self.gettext("unknown")
        else:
            formatted = self.formatter.format_date(nc, value)
        return formatted

    def format_time(self, nc, value):
        """Return the time value as a formatted string.
        If value is None, the method returns an empty string. If the
        value is already a string (should only happen if the user tried
        to input values that cannot be parsed as a time) the method
        returns that string. Otherwise the value should be a time object
        and the method formats the time with the format_time method of
        the renderer's formatter instance.
        """
        if value is None:
            formatted = ""
        elif isinstance(value, basestring):
            formatted = value
        else:
            formatted = self.formatter.format_time(nc, value)
        return formatted

    def format_duration(self, nc, value):
        """Return the duration/timedelta value as a formatted string.
        If value is None, the method returns an empty string. If the
        value is already a string (should only happen if the user tried
        to input values that cannot be parsed as a date) the method
        returns that string. Otherwise the value should be a timedelta
        object and the method formats the duration as a string with the
        length of the interval in minutes as a decimal number
        """
        if value is None:
            formatted = ""
        elif isinstance(value, basestring):
            formatted = value
        else:
            formatted = str(int(value.total_seconds() // 60))
        return formatted

    def render_help_link(self, nc, kind):
        return [tag("a", tabindex=-1, class_="annotation a-help", data_toggle="popover",
                    data_trigger="focus", data_html="true",
                    data_placement="top",
                    title=self.gettext("Help for field '%s'") % nc.getDescription(),
                    data_content=nc.getHelp()),
                tag("svg", class_="svg-icon-form", aria_labelledby="title", role_="img"),
                '<title id="title">' + self.gettext('Help') + '</title>'
                '<use xlink:href="#%s"></use>' % kind,
                '</svg>',
                '</a>']


    def render_required_link(self, nc, kind):
        return [tag("a", class_="annotation", 
                    title_=self.gettext("Required field"), 
                    action=kind, id=nc.getName()),
                tag("svg", class_="svg-icon-form", role_="img"),
                '<use xlink:href="#%s"></use>' % kind,
                '</svg>',
                '</a>']

    def render_evaluation_link(self, nc, kind):
        return [tag("a", class_="annotation", 
                    title_=self.gettext("Evaluation field"), 
                    action=kind, id=nc.getName()),
                tag("svg", class_="svg-icon-form", role_="img"),
                '<use xlink:href="#%s"></use>' % kind,
                '</svg>',
                '</a>']


    def render_annotation_links(self, nc):
        out=[]
        if nc.is_required():
            out.extend(self.render_required_link(nc, "required"))
        if is_evaluation_relevant(nc.getFlags()):
            out.extend(self.render_evaluation_link(nc, "evaluation"))
        if nc.hasHelp():
            out.extend(self.render_help_link(nc, "help"))
        return out

    def render_page_headings(self, all=False):
        """Render the HTML page headings for the form.
        Normally, the description of the renderer's view_root is used as
        the heading together with links for any annotations available
        for the view_root.  However, if the parameter all is given and
        true (default: False), the descriptions of the parents of the
        view_root are also rendered as headings, ordered from the
        formed-root to the view_root.
        """
        if all:
            # omit first two items of pathToRoot because those are the
            # root node and the case node.
            nodes = self.view_root.pathToRoot()[2:]
        else:
            nodes = [self.view_root]

        out = []
        for depth, node in enumerate(nodes):
            heading = "h%d" % (depth + 3,)
            out.append("<%s>%s %s %s</%s>"
                       % (heading, escape(node.getDescription()),
                          "".join(self.render_annotation_links(node)),
                          "".join(self.render_page_heading_buttons(node)),
                          heading))
        return "".join(out)

    def render_page_heading_buttons(self, node):
        return []


class HtmlRenderer(BaseRenderer):

    def __init__(self, data, view_root, url_factory, read_only=True,
                 languages=None, set_modification_attrs=None):
        super(HtmlRenderer, self).__init__(data, view_root, url_factory,
                                           read_only=read_only,
                                           languages=languages)
        self.tabindex = 3
        self.disabled_fields = set()
        self.warnings = defaultdict(list)
        self.notices = defaultdict(list)
        self.errors = defaultdict(list)
        self.widget_order = []
        self.save_next = False
        self.current_rg_deletion_id = None
        self.current_rg_add_id = None

        self.init_folding = []

        if set_modification_attrs is None:
            set_modification_attrs = {}
        self.set_modification_attrs = set_modification_attrs

        # currently the root of the view must be a page
        assert isinstance(self.view_root, PageNode)

    def next_tabindex(self):
        tabindex = self.tabindex
        self.tabindex += 1
        return str(tabindex)

    def render(self, show_next=True, hidden_fields=()):
        out = []
        self.show_next = show_next

        if not self.read_only:
            out.append(self.render_form_tag())
            out.extend(self.render_submit_buttons_top(show_next))

        for name, value in hidden_fields:
            out.append(tag("input", type="hidden", name=name, id=name,
                           value=value))
        out.extend(self.render_page(self.view_root))

        if not self.read_only:
            out.extend(self.render_submit_buttons_bottom(show_next))
            out.append("</form>")

        if len(self.init_folding):
            out.append(folding_init(self.init_folding))

        return "".join(out)

    def render_form_tag(self):
        return tag("form", id="formularpage",
                   action=self.url(controller="formularpage",
                                   action="save"),
                   enctype="multipart/form-data",
                   accept_charset="UTF-8",
                   method="POST")

    def render_submit_buttons_top(self, show_next=True):
        return self.render_submit_buttons(show_next, position="Top")

    def render_submit_buttons_bottom(self, show_next=True):
        return self.render_submit_buttons(show_next, position="Bottom")

    def render_submit_buttons(self, show_next, position='', from_tag=False):
        out = []
        if not from_tag:
            out.append(tag("div"))
        out.append(self.render_submit_button(self.gettext('Save'), name="save",
                                             css_class="btn-primary"))
        if show_next:
            out.append("&nbsp;")
            out.append(self.render_submit_button(
                       self.gettext('Save and proceed'), name="savenext",
                       css_class="btn-outline-primary"))
        if not from_tag:
            out.append('</div>')
        return out

    def render_submit_button(self, label, name, css_class=None):
        classes = ["btn",
                   css_class if css_class is not None else "btn-primary",
                   ]
        return tag("input", type="submit", name=name,
                   class_=" ".join(classes), value=label,
                   tabindex=self.next_tabindex())

    def render_page(self, nc):
        parts = []
        parts.append(self.render_children(nc))
        return parts

    def render_children(self, nc, child_renderer=None):
        if child_renderer is None:
            child_renderer = self.render_nc

        out = []
        for child in nc.getChildren():
            if isinstance(child, (RuleLeaf, VirtualLeaf)):
                continue
            content = child_renderer(child)
            if content is not None:
                args = ()
                if isinstance(content, Col):
                    args = (child.get_span(),)
                out.append(content.render(*args))
        return "".join(out)

    def update_disabled_fields(self, nc):
        if self.read_only:
            self.disabled_fields.add(self.data.get_id(nc))

    def render_nc(self, nc):
        self.check_rules(nc)
        self.update_disabled_fields(nc)
        if isinstance(nc, TextLeaf):
            return self.render_text(nc)
        if isinstance(nc, TextAreaLeaf):
            return self.render_text_area(nc)
        if isinstance(nc, DateLeaf):
            return self.render_date(nc)
        if isinstance(nc, NumericLeaf):
            return self.render_numeric(nc)
        if isinstance(nc, IntLeaf):
            return self.render_int(nc)
        if isinstance(nc, BoolLeaf):
            return self.render_bool(nc)
        if isinstance(nc, InfoLeaf):
            return self.render_info(nc)
        if isinstance(nc, SaveLeaf):
            return self.render_save(nc)
        if isinstance(nc, ChoiceNode):
            return self.render_choice(nc)
        if isinstance(nc, RadioNode):
            return self.render_radio(nc)
        if isinstance(nc, PageNode):
            return self.render_page(nc)
        if isinstance(nc, MatrixNode):
            return self.render_radio_matrix(nc)
        if isinstance(nc, RepeatNode):
            return self.render_inline_repeat_group(nc)
        if isinstance(nc, GroupNode):
            return self.render_group(nc)
        if isinstance(nc, RowNode):
            return self.render_row(nc)
        if isinstance(nc, ConditionalNode):
            return self.render_conditional(nc)
        if isinstance(nc, ReferenceNode):
            return self.render_reference(nc)
        if isinstance(nc, FileLeaf):
            return self.render_file(nc)
        if isinstance(nc, TimeLeaf):
            return self.render_time(nc)
        if isinstance(nc, DurationLeaf):
            return self.render_duration(nc)
        else:
            raise ValueError("Cannot render nc %r" % nc)

    def render_text(self, nc):
        value = self.data.get_value(nc)
        id = self.data.get_id(nc)

        out = []
        out.extend(self.render_label(nc, id))

        out.append(tag('input', name=id, id=id, value=value,
                       class_=self.error_css_class(nc) + " form-control form-control-sm",
                       type="text",
                       maxlength=nc.getMaxLength(), disabled=self.read_only,
                       **self.set_modification_attrs))
        self.render_tooltip(nc, out)
        return Col("".join(out))

    def render_text_area(self, nc):
        value = self.data.get_value(nc)
        id = self.data.get_id(nc)
        cols = nc.getCols() or "0"
        rows = nc.getRows() or "0"

        out = []
        out.extend(self.render_label(nc, id))

        out.append(tag('textarea', id=id, name=id, rows=rows, cols=cols,
                       placeholder=nc.getPlaceholder(),
                       class_=self.error_css_class(nc) + " form-control form-control-sm",
                       disabled=self.read_only, **self.set_modification_attrs))
        out.append(escape(value))
        out.append("</textarea>")
        self.render_tooltip(nc, out)
        return Col("".join(out))

    def render_choice(self, nc):
        value = self.data.get_value(nc)
        if not nc.getMultiple():
            # Representing the value of a choice field as a list of ints
            # even if the field does not allow multiple selections
            # allows us to handle the rendering in a more uniform way.
            value = [value]
        value = set(value)
        id = self.data.get_id(nc)

        out = []
        if not self.read_only and nc.getMultiple():
            out.append(hidden_multiselect_tag(id))

        if nc.getInput() == "checkboxes":
            out.extend(self.render_choice_checkboxes(nc, value))
        else:
            out.extend(self.render_choice_select(nc, value))
        return Col("".join(out))

    def render_choice_checkboxes(self, nc, value):
        multiselect_id = self.data.get_id(nc)
        out = []
        group_level = min(group_nesting_level(nc), 4)
        out.append(tag("fieldset", class_="group-level-%d" % (group_level + 1,)))
        out.append(tag("div", class_="row"))
        out.append(tag("div", class_="col-auto"))
        out.append("<legend>%s</legend>"
                   % ("".join(self.render_label(nc, self.data.get_id(nc)))))
        out.append("</div>")
        out.append("</div>")

        out.append(tag("div"))
        for node in nc.get_option_nodes():
            checkboxid = "%s@%s" % (multiselect_id, node.getValue())
            span = node.get_span()
            if span is not None:
                col_class = " col-%d" % (span,)
            else:
                col_class = ""
            out.append(tag("div", class_="form-check col-form" + col_class))
            out.append(tag('input', type="checkbox",
                           class_="form-check-input",
                           id=checkboxid, name=multiselect_id,
                           value=node.getValue(), disabled=self.read_only,
                           checked=node.getValue() in value,
                           tabindex=self.next_tabindex(),
                           **self.set_modification_attrs))
            out.append(tag('label', class_="form-check-label text-nowrap",
                           for_=checkboxid))
            out.append(escape(node.getDescription()))
            out.append('</label>')
            out.append("</div>")
        out.append("</div>")
        self.render_tooltip(nc, out)
        out.append("</fieldset>")

        return "".join(out)

    def render_choice_select(self, nc, value):
        id = self.data.get_id(nc)
        out = []
        out.extend(self.render_label(nc, id))
        out.append(tag('select', id=id, name=id,
                       multiple=bool(nc.getMultiple()), disabled=self.read_only,
                       class_=self.error_css_class(nc) + " custom-select custom-select-sm",
                       **self.set_modification_attrs))
        out.append(self.render_choice_children(nc, value))
        out.append('</select>')
        self.render_tooltip(nc, out)
        return out

    def render_choice_children(self, nc, value):
        def child_renderer(nc):
            return self.render_choice_nc(nc, value)

        return self.render_children(nc, child_renderer)

    def render_choice_nc(self, nc, value):
        if isinstance(nc, BoolLeaf):
            return self.render_choice_bool(nc, value)
        if isinstance(nc, GroupNode):
            return self.render_choice_group(nc, value)
        if isinstance(nc, ExternalChoiceListLeaf):
            return self.render_external_choice_list(nc, value)
        if isinstance(nc, ConditionalNode):
            return self.render_choice_conditional(nc, value)
        raise ValueError("Cannot render choice nc %r" % nc)

    def render_choice_group(self, nc, value):
        out = []
        out.append(tag("optgroup", label=nc.getDescription()))
        out.append("\n")
        out.append(self.render_choice_children(nc, value))
        out.append("</optgroup>\n")
        return Text("".join(out))

    def render_choice_bool(self, nc, value):
        selected = nc.getValue() in value
        out = []
        if not nc.isInvisible() or selected:
            out.append(tag("option", selected=selected, value=nc.getValue()))
            out.append(escape(nc.getDescription()))
            out.append("</option>")

        return Text("".join(out))

    def render_external_choice_list(self, nc, value):
        out = []
        for child in nc.getChildren():
            child_value = child.getValue()
            out.append(tag("option", value=child_value,
                           selected=(child_value in value)))
            out.append(escape(child.getDescription()))
            out.append("</option>")
        return Text("\n".join(out))

    def render_choice_conditional(self, nc, value):
        return self.do_render_conditional(nc,
                                          self.render_choice_children, value)

    def render_date(self, nc):
        id = self.data.get_id(nc)
        formatted = self.format_date(nc, self.data.get_value(nc))
        maxlength = "10"

        out = []
        out.extend(self.render_label(nc, id))

        css_class = " ".join(["datefield", "mpuls-datepicker",
                              self.error_css_class(nc)])

        out.append(tag("input", name=id, id=id, value=formatted,
                       type="text", maxlength=maxlength,
                       disabled=self.read_only,
                       class_=css_class + " form-control form-control-sm",
                       placeholder=nc.getPlaceholder(),
                       **self.set_modification_attrs))
        self.render_tooltip(nc, out)
        return Col("".join(out))

    def render_time(self, nc):
        id = self.data.get_id(nc)
        formatted = self.format_time(nc, self.data.get_value(nc))
        maxlength = "5"

        out = []
        out.extend(self.render_label(nc, id))

        css_class = self.error_css_class(nc)

        out.append(tag("input", name=id, id=id, value=formatted,
                       type="text", maxlength=maxlength,
                       disabled=self.read_only,
                       class_=css_class + " form-control form-control-sm",
                       placeholder=nc.getPlaceholder(),
                       **self.set_modification_attrs))
        self.render_tooltip(nc, out)
        return Col("".join(out))

    def render_duration(self, nc):
        id = self.data.get_id(nc)
        formatted = self.format_duration(nc, self.data.get_value(nc))
        maxlength = "5"

        out = []
        out.extend(self.render_label(nc, id))

        css_class = self.error_css_class(nc)
        out.append(tag("input", name=id, id=id, value=formatted,
                       type="text", maxlength=maxlength,
                       disabled=self.read_only,
                       class_=css_class + " form-control form-control-sm",
                       placeholder=nc.getPlaceholder(),
                       **self.set_modification_attrs))
        self.render_tooltip(nc, out)
        return Col("".join(out))

    def render_numeric(self, nc):
        id = self.data.get_id(nc)
        formatted = self.format_numeric(nc, self.data.get_value(nc))

        out = []
        out.extend(self.render_label(nc, id))

        out.append(tag('input', name=id, id=id, value=formatted, type="text",
                       placeholder=nc.getPlaceholder(),
                       class_=self.error_css_class(nc) + " form-control form-control-sm",
                       disabled=self.read_only, **self.set_modification_attrs))
        self.render_tooltip(nc, out)

        return Col("".join(out))

    def render_int(self, nc):
        id = self.data.get_id(nc)
        value = self.data.get_value(nc)
        if value == UNKNOWN_INT:
            value = self.gettext("unknown")

        out = []
        out.extend(self.render_label(nc, id))

        out.append(tag('input', name=id, id=id, value=value, type="text",
                       placeholder=nc.getPlaceholder(),
                       class_=self.error_css_class(nc) + " form-control form-control-sm",
                       disabled=self.read_only, **self.set_modification_attrs))
        self.render_tooltip(nc, out)

        return Col("".join(out))

    def render_bool(self, nc):
        id = self.data.get_id(nc)
        value = self.data.get_value(nc)

        out = []
        out.append(tag('input', type="checkbox",
                       class_=" ".join(["form-check-input",
                                        self.error_css_class(nc)]),
                       id=id, name=id, value="1", disabled=self.read_only,
                       checked=(value == 1), tabindex=self.next_tabindex(),
                       **self.set_modification_attrs))
        if not self.read_only:
            out.append(hidden_bool_tag(id))
        out.append(tag('label', class_="form-check-label text-nowrap", for_=id))
        out.append(escape(nc.getDescription()))
        out.append('</label>')
        self.render_tooltip(nc, out)
        return Col("".join(out), cssclass="form-check")

    def render_radio(self, nc):
        id = self.data.get_id(nc)
        value = self.data.get_value(nc)

        out = []
        out.extend(self.render_label(nc, id))

        css_class, messages = self.error_info(nc)
        if messages and not self.read_only:
            out.append(tag("div", class_=css_class))
        else:
            out.append(tag("div"))


        for child in nc.getChildren():
            if child.isInvisible():
                continue
            child_value = child.getValue()

            # the child name is usually only unique within the radio
            # node, so prefix the id used in the HTML form with the
            # radio instance id so that it hopefully is unique on the
            # page.
            radio_id = id + "." + child.getName()

            out.append('<div class="form-check form-check-inline">')
            out.append(tag('input', type="radio", id=radio_id, name=id,
                           value=child_value, disabled=self.read_only,
                           checked=(child_value == value),
                           tabindex=self.next_tabindex(),
                           class_=(self.error_css_class(nc)
                                   + " form-check-input"),
                           **self.set_modification_attrs))
            out.append(tag("label", class_="form-check-label", for_=radio_id))
            out.append(tag("a", name="f_" + radio_id))
            out.append(escape(child.getDescription()))
            out.append("</a>")
            out.append("</label>")
            out.append("</div>")

        out.append('</div>')
        self.render_tooltip(nc, out)

        return Col("".join(out))

    def render_radio_matrix(self, nc):
        legend = []

        out = []
        out.append('<div class="table-responsive">')
        out.append('<table class="radiomatrix table table-sm table-borderless">')

        # write table header
        children = nc.getChildren()[0].getChildren()
        nr_of_children = len(children) + 1

        if nc.getDescription() != "":
            out.append(tag("legend", class_="text-muted"))
            out.extend(self.render_label(nc, ""))
            out.append('</legend>')

        out.append('<thead>')
        out.append('<tr>')
        out.append('<th></th>')
        for hcol in children:
            alternative = hcol.getAlternative()
            description = hcol.getDescription()
            if not alternative:
                header = description
            else:
                header = alternative
                legend.append("<small>%s:%s</small>"
                              % (escape(alternative), escape(description)))
            out.append('<th scope="col">%s</th>' % escape(header))
        out.append('</tr>')
        out.append('</thead>')

        # write table body
        out.append('<tbody>')
        for num, row in enumerate(nc.getChildren()):
            if isinstance(row, RuleLeaf):
                continue
            self.check_rules(row)
            id = self.data.get_id(row)
            out.append('<tr>')
            out.append(tag("th", scope_="row"))
            out.extend(self.render_label(row, id))
            out.append('</th>')
            value = self.data.get_value(row)
            for col in row.getChildren():
                col_value = col.getValue()
                out.append('<td>')
                out.append(tag("input", type="radio", name=id,
                               value=col_value, disabled=self.read_only,
                               checked=(value == col_value),
                               tabindex=self.next_tabindex(),
                               class_=self.error_css_class(nc),
                               **self.set_modification_attrs))
                out.append('</td>')
            out.append("</tr>")

        # write legend
        out.append('<tr>')
        out.append(tag("td", colspan=nr_of_children))
        out.append('</td></tr>')
        if legend:
            out.append('<tr>')
            out.append(tag("td", class_="legend", colspan=nr_of_children))
            out.append(" ".join(legend))
            out.append('</td>')
            out.append('</tr>')
        out.append('<tr>')
        out.append(tag("td", class_="label", colspan=nr_of_children))
        out.append('</td>')
        out.append('</tr>')

        out.append('</tbody>')
        out.append('</table>')
        out.append('</div>')

        self.render_tooltip(nc, out)

        return Col("".join(out))

    def render_info(self, nc):
        if nc.isInvisible():
            return
        text = ""
        for item in nc.get_content():
            if isinstance(item, ExprLeaf):
                value = self.data.evaluate_expression(item)
                if isinstance(value, basestring):
                    text += value
                elif isinstance(value, datetime.date):
                    text += self.format_date(item, value)
                # todo: check for numeric values and format them
                else:
                    text += str(value)
            elif isinstance(item, basestring):
                text += item
            else:
                log.error("InfoLeaf %r contains not supported child %r",
                          nc, item)
        return self.render_info_text(text)

    def render_info_text(self, text):
        if not text:
            return Text("")
        return Col("".join([tag("small", class_="form-text text-muted",
                                id="InfoBlock"),
                            escape(text),
                            '</small>']))

    def render_save(self, nc):
        if nc.isInvisible():
            return
        if self.read_only:
            return
        id = self.data.get_id(nc)
        out = []
        out.append(tag("a", class_="anchor", id=id))
        out.extend(self.render_submit_buttons(show_next=False, position=id,
                                              from_tag=True))
        out.append('</a>')
        return Item("".join(out))

    def render_group(self, nc):
        description = nc.getDescription()

        item_kw = {}
        out = []
        if nc.isInvisible():
            if description:
                out.append("<!-- %s -->" % escape(description))
        else:
            item_kw["htmlid"] = "form-" + self.data.get_id(nc)
            if self.conditional_css_class:
                item_kw["cssclass"] = self.conditional_css_class

            # Maximum group level is 4 because that's the maximum
            # reached in JMD.
            group_level = min(group_nesting_level(nc), 4)
            out.append(tag('fieldset', id=self.data.get_id(nc),
                           class_="group-level-%d" % (group_level,)))
            if description:
                out.append(tag("div", class_="row"))
                out.append(tag("div", class_="col-auto"))
                out.append("<legend>%s</legend>"
                           % ("".join(self.render_label(nc,
                                                        self.data.get_id(nc))))
                           )

                out.append("</div>")

                out.extend(self.render_instance_rg_buttons(nc))

                out.append("</div>")
            if nc.isFolding():
                out.append(folding_helper_functions())
                out.append(tag("div", id=nc.getName()))
                self.init_folding.append('%s' % nc.getName())
            else:
                out.append("<div>")
            out.append(self.render_children(nc))
            out.append("</div></fieldset>")
            self.render_tooltip(nc, out)

        return Col("".join(out), **item_kw)

    def render_row(self, nc):
        out = [tag("div", class_="form-row")]
        out.append(self.render_children(nc))
        out.append("</div>")
        return Item("".join(out))

    def render_conditional(self, nc):
        return self.do_render_conditional(nc, self.render_children)

    def server_side_only(self, nc):
        """Return whether a conditional can be handled completely server side.

        The parameter is the conditional's node component. By default
        this method just returns False, so it's basically assuming that
        all conditionals may have to be handle by the client, too.
        Derived classes may override this method if they have better
        information. E.g. a conditional that does not reference any
        field on the current page does not need to be evaluated by the
        client at all.
        """
        return False

    def do_render_conditional(self, nc, render_children, *args):
        result = None
        editable = self.data.evaluate_conditional(nc)
        old_read_only = self.read_only
        old_conditional_css_class = self.conditional_css_class
        try:
            self.read_only = self.read_only or not editable
            if nc.isInvisible() and not editable:
                self.conditional_css_class = "invisible_conditional"
            if editable or not (nc.isInvisible() and self.server_side_only(nc)):
                result = render_children(nc, *args)
        finally:
            self.read_only = old_read_only
            self.conditional_css_class = old_conditional_css_class
        if result is not None:
            result = Text(result)
        return result

    def render_reference(self, nc):
        return self.render_nc(self.data.get_referenced_meta_node(nc))

    def render_inline_repeat_group(self, nc):
        description = nc.getDescription()
        out = []
        if not nc.isInsideFolding():
            out.append(tag("fieldset", id=self.data.get_id(nc)))
            out.append('<legend>%s %s</legend>'
                       % (escape(description),
                          "".join(self.render_annotation_links(nc))))
        out.append('<div class="repeatgroup mb-4">')
        id = self.data.get_id(nc)
        out.append(tag("a", class_=self.error_css_class(nc),
                       name="%s" % id))
        out.append("</a>")
        out.extend(self.render_repeat_list(nc))
        out.append('</div>')
        if not nc.isInsideFolding():
            out.append('</fieldset>')
        return Item("".join(out))

    def render_repeat_list(self, nc):
        show_expr = nc.getShowControls()
        min_children = nc.getMin()
        max_children = nc.getMax()
        child_rgs = self.data.get_ordered_repeat_group(nc)
        len_child_rgs = len(child_rgs)

        show_delete_button = len_child_rgs > min_children
        show_add_button = max_children is None or len_child_rgs < max_children
        if show_expr:
            var_provider = self.data.get_variable_provider()
            show_controls = show_expr.evaluate(var_provider)
            show_delete_button = show_delete_button and show_controls & 2
            show_add_button = show_add_button and show_controls & 1

        out = []
        if len_child_rgs > 0:
            for rg in child_rgs:
                rg_id = self.data.get_id(nc)
                self.data.push_node(rg)
                try:
                    out.append('<div class="col-8 repeat-entry p-3">')
                    out.append('<div class="form-group">')
                    if not self.read_only:
                        # indicate which buttons should be rendered
                        # inside the instance. The currenty
                        # implementation allows only one repeat group to
                        # have such information at any one time, because
                        # that's simple to implement and we don't
                        # actually need to support nested repeat groups
                        # here, because we currently do not have nested
                        # *inline* repeat groups. To avoid problems if
                        # ever need to support them, we assert that the
                        # assumption that we're not inside a repeat
                        # group instance yet, by checking that the
                        # instance variables are None.
                        if show_delete_button:
                            assert self.current_rg_deletion_id is None
                            self.current_rg_deletion_id = self.data.get_rg_id(rg)
                        if show_add_button:
                            assert self.current_rg_add_id is None
                            self.current_rg_add_id = rg_id
                    out.append(self.render_children(nc))
                    self.current_rg_add_id = None
                    self.current_rg_deletion_id = None
                    out.append('</div>')
                    out.append('</div>')
                finally:
                    self.data.pop_node()
        else:
            out.extend(self.render_empty_repeat_list())

        if not self.read_only and show_add_button:
            out.append('<div class="col-8" style="text-align: right; '
                       'margin-top: -20px;">')
            out.append(
                self.render_general_inline_rg_create_button(self.data.get_id(nc)))
            out.append("</div>")
        elif self.read_only and len_child_rgs == 0:
            out.extend(self.render_add_rg_info_text())
        else:
            out.append('<div class="col-8" style="text-align: right;">'
                       '&nbsp;</div>')
        return out

    def render_empty_repeat_list(self):
        out = []
        return out

    def render_add_rg_info_text(self):
        out = []
        out.append('<div class="form-row"><div class="col-xl-8">'
                   '  <small class="form-text text-muted">%s</small>'
                   '</div></div>'
                   % escape(self.gettext("To add a RG, edit the document")))
        return out

    def render_instance_rg_buttons(self, nc):
        """Render the RG create/delete buttons associated with an RG intance."""
        out = []
        if self.current_rg_deletion_id:
            out.append(self.render_inline_rg_delete_button(
                self.current_rg_deletion_id))
            self.current_rg_deletion_id = None
        if self.current_rg_add_id:
            out.append(self.render_inline_rg_create_button(
                self.current_rg_add_id))
            self.current_rg_add_id = None
        if out:
            out.insert(0, tag("div", class_="col modusfield"))
            out.append("</div>")
        return out

    def render_inline_rg_delete_button(self, rg_id):
        """Return HTML-fragment for the delete button for inline repeat groups.
        """
        out = []
        out.append(tag("button", class_="icon modusfield_icon_rg",
                       type="submit", name="del_rg|%s" % rg_id,
                       title=self.gettext('Delete RG')))
        out.append(tag("svg", class_="svg-icon", role_="img"))
        out.append('<use xlink:href="#delete" />')
        out.append("</svg>")
        out.append("</button>")
        return "".join(out)

    def render_inline_rg_create_button(self, rg_id):
        """Return HTML-fragment for the create button for inline repeat groups.
        """
        out = []
        out.append(tag("button", class_="modusfield_icon_rg",
                       type="submit", name="add_rg|%s" % rg_id,
                       title=self.gettext('Add new RG')))
        out.append(tag("svg", class_="svg-icon", role_="img"))
        out.append('<use xlink:href="#plus-circle-solid" />')
        out.append("</svg>")
        out.append("</button>")
        return "".join(out)

    def render_general_inline_rg_create_button(self, rg_id):
        """Return HTML-fragment for the create button for inline repeat groups.
        """
        out = []
        out.append(tag("button", class_="btn btn-sm ico-btn",
                       type="submit", name="add_rg|%s" % rg_id,
                       title=self.gettext('Add new RG')))
        out.append(tag("svg", class_="svg-icon", role_="img"))
        out.append('<use xlink:href="#plus-circle-solid" />')
        out.append("</svg>")
        out.append("</button>")
        return "".join(out)

    def get_inline_rg_add_text(self):
        return self.gettext('Add new RG')



    def render_tooltip(self, nc, out):
        css_classes = self.error_css_class(nc, tooltip_only=False)
        if css_classes == "warning":
            out.append(tag("small", class_="alert-warning warningtip"))
            out.extend(self.render_errors(nc))
            out.append("</small>")
        elif css_classes == "error":
            out.append(tag("small", class_="alert-danger errortip"))
            out.extend(self.render_errors(nc))
            out.append("</small>")

    def render_label(self, nc, id):
        out = []
        if isinstance(nc.getParent(), MatrixNode):
            out.append(tag("label", for_=id))
        else:
            out.append(tag("div", class_="text-nowrap"))
            out.append(tag("label", for_=id))
        if isinstance(nc, GroupNode) and nc.isFolding():
            out.append(tag("a",
                           class_="folded ",
                           name="f_%s" % id,
                           title=nc.getDescription() + ": " + self.gettext("Fold/Unfold"),
                           id="label_%s" % nc.getName(),
                           onClick="toggle_folding('%s');" % nc.getName()))
            out.append(escape(nc.getDescription()))
            out.append("</a>")
        else:
            out.append(escape(nc.getDescription()))
        out.append("</label>")
        out.append(tag("a", name="f_%s" % id, class_="anchor"))
        out.append("</a>")
        out.extend(self.render_annotation_links(nc))
        out.append("</div>")
        return out


    def render_file(self, nc):
        id = self.data.get_id(nc)

        out = []
        out.extend(self.render_label(nc, id))
        out.append(tag("input", name=id, id=id,
                       class_="form-control-file " + self.error_css_class(nc),
                       type="file", disabled=self.read_only,
                       multiple=nc.is_multiple(), **self.set_modification_attrs))
        return Col("".join(out))

    def render_errors(self, nc):
        """Render error hints for the item indicated by nc.
        The return value is a sequence of strings with an error or
        warning message if there are any.
        """
        css_class, messages = self.error_info(nc)

        out = []
        if messages:
            out.extend([", ".join(escape(msg) for msg in messages)
                        ])
        return out

    def error_info(self, nc):
        """Return error information for the item indicated by nc.

        The return value should be a tuple (CSS-CLASS, MESSAGES) where
        CSS-CLASS is a string with the css class to use when rendering
        the span with the messages. Typical classes are 'error' and
        'warning'. It should normally be the same as the class returned
        by the error_css_class method. MESSAGES is a list if strings
        containing the messages. If there are no errors, the MESSAGES
        list should be empty.
        """
        return ("error", self.get_errors(nc))

    def error_css_class(self, nc, tooltip_only=None):
        if tooltip_only is None:
            tooltip_only = self.read_only
        classes = []
        if self.has_errors(nc):
            classes.append("error")
        elif self.has_warnings(nc):
            if not tooltip_only:
                classes.append("warning")
        return " ".join(classes)

    def check_warnings(self, nc):
        """
        This method is deprecated. Please use check_rules instead.
        """
        self.check_rules(nc)

    def check_rules(self, nc):
        """Check node for warnings, notices and errors by evaluating the
        relevant rules.

        Found warnings, notices and rules are stored in the corresponding member
        dict as a list for the requested node.
        """
        full_name = self.data.full_item_name(nc)

        # add all checked widgets to warning_order, even if they have no
        # warnings or no warning rules so that the widget_order list can
        # be used to order the errors too.
        self.widget_order.append(full_name)

        warnings, notices, errors = self.data.evaluate_rules(nc)

        self.warnings[full_name].extend(rule.getValue() for rule in warnings)
        self.notices[full_name].extend(rule.getValue() for rule in notices)
        self.errors[full_name].extend(rule.getValue() for rule in errors)

    def has_warnings(self, nc):
        return bool(self.warnings.get(self.data.full_item_name(nc)))

    def get_warnings(self, nc):
        return self.warnings.get(self.data.full_item_name(nc), ())

    def has_errors(self, nc):
        """
        Returns True if the meta node has an error

        An error is present if either a rule with a error mark is invalid or the
        provided data has an error caused by a rule violation.
        """
        full_name = self.data.full_item_name(nc)
        return bool(self.errors[full_name]) or self.data.has_errors(nc)

    def get_errors(self, nc):
        """
        Returns a list of error messages for the meta node nc
        """
        messages = self.data.get_errors(nc)
        errors = self.errors.get(self.data.full_item_name(nc), [])
        return list(messages) + list(errors)

class RepeatGroupContext:
    """Context manager for rendering repeat groups.
    Abstract away fuddeling with manually pushing and popping nodes
    and make a "context-switch" explicit.
    """
    def __init__(self, data, rg):
        self.data = data
        self.rg = rg

    def __enter__(self):
        self.data.push_node(self.rg)

    def __exit__(self, type, value, traceback):
        self.data.pop_node()

class RepeatGroupRenderer(BaseRenderer):

    """Renderer for repeat group digests

    Repeat groups that contain entire form pages have a digest page that
    shows an overview of the existing instances. This overview is a list
    with one entry for each instance and buttons to delete or create new
    instances. The entries for the individual repeat group show a subset
    of the fields of the repeat group.

    Which fields are shown in an item for a repeat group and in which
    order the repeat groups are shown is determined by the digest and
    orderby attributes of the meta node of the repeat group.

    The parameter page_errors, if given, should be a dictionary mapping
    instace IDs to error dictionaries. Instances which have errors
    according to this dictionary are marked in the digest.

    The parameter digests is not used and is only there for backwards
    compatibility reasons. There is still code that passes it to the
    constructor, although no mpuls application actually uses it anymore.

    The digest could be rendered in two ways.
    Either in the classical way as a pure digest or as a flat table.
    If you want to use the flat structure, you have to set the attribute
    in the according formedtree.xml like

    <repeat flat="true" ...
    """

    def __init__(self, data, view_root, url_factory, page_errors=None,
                 read_only=True, digests=None, languages=None):
        super(RepeatGroupRenderer, self).__init__(data, view_root, url_factory,
                                                  read_only=read_only,
                                                  languages=languages)
        self.disabled_fields = set()
        self.warnings = {}
        self.notices = {}
        self.errors = {}
        self.page_errors = page_errors if page_errors is not None else {}
        self.widget_order = []

        # the root of the repeat group renderer must be a repeat node
        assert isinstance(self.view_root, RepeatNode)

        # view_root is the RepeatNode and the first child must be the
        # page that's repeated.
        self._page = self.view_root.getChildren()[0]
        assert isinstance(self._page, PageNode)

        self._fields = self.view_root.getDigest()

    def _generate_orderby_attribute(self):
        orderby = self.view_root.getOrderBy()
        if len(orderby) > 1: # Only simple orderby statements are supported atm
            raise NotImplementedError()
        field, sort_order = orderby[0]
        if sort_order not in ["asc", "desc"]: # ignore custom sorting
            sort_order = "asc"
        field_index = self._fields.index(field) + 1
        return "{},{}".format(field_index, sort_order)

    def _render_flattable(self, instances, show_controls, show_expr):
        """Renders the flat tablestructure.
        """
        orderby = self._generate_orderby_attribute()
        out = self._render_table_with(
            self._render_flattable_content(
                [],
                instances,
                show_controls,
                show_expr
            ), id_="rg-flat", orderby=orderby
        )
        out = self._render_regular_footer(out, instances, show_controls)
        return "\n".join(out)

    def _render_flattable_headers(self, out):
        """Renders the thead section of a flat table
        """
        out.append(tag("thead"))
        out.append(tag("tr"))
        out.append(tag("th")) # Add empty column for sorting | display of errors
        out.append("</th>")
        for field in self._fields:
            title = self.get_short_description(field)
            out.append(tag("th"))
            out.append(escape(title))
            out.append("</th>")
        out.append(tag("th")) # Add empty column for controls
        out.append("</th>")
        out.append("</thead>")
        return out

    def _render_flattable_body(self, out, instances, show_controls, show_expr):
        """Renders the tbody section of a flat table
        """
        out.append(tag("tbody"))
        for num, rg in enumerate(instances):
            with RepeatGroupContext(self.data, rg):
                out = out + self._render_clickable_row_with(
                    self._render_flattable_row_content([], rg, show_controls)
                )

        out.append("</tbody>")
        return out

    def _render_flattable_row_content(self, out, rg, show_controls):
        """The content of a row for flattable is rendered here.
        """
        out = self._render_index_number_cell(out, None, rg)
        for field in self._fields:
            nc = self.data.get_meta_node(field)
            value = self.format_nc_value(nc)
            if not value and isinstance(nc, DateLeaf):
                out.append(tag("td", data_sort="0"))
            elif value and isinstance(nc, DateLeaf):
                data_sort = self.data.get_value(nc)
                out.append(tag("td", data_sort=data_sort))
            else:
                out.append(tag("td", data_sort=value))
            out.append('%s' % escape(NA(value)))
            out.append("</td>")
        out = self._render_controls_cell(out, show_controls, rg)
        return out

    def _render_flattable_content(self,  out, instances, show_controls, show_expr):
        """Renders the content of the flat table.
        It contains a thead section and a tbody section.
        """
        out = self._render_flattable_headers(out)
        out = self._render_flattable_body(out, instances, show_controls, show_expr)
        return out

    def _render_table_with(self, out, id_="rg", orderby=None):
        """Renders the given table structure within table tags.
        """
        table_div = []
        table_div.append(tag("div", class_="table-responsive"))
        table_div.append(tag("table", id=id_, orderby=orderby,
                       class_="table table-sm table-hover custom-table-striped digest col-11"))

        table_div = table_div + out
        table_div.append("</table>")
        table_div.append("</div>")
        return table_div

    def _is_field_empty(self, field):
        """Determines whether field has no value.
        """
        nc = self.data.get_meta_node(field)
        value = self.format_nc_value(nc)
        return not value

    def _determine_sort_order_for_index_cell(self, rg):
        """The ordering of the first cell is defined as follows:
        If the item has all fields None it is likely that it is new
        item and should be sorted on top of others.
        If the item has page_errors it should be on top of ones without
        errors.
        Last are neither new items nor ones with page_errors

        If you sort colums in descending order, you get "NEW" > "ERROR" > "".

        """
        values_empty = [ self._is_field_empty(field) for field in self._fields]
        if all(values_empty):
            return "NEW"
        if self.page_errors.get(self.data.get_rg_id(rg)):
            return "ERROR"
        return ""

    def _render_index_number_cell(self, out, num, rg):
        """The first cell contains index number and infos about errors.
        This cell is rendered here.
        """
        data_sort = self._determine_sort_order_for_index_cell(rg)
        out.append(tag("td", data_sort=data_sort, width="5%"))
        if num is not None:
            out.append('%s' % (num + 1))
        if self.page_errors.get(self.data.get_rg_id(rg)):
            out.append(tag("svg", class_="svg-icon-alert", aria_labelledby="title", role_="img"))
            out.append('<use xlink:href="#alert" />')
            out.append(tag("title", id="title"))
            out.append(escape(self.gettext('Formular errors')))
            out.append("</title>")
            out.append('</svg>')
        out.append('</td>')
        return out

    def _render_controls_cell(self, out, show_controls, rg):
        """The rightmost cell is reserved for controls like "delete".
        This is rendered here.
        """
        out.append(tag("td", class_="icon-td"))
        if not self.read_only and show_controls & 2:
            out.append(tag("a",
                           role="button",
                           href=self.url(controller="repeatgroup",
                                         action="delete",
                                         id=self.data.get_rg_id(rg)),
                       title=self.gettext('Delete RG')))
            out.append(tag("svg", class_="svg-icon", role="img"))
            out.append('<use xlink:href="#delete" />')
            out.append("</svg>")
            out.append("</a>")
        out.append("</td>")
        return out

    def _render_regular_row_content(self, out, num, rg, show_controls):
        """Renders the digest for default render method
        """
        out = self._render_index_number_cell(out, num, rg)
        out.append(self.render_digest(rg))
        out = self._render_controls_cell(out, show_controls, rg)
        return out

    def _render_clickable_row_with(self, out):
        """Renders the output to be wrapped in a clickable row (tr) tag.
        """
        row = []
        row.append(tag("tr", class_="clickable-row", data_href=self.url(controller="navigation", action="select_branch", id=self.data.get_id(self._page))))
        row = row + out
        row.append('</tr>')
        return row

    def _render_regular_body(self, out, instances, show_controls, show_expr):
        """Renders the tbody section of the default table layout.
        """
        out.append(tag("tbody"))
        for num, rg in enumerate(instances):
            with RepeatGroupContext(self.data, rg):
                content = self._render_regular_row_content([], num, rg, show_controls)
                out = out + self._render_clickable_row_with(content)
        out.append("</tbody>")
        return out

    def _render_regular_footer(self, out, instances, show_controls):
        """Renders the footer section for a RG-table.
        """
        out.append('<div class="col-form col-11 text-right">')
        if not self.read_only and show_controls & 1:
            out.extend(self.render_add_rg_button())
        out.append('</div>')
        if not instances and self.read_only:
            out.extend(self.render_add_rg_info_text())
        out.append('</div>')
        out.append('<div class="offset-1">%s</div>'
                   % (escape(self.ungettext('%d Item found', '%d Items found',
                                            len(instances))
                             % len(instances))))
        return out

    def _render_regular(self, instances, show_controls, show_expr):
        """Renders the Repeatgroup in the default layout with digest.
        """
        out = self._render_table_with(
            self._render_regular_body(
                [],
                instances,
                show_controls,
                show_expr
            )
        )
        out = self._render_regular_footer(out, instances, show_controls)
        return "\n".join(out)

    def render(self, show_next=False, hidden_fields=()):
        show_controls = 3  # show both buttons
        show_expr = self.view_root.getShowControls()
        if show_expr:
            var_provider = self.data.get_variable_provider()
            show_controls = show_expr.evaluate(var_provider)

        instances = self.data.get_ordered_repeat_group(self.view_root)
        if self.view_root.attributes.get("flat") == "true":
            return self._render_flattable(instances, show_controls, show_expr)
        return self._render_regular(instances, show_controls, show_expr)

    def render_add_rg_button(self):
        out = []
        if not self.read_only:
            out.append(tag("a",
                           role="button",
                           title=self.gettext('Add new RG'),
                           href=self.url(controller="repeatgroup",
                                         action="new",
                                         id=self.data.get_id(self.view_root))))

            out.append(tag("svg", class_="svg-icon", role="img"))
            out.append('<use xlink:href="#plus-circle-solid" />')
            out.append("</svg>")
            out.append("</a>")
        return out

    def render_add_rg_info_text(self):
        out = []
        out.append('<div class="row">')
        out.append('<div class="col">')
        out.append(escape(self.gettext("To add a RG, edit the document")))
        out.append('</div>')
        out.append('</div>')
        return out

    def render_digest(self, rg):
        out = []
        out.append('<td>')
        for field in self._fields:
            description, value = self.get_field_info(field)
            if description is None:
                continue
            out.append('<div class="row rg-row">')
            out.append('<div class="col-md-4">')
            out.append(escape(description))
            out.append('</div>')
            out.append('<div class="col-md">')
            out.append('%s' % escape(NA(value)))
            out.append('</div>')
            out.append('</div>')
        out.append('</td>')
        return "".join(out)

    def get_short_description(self, field):
        """Retrieve the short-description from the node component.
        """
        nc = self.data.get_meta_node(field)
        return nc.getShortDescription()

    def get_title(self, field):
        """Retrieve the field-title from the node component.
        """
        nc = self.data.get_meta_node(field)
        return nc.getTitle()

    def get_field_description(self, field):
        """Retrieve the description from the node component.
        """
        nc = self.data.get_meta_node(field)
        return nc.getDescription()

    def get_field_info(self, field):
        """Return description and formatted value for the given digest field.
        If the item referred to by the field would be invisible in the
        rendered repeat group instance due to visibility conditionals,
        the method returns (None, None) to indicate that it shouldn't be
        rendered in the digest either.
        """
        nc = self.data.get_meta_node(field)

        # check whether the item would be visible according to the
        # conditionals containing it.
        for conditional in conditional_parents(nc):
            if (conditional.isInvisible()
                    and not self.data.evaluate_conditional(conditional)):
                return None, None

        return (nc.getDescription(), self.format_nc_value(nc))

    def format_nc_value(self, nc):
        """Return the formatted value for the meta node nc.
        If nc is a ReferenceNode, the value is that of the referenced
        field.
        """
        if isinstance(nc, ReferenceNode):
            nc = self.data.get_referenced_meta_node(nc)
        value = self.data.get_value(nc)
        if isinstance(nc, DateLeaf):
            value = self.format_date(nc, value)
        elif isinstance(nc, NumericLeaf):
            value = self.format_numeric(nc, value)
        if isinstance(nc, ChoiceNode):
            value = nc.getValueMapping().get(value, "?")
        return value

    def render_page_heading_buttons(self, node):
        return self.render_add_rg_button()


class ErrorRenderer(BaseRenderer):

    def __init__(self, data, meta_tree, errors, warnings, notices, widget_order,
                 url_factory, languages=None, disabled_fields=None,
                 show_error_undo=True):
        super(ErrorRenderer, self).__init__(data, None, url_factory,
                                            languages=languages)
        self.meta_tree = meta_tree
        self.errors = errors
        self.warnings = warnings
        self.notices = notices
        self.widget_order = widget_order
        if disabled_fields is None:
            disabled_fields = {}
        self.disabled_fields = disabled_fields
        self.show_error_undo = show_error_undo

    def url(self, controller, action, id=None):
        return self.url_factory.url(controller, action, id=id)

    def icon(self, name):
        return self.url_factory.icon(name)

    def in_widget_order(self, items):
        """Generate the items in the dictionary items in widget order.
        The dictionary should map widget ids to errors or warnings.
        The widget order is defined by the widget_order parameter given
        to the constructor when the ErrorRenderer was instantiated.
        """
        copy = items.copy()
        for id in self.widget_order:
            value = copy.pop(id, None)
            if value is not None:
                yield (id, value)
        if copy:
            # This shouldn't happen, but if it does happen, log a debug
            # message and yield the rest of the items in an arbitrary
            # order, so that all are at least shown to the user.
            log.debug("ErrorRenderer: not all items occur in the widget_order"
                      " list")
            for key, value in copy.iteritems():
                yield (key, value)

    def render(self):
        out = []
        out.extend(self.render_items(self.errors,
                                     self.gettext('Formular errors'),
                                     show_undo=self.show_error_undo,
                                     css_class='col-8 alert-danger'))
        out.extend(self.render_items(self.notices,
                                     self.gettext('Formular notices'),
                                     css_class='col-8 alert-info',
                                     show_undo=False,
                                     id_prefix="formed-notice"))
        out.extend(self.render_items(self.warnings,
                                     self.gettext('Formular warnings'),
                                     css_class='col-8 alert-warning',
                                     show_undo=False,
                                     id_prefix="formed-warning"))
        return "".join(out)

    def render_items(self, items, title, css_class="",
                     show_undo=True,
                     item_url_function=None, id_prefix="formed-error"):
        items = dict((key, value) for key, value in items.items() if value)
        if not items:
            return ""

        if item_url_function is None:
            item_url_function = self.error_item_url

        out = []
        out.append('<h5>')
        out.append(escape(title))
        out.append('</h5>')
        out.append('<ol>')

        has_visible_items = False
        for k, values in self.in_widget_order(items):
            li_kwargs = {}
            if "warning" in id_prefix and k in self.disabled_fields:
                li_kwargs["style"] = "display: none;"
            else:
                has_visible_items = True
            element = self.data.get_meta_node_from_item_name(k)
            description = element.getDescription()
            # the values in the items dict may either be strings or
            # lists of strings.
            if not isinstance(values, list):
                values = [values]
            for (iv, v) in enumerate(values):
                out.append(tag('li', id="%s-%d-%s" % (id_prefix, iv, k),
                               **li_kwargs))
                if show_undo:
                    out.append(tag("a", class_="alert-link",
                                   href=self.url(controller="formularpage",
                                                 action="undo", id=k)))
                    out.append(tag("svg", class_="svg-icon-from-error", aria_labelledby="title", role="img"))
                    out.append('<use xlink:href="#undo" />')
                    out.append(tag("title", id="title"))
                    out.append(escape(self.gettext('Restore last valid value')))
                    out.append("</title>")
                    out.append('</svg>')
                    out.append('</a>')
                out.append(tag("a", class_="alert-link errortip",
                               href=item_url_function(k)))
                out.append("%s:</a>" % escape(description))
                out.append(" %s" % escape(v))
                out.append("</li>")
        out.append('</ol>')

        # Surround the list with a <div>. We need to that after
        # rendering the list items because the style of the div depends
        # on whether any of them are visible.
        out.insert(0, tag("div", class_=("alert %s" % css_class),
                          role_="alert",
                          style="" if has_visible_items else "display:none;"))
        out.append('</div>')
        return out

    def error_item_url(self, full_name):
        """Return the URL for a link to the form element given by full_name"""
        return ("#f_%s" % self.data.form_name_from_full_name(full_name))
